/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */
 
/**
 * @file module/client-server.c
 * @brief Implementation of the server for client applications of DHT.
 * @author Marko Rih
 * @version $Id: client-server.c,v 1.64 2004/05/24 12:28:22 mjraiha Exp $
 */

#ifndef CS_TEST
#include "dht.h"
#else
#include "test/client-server_test.h"
#endif

#include "dht_util.h"
#include "gnunet_util.h"
#include "api_structs.h"
#include "client-server-api.h"
#include "datalayer.h"
#include "../util/networking.h"

static unsigned short apiIdCount;

static Mutex csMutex;

/*
 * Internal helper functions.
 */
static void replyFailure(ClientHandle   client,
                         unsigned short apiId,
                         unsigned short requestId,
                         unsigned short failureReason) {

  unsigned short         msgSize = sizeof(DHT_CS_REPLY_FAILURE);
  DHT_CS_REPLY_FAILURE * msg     = NULL;

  msg                        = MALLOC(msgSize);
  msg->header.size           = htons(msgSize);
  msg->header.tcpType        = htons(DHT_CS_PROTO_REPLY_FAILURE);
  msg->dhtCSHeader.apiId     = htons(apiId);
  msg->dhtCSHeader.requestId = htons(requestId);
  msg->failureReasonNumber   = htons(failureReason);
  coreAPI->sendToClient(client, (CS_HEADER*)msg);
  if (msg != NULL) FREE(msg);
}

static void replyStandard(ClientHandle   client,
                          unsigned short apiId,
                          unsigned short requestId,
                          unsigned short requestStatus) {

  unsigned short          msgSize = sizeof(DHT_CS_REPLY_STANDARD);
  DHT_CS_REPLY_STANDARD * msg     = NULL;

  msg                        = MALLOC(msgSize);
  msg->header.size           = htons(msgSize);
  msg->header.tcpType        = htons(DHT_CS_PROTO_REPLY_STANDARD);
  msg->dhtCSHeader.apiId     = htons(apiId);
  msg->dhtCSHeader.requestId = htons(requestId);
  msg->requestStatusNumber   = htons(requestStatus);
  coreAPI->sendToClient(client, (CS_HEADER*)msg);
  if (msg != NULL) FREE(msg);
}

static DHT_CS_REPLY_RESULTS * initReplyResults(unsigned short apiId,
                                               unsigned short requestId,
                                               unsigned short msgSize) {

  DHT_CS_REPLY_RESULTS * msg = NULL;

  msg                        = MALLOC(msgSize);
  msg->header.size           = htons(msgSize);
  msg->header.tcpType        = htons(DHT_CS_PROTO_REPLY_RESULTS);
  msg->dhtCSHeader.apiId     = htons(apiId);
  msg->dhtCSHeader.requestId = htons(requestId);

  return msg;
}

static void doneReplyResults(DHT_CS_REPLY_RESULTS * msg) {

  if(msg != NULL) FREE(msg);
}

static void replyStatus(ClientHandle   client,
                        unsigned short apiId,
                        unsigned short requestId,
                        unsigned short operationStatusNumber) {

  unsigned short        msgSize = sizeof(DHT_CS_REPLY_STATUS);
  DHT_CS_REPLY_STATUS * msg     = NULL;

  msg                        = MALLOC(msgSize);
  msg->header.size           = htons(msgSize);
  msg->header.tcpType        = htons(DHT_CS_PROTO_REPLY_STATUS);
  msg->dhtCSHeader.apiId     = htons(apiId);
  msg->dhtCSHeader.requestId = htons(requestId);
  msg->operationStatusNumber = htons(operationStatusNumber);
  coreAPI->sendToClient(client, (CS_HEADER*)msg);
  if (msg != NULL) FREE(msg);
}

/**
 * CS handler for asking ID for API to use.
 */
static int csApiId(ClientHandle   client,
                   CS_HEADER    * message) {

  DHT_CS_REQUEST_API_ID * msg   = NULL;
  DHT_CS_REPLY_RESULTS  * reply = NULL;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) != sizeof(DHT_CS_REQUEST_API_ID)) {
    LOG(LOG_WARNING,
      "client_server::csApiId() - invalid message from client. %d != %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_API_ID));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_API_ID*)message;

  // create and send reply message
  reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                           ntohs(msg->dhtCSHeader.requestId),
                           sizeof(DHT_CS_REPLY_RESULTS));
  MUTEX_LOCK(&csMutex);
  reply->dhtCSHeader.apiId = htons(apiIdCount++);
  MUTEX_UNLOCK(&csMutex);
  coreAPI->sendToClient(client, (CS_HEADER*)reply);
  doneReplyResults(reply);

  LOG(LOG_DEBUG, "client_server::csApiId() - done.\n");
  return OK;
}

/**
 * CS handler for creating a new DHT-table.
 */
static int csCreate(ClientHandle   client,
                    CS_HEADER    * message) {

  DHT_CS_REQUEST_CREATE * msg           = NULL;
  DHT_CS_REPLY_RESULTS  * reply         = NULL;
  DHT_TableMetaData     * tableMetaData = NULL;
  DHT_TableConfig       * tableConfig   = NULL;
  DHT_TableId           * tableId       = NULL;
  int                     errorCode     = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) <= sizeof(DHT_CS_REQUEST_CREATE)) {
    LOG(LOG_WARNING,
      "client_server::csCreate() - invalid message from client. %d <= %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_CREATE));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_CREATE*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csCreate() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // create table
  tableMetaData       = MALLOC(sizeof(DHT_TableMetaData));
  tableMetaData->name = MALLOC(msg->header.size - sizeof(DHT_CS_REQUEST_CREATE));
  strcpy(tableMetaData->name, msg->name);

  tableConfig                        = MALLOC(sizeof(DHT_TableConfig));
  tableConfig->replicationCount      = ntohl(msg->tableConfig.replicationCount); 
  tableConfig->parallelismCount      = ntohl(msg->tableConfig.parallelismCount);
  tableConfig->maximumValuesPerKey   = ntohl(msg->tableConfig.maximumValuesPerKey);
  tableConfig->expirationTimeSeconds = ntohl(msg->tableConfig.expirationTimeSeconds);
  tableConfig->flags                 = ntohl(msg->tableConfig.flags);
  tableConfig->cacheTimeMultiplier   = ntohl(msg->tableConfig.cacheTimeMultiplier);

  errorCode = create(tableConfig, tableMetaData, &tableId);

  // create reply
  if (OK == errorCode && tableId != NULL) {
    // table created successfully
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             sizeof(DHT_CS_REPLY_RESULTS) + sizeof(DHT_TableId));
    htonHashCode160Copy((HashCode160 *)&reply->data, tableId); 
    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);

    LOG(LOG_DEBUG, "client_server::csCreate() - done. tableId: %s\n",
        hash160ToString(tableId));
  } else {
    // creating table failed
    LOG(LOG_WARNING,
      "client_server::csCreate() - creating table failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (tableId             != NULL) FREE(tableId);
  if (tableMetaData->name != NULL) FREE(tableMetaData->name);
  if (tableMetaData       != NULL) FREE(tableMetaData);
  if (tableConfig         != NULL) FREE(tableConfig);
  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for joining existing DHT-table.
 */
static int csJoin(ClientHandle   client,
                  CS_HEADER    * message) {

  DHT_CS_REQUEST_JOIN  * msg         = NULL;
  DHT_CS_REPLY_RESULTS * reply       = NULL;
  DHT_NodeAddress      * nodeAddress = NULL;
  DHT_TableId          * tableId     = NULL;
  DHT_TableId          * tableIdPtr  = NULL;
  int                    errorCode   = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) != sizeof(DHT_CS_REQUEST_JOIN)) {
    LOG(LOG_WARNING,
      "client_server::csJoin() - invalid message from client. %d != %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_JOIN));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_JOIN*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csJoin() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // join table
  nodeAddress = MALLOC(sizeof(DHT_NodeAddress));
  ntohHashCode160Copy((HashCode160 *)nodeAddress, (HashCode160 *)&msg->helperNodeAddress);

  tableId = MALLOC(sizeof(DHT_TableId));
  ntohHashCode160Copy(tableId, &msg->joinedTableId);

  errorCode = join(nodeAddress, tableId);

  // create reply
  if(OK == errorCode) {
    // joined successfully
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             sizeof(DHT_CS_REPLY_RESULTS) + sizeof(DHT_TableId));
    // FIXME
    tableIdPtr  = (DHT_TableId *)reply->data;
    *tableIdPtr = msg->joinedTableId;
    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csJoin() - done.\n");
  } else {
    // joining failed
    LOG(LOG_WARNING,
      "client_server::csJoin() - joining table failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (nodeAddress != NULL) FREE(nodeAddress);
  if (tableId     != NULL) FREE(tableId);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for leaving DHT-table.
 */
static int csLeave(ClientHandle   client,
                   CS_HEADER    * message) {

  DHT_CS_REQUEST_LEAVE * msg       = NULL;
  DHT_CS_REPLY_RESULTS * reply     = NULL;
  DHT_TableId          * tableId   = NULL;
  int                    errorCode = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) != sizeof(DHT_CS_REQUEST_LEAVE)) {
    LOG(LOG_WARNING,
      "client_server::csLeave() - invalid message from client. %d != %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_LEAVE));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_LEAVE*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csLeave() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // leave table
  tableId = MALLOC(sizeof(DHT_TableId));
  ntohHashCode160Copy(tableId, &msg->leavedTableId);

  errorCode = leave(tableId);

  // create and send reply message
  if(OK == errorCode) {
    // leaved successfully
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             sizeof(DHT_CS_REPLY_RESULTS));
    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csLeave() - done.\n");
  } else {
    // leaving failed
    LOG(LOG_WARNING,
      "client_server::csLeave() - leaving table failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (tableId != NULL) FREE(tableId);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for inserting <key,value>-pair into DHT-table.
 */
static int csInsert(ClientHandle   client,
                    CS_HEADER    * message) {

  DHT_CS_REQUEST_INSERT * msg       = NULL;
  DHT_CS_REPLY_RESULTS  * reply     = NULL;
  DHT_TableId           * tableId   = NULL;
  DHT_DataContainer     * key       = NULL;
  DHT_DataContainer     * value     = NULL;
  int                     errorCode = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) <= sizeof(DHT_CS_REQUEST_INSERT)) {
    LOG(LOG_WARNING,
      "client_server::csInsert() - invalid message from client. %d <= %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_INSERT));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_INSERT*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csInsert() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // insert pair
  tableId = MALLOC(sizeof(DHT_TableId));
  ntohHashCode160Copy(tableId, &msg->targetTableId);

  key             = MALLOC(sizeof(DHT_DataContainer));
  key->dataLength = ntohs(msg->keyLength);
  key->data       = MALLOC(key->dataLength);
  memcpy(key->data,
         msg->keyAndValue,
         key->dataLength);

  LOG(LOG_DEBUG, "client-server::csInsert() - key.size=%d & key=%s.\n",
    key->dataLength, key->data);

  value             = MALLOC(sizeof(DHT_DataContainer));
  value->dataLength = ntohs(msg->valueLength);
  value->data       = MALLOC(value->dataLength);
  memcpy(value->data,
         &msg->keyAndValue[key->dataLength],
         value->dataLength);

  errorCode = insert(tableId, key, value);

  // create and send reply message
  if(OK == errorCode) {
    // inserted successfully
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             sizeof(DHT_CS_REPLY_RESULTS));
    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csInsert() - done.\n");
  } else {
    // leaving failed
    LOG(LOG_WARNING,
      "client_server::csInsert() - inserting into table failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (tableId != NULL) FREE(tableId);
  if (key     != NULL) FREE(key);
  if (value   != NULL) FREE(value);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for fetching <key,value>-pairs from DHT-table.
 */
static int csFetch(ClientHandle   client,
                   CS_HEADER    * message) {

  DHT_CS_REQUEST_FETCH * msg        = NULL;
  DHT_CS_REPLY_RESULTS * reply      = NULL;
  DHT_ResultSet        * resultSet  = NULL;
  DHT_TableId          * tableId    = NULL;
  DHT_DataContainer    * key        = NULL;
  DHT_ResultSetItem    * item       = NULL;
  unsigned short       * lengthPtr  = NULL;
  char                 * dataPtr    = NULL;
  int                    replySize  = 0;
  int                    replyCount = 0;
  int                    errorCode  = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) <= sizeof(DHT_CS_REQUEST_FETCH)) {
    LOG(LOG_WARNING,
      "client_server::csFetch() - invalid message from client. %d <= %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_FETCH));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_FETCH*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csFetch() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // fetch
  tableId = MALLOC(sizeof(DHT_TableId));
  ntohHashCode160Copy(tableId, &msg->targetTableId);

  key             = MALLOC(sizeof(DHT_DataContainer));
  key->dataLength = ntohl(msg->keyLength);
  key->data       = MALLOC(key->dataLength);
  memcpy(key->data,
         msg->key,
         key->dataLength);

  LOG(LOG_DEBUG, "client-server::csFetch() - key.size=%d & key=%s.\n",
    key->dataLength, key->data);

  errorCode = fetch(tableId, key, &resultSet);

  // create and send reply message
  if(OK == errorCode && resultSet != NULL) {
    // fetched successfully

    // calculate reply size
    replySize = sizeof(DHT_CS_REPLY_RESULTS) + sizeof(unsigned short); /* one unsigned short for valueCount (=replyCount) */
    item = resultSet->firstItem;

    while ((item != NULL && item->data != NULL) && item->data->data != NULL ) {
      replyCount++;
      replySize += sizeof(unsigned short);  // one length per result
      replySize += item->data->dataLength;  // result data
      item = item->nextItem;
    }

    // initialize reply
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             replySize);

    // FIXME: limit message size

    // write valueCount (=replyCount) to the reply message.
    lengthPtr  = (unsigned short *)reply->data;
    *lengthPtr = htons(replyCount);
    lengthPtr++;

    // create array of reply items sizes
    item = resultSet->firstItem;
    while ((item != NULL && item->data != NULL ) && item->data->data != NULL) {
      *lengthPtr = htons((unsigned short)item->data->dataLength);
      lengthPtr++;
      item = item->nextItem;
    }

    LOG(LOG_DEBUG, "client-server::csFetch() - wrote %d replies.\n", replyCount);

    // add all replys
    item = resultSet->firstItem; 
    dataPtr = (char *)lengthPtr;
    while ((item != NULL && item->data != NULL ) && item->data->data != NULL) {
      memcpy(dataPtr,
             item->data->data,
             item->data->dataLength);
      dataPtr += item->data->dataLength;
      item = item->nextItem;
    }

    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csFetch() - done.\n");
  } else {
    // fetching failed
    LOG(LOG_WARNING,
      "client_server::csFetch() - fetching from table failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (tableId   != NULL) FREE(tableId);
  if (key       != NULL) FREE(key);
  if (resultSet != NULL) freeResultSet(&resultSet);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for fetching list of tables in given DHT-node.
 */
static int csTables(ClientHandle   client,
                    CS_HEADER    * message) {

  DHT_CS_REQUEST_TABLES * msg         = NULL;
  DHT_CS_REPLY_RESULTS  * reply       = NULL;
  DHT_NodeAddress       * nodeAddress = NULL;
  DHT_TableSet          * tableSet    = NULL;
  DHT_TableSetItem      * item        = NULL;
  unsigned short        * lengthPtr   = NULL;
  char                  * dataPtr     = NULL; 
  int                     replyCount  = 0;
  int                     replySize   = 0;
  int                     errorCode   = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) != sizeof(DHT_CS_REQUEST_TABLES)) {
    LOG(LOG_WARNING,
	"client_server::csTables() - invalid message from client. %d != %d\n",
	ntohs(message->size), 
	sizeof(DHT_CS_REQUEST_TABLES));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_TABLES*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csTables() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // fetch tables
  nodeAddress = MALLOC(sizeof(DHT_NodeAddress));
  ntohHashCode160Copy((HashCode160 *)nodeAddress, (HashCode160 *)(&(msg->nodeAddress)));

  errorCode = tables(nodeAddress, &tableSet);

  // create and send reply message
  if(OK == errorCode && tableSet != NULL) {
    // fetched tables successfully

    // calculate reply size
    replySize  = sizeof(DHT_CS_REPLY_RESULTS) + sizeof(unsigned short); // data has at least 1 entry (for valueCount=replyCount).
    item       = tableSet->firstItem;

    while (item != NULL) {
      LOG(LOG_DEBUG, "client-server::csTables() - tableId: %s name: %s.\n", 
        hash160ToString(&item->tableId), item->tableMetaData.name); 
      replyCount += 2;                                    // add two replies per result
      replySize  += 2 * sizeof(unsigned short);           // two lengths per result
      replySize  += sizeof(DHT_TableId);                  // one tableId per result
      replySize  += strlen(item->tableMetaData.name) + 1; // one name per result
      item = item->nextItem;
    }
    
    // initialize reply
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             replySize);

    // FIXME: limit message size

    // write valueCount (=replyCount) to reply 
    lengthPtr  = (unsigned short *)reply->data; 
    *lengthPtr = htons(replyCount); 
    lengthPtr++;

    // create array of reply item sizes
    item       = tableSet->firstItem;
    while (item != NULL) {
      *lengthPtr = htons(sizeof(DHT_TableId));
      lengthPtr++;
      *lengthPtr = htons(strlen(item->tableMetaData.name) + 1);
      lengthPtr++;
      item = item->nextItem;
    }

    LOG(LOG_DEBUG, "client-server::csTables() - wrote %d replies.\n", replyCount); 

    // write all values (=replys)
    item       = tableSet->firstItem;
    dataPtr = (char *)lengthPtr; 
    while (item != NULL) {
      htonHashCode160Copy((HashCode160 *)dataPtr, &(item->tableId));
      dataPtr += sizeof(DHT_TableId);

      memcpy(dataPtr, 
             item->tableMetaData.name,
             strlen(item->tableMetaData.name) + 1);
      dataPtr += strlen(item->tableMetaData.name) + 1; 

      LOG(LOG_DEBUG,"client-server::csTables() - tablename %s added into reply.\n",
        item->tableMetaData.name
      );

      item = item->nextItem;
    }
    
    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csTables() - done.\n");
  } else {
    // fetching tables failed
    LOG(LOG_WARNING,
      "client_server::csTables() - fetching from table failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (nodeAddress != NULL) FREE(nodeAddress);
  if (tableSet    != NULL) freeTableSet(&tableSet);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for listing <key,value>-pairs inserted by this node.
 */
static int csInserted(ClientHandle   client,
                      CS_HEADER    * message) {

  DHT_CS_REQUEST_INSERTED * msg        = NULL;
  DHT_CS_REPLY_RESULTS    * reply      = NULL;
  DHT_TableId             * tableId    = NULL;
  DHT_DataList            * dataList   = NULL;
  DHT_DataListItem        * item       = NULL;
  unsigned short          * lengthPtr  = NULL;
  char                    * dataPtr    = NULL;
  int                       replyCount = 0;
  int                       replySize  = 0;
  int                       errorCode  = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) != sizeof(DHT_CS_REQUEST_INSERTED)) {
    LOG(LOG_WARNING,
      "client_server::csInserted() - invalid message from client. %d != %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_INSERTED));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_INSERTED*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csInserted() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // fetch list of inserted
  tableId = MALLOC(sizeof(DHT_TableId));
  ntohHashCode160Copy(tableId, &msg->insertedTableId);

  errorCode = inserted(tableId, &dataList);

  // create and send reply message
  if(OK == errorCode && dataList != NULL) {
    // fetched list of inserted successfully

    // calculate reply size
    replySize  = sizeof(DHT_CS_REPLY_RESULTS) + sizeof(unsigned short);
    item       = dataList->firstItem;

    if (item != NULL) {
      replyCount += 4;                         // add four replies per result
      replySize += 4 * sizeof(unsigned short); // four lengths per result
      replySize += item->key->dataLength;      // one key per result
      replySize += item->value->dataLength;    // one value per result
      replySize += sizeof(HashCode160);        // one hash of key
      replySize += sizeof(cron_t);             // one insertion time
      item = item->nextItem;
    }

    // initialize reply
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             replySize);

    // FIXME: limit message size

    // create array of reply items sizes
    lengthPtr  = (unsigned short *)reply->data;
    *lengthPtr = htons(replyCount);
    lengthPtr++;

    LOG(LOG_DEBUG, "client-server::csInserted() - wrote %d replies.\n", replyCount);

    item = dataList->firstItem;
    while (item != NULL) {
      *lengthPtr = htons(item->key->dataLength);
      lengthPtr++;

      *lengthPtr = htons(item->value->dataLength);
      lengthPtr++;

      *lengthPtr = htons(sizeof(HashCode160));
      lengthPtr++;

      *lengthPtr = htons(sizeof(cron_t));
      lengthPtr++;

      item = item->nextItem;
    }

    // add all replys
    item    = dataList->firstItem;
    dataPtr = (char *)lengthPtr;
    while (item != NULL) {
      memcpy(dataPtr,
             item->key->data,
             item->key->dataLength);
      dataPtr += item->key->dataLength;

      memcpy(dataPtr,
             item->value->data,
             item->value->dataLength);
      dataPtr += item->value->dataLength;

      htonHashCode160Copy((HashCode160 *)dataPtr, &(item->uniqueReference.hashCode));
      dataPtr += sizeof(HashCode160);

      memcpy(dataPtr,
             &item->uniqueReference.insertionTime,
             sizeof(cron_t));
      dataPtr += sizeof(cron_t);

      LOG(LOG_WARNING, "client-server::csInserted() - key: %s length: %d value: %s length: %d uref: %s itime %llu.\n",
        item->key->data,
        item->key->dataLength,
        item->value->data,
        item->value->dataLength,
        hash160ToString(&item->uniqueReference.hashCode),
        &item->uniqueReference.insertionTime);

      item = item->nextItem;
    }

    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csInserted() - done.\n");
  } else {
    // fetching list of inserted failed
    LOG(LOG_WARNING,
      "client_server::csInserted() - fetching list of inserted failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (tableId  != NULL) FREE(tableId);
  if (dataList != NULL) freeDataList(&dataList);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for removing <key,value>-pairs inserted by this node.
 */
static int csDrop(ClientHandle   client,
                  CS_HEADER    * message) {

  DHT_CS_REQUEST_DROP     * msg       = NULL;
  DHT_CS_REPLY_RESULTS    * reply     = NULL;
  DHT_TableId             * tableId   = NULL;
  DHT_StoredDataReference * dataRef   = NULL;
  int                       errorCode = OK;

  // check size of incoming message and convert it if ok
  if (ntohs(message->size) != sizeof(DHT_CS_REQUEST_DROP)) {
    LOG(LOG_WARNING,
      "client_server::csDrop() - invalid message from client. %d != %d\n",
      ntohs(message->size), sizeof(DHT_CS_REQUEST_DROP));
    replyFailure(client, 0, 0, FAILURE_REASON_MSG_LENGTH_MISMATCH);
    return SYSERR;
  }
  msg = (DHT_CS_REQUEST_DROP*)message;

  // accept request
  LOG(LOG_DEBUG, "client_server::csDrop() - accepted request.\n");
  replyStandard(client,
                ntohs(msg->dhtCSHeader.apiId),
                ntohs(msg->dhtCSHeader.requestId),
                REQUEST_STATUS_ACCEPTED);

  // drop
  tableId = MALLOC(sizeof(DHT_TableId));
  ntohHashCode160Copy(tableId, &msg->insertedTableId);

  dataRef = MALLOC(sizeof(DHT_StoredDataReference));
  ntohHashCode160Copy(&dataRef->hashCode, &msg->droppedDataReference.hashCode);
  dataRef->insertionTime = ntohs(msg->droppedDataReference.insertionTime);

  errorCode =drop(tableId, dataRef);
  
  // create and send reply message
  if(OK == errorCode) {
    // dropped successfully
    reply = initReplyResults(ntohs(msg->dhtCSHeader.apiId),
                             ntohs(msg->dhtCSHeader.requestId),
                             sizeof(DHT_CS_REPLY_RESULTS));
    coreAPI->sendToClient(client, (CS_HEADER*)reply);
    doneReplyResults(reply);
    LOG(LOG_DEBUG, "client_server::csDrop() - done.\n");
  } else {
    // fetching list of inserted failed
    LOG(LOG_WARNING,
      "client_server::csDrop() - drop failed.\n");
    replyFailure(client,
                 ntohs(msg->dhtCSHeader.apiId),
                 ntohs(msg->dhtCSHeader.requestId),
                 errorCode);
  }

  if (tableId != NULL) FREE(tableId);

  return errorCode == OK ? OK : SYSERR;
}

/**
 * CS handler for querying status of operations in this node.
 */
static int csStatus(ClientHandle client,
                    CS_HEADER * message) {

  LOG(LOG_FAILURE, "client_server::csStatus() - doing nothing for now.\n");
  return OK;
}

/**
 * CS handler for handling exiting client.
 */
static void csClientExit(ClientHandle client) {

  LOG(LOG_FAILURE, "client_server::csClientExit() - doing nothing for now.\n");
}

/**
 * Initialization of CS server handles and memory structures.
 */
int initClientServer(CoreAPIForApplication * capi) {

  int status = OK;

  MUTEX_CREATE(&csMutex);
  apiIdCount = 42;

  LOG(LOG_DEBUG, "client_server::initClientServer() - begin.\n");

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_API_ID,
                                            (CSHandler) &csApiId))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_CREATE,
                                            (CSHandler) &csCreate))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_JOIN,
                                            (CSHandler) &csJoin))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_LEAVE,
                                            (CSHandler) &csLeave))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_INSERT,
                                            (CSHandler) &csInsert))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_FETCH,
                                            (CSHandler) &csFetch))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_TABLES,
                                            (CSHandler) &csTables))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_INSERTED,
                                            (CSHandler) &csInserted))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_DROP,
                                            (CSHandler) &csDrop))
    status = SYSERR;

  if (SYSERR == capi->registerClientHandler(DHT_CS_PROTO_REQUEST_STATUS,
                                            (CSHandler) &csStatus))
    status = SYSERR;

  if (SYSERR == capi->registerClientExitHandler(&csClientExit))
    status = SYSERR;

  if (status == SYSERR) {
    LOG(LOG_FAILURE,
      "client_server::initClientServer() - registering handlers failed.\n");
  } else{
    LOG(LOG_DEBUG, "client_server::initClientServer() - done.\n");
  }

  return status;
}

/**
 * Unregisters handlers, cleans memory structures etc when this node exits.
 */
int doneClientServer(CoreAPIForApplication * capi) {

  MUTEX_DESTROY(&csMutex);

  LOG(LOG_DEBUG, "client_server::doneClientServer() - begin.\n");

  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_API_ID,
                                (CSHandler) &csApiId);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_CREATE,
                                (CSHandler) &csCreate);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_JOIN,
                                (CSHandler) &csJoin);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_LEAVE,
                                (CSHandler) &csLeave);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_INSERT,
                                (CSHandler) &csInsert);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_FETCH,
                                (CSHandler) &csFetch);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_TABLES,
                                (CSHandler) &csTables);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_INSERTED,
                                (CSHandler) &csInserted);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_DROP,
                                (CSHandler) &csDrop);
  capi->unregisterClientHandler(DHT_CS_PROTO_REQUEST_STATUS,
                                (CSHandler) &csStatus);
  capi->unregisterClientExitHandler(&csClientExit);

  LOG(LOG_DEBUG, "client_server::doneClientServer() - done.\n");

  return OK;
}
