 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.
 
      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */
 
/**
 * This file provides the implementation of the localstorage of a DHT module.
 * 
 * @file module/datalayer.h
 * @brief implementation of the localstorage of a DHT module.
 * @author Simo Viitanen
 * @version $Id: datalayer.c,v 1.36 2004/05/23 14:03:17 szviitan Exp $ 
 */

#include "datalayer.h"
#include "datalayer_persistance.h"
#include "api_structs.h"
#include "dht_util.h"
#include "api_errorcodes.h"
#include <gnunet_util.h>

#define OWN_DATA_FLAG   (1<<0)
#define REPUBLISH_FLAG  (1<<1)

#define INITIAL_TABLES_ARRAY_SIZE 2

/* implementation of a table*/

/**
 * 
 * @brief datastructure for one node in table.  
 **/ 
typedef struct tnode_t {
  DHT_LocalStoreUnit   * storedUnit;
  struct tnode_t      * next;
  struct tnode_t      * prev;
  /* TODO finer datastructures (implementation here ??)*/
} TableNode;

/**
 * @brief the table struct.
 **/ 
typedef struct table_t {
  Mutex               lock;
  DHT_TableDef        * tableDef;
  TableNode           * first;
} LocalStoreTable;

/* implementation of searchresult */
/**
 * 
 * @brief datastructure for one node in search result.
 **/ 
typedef struct snode_t {
  /**
   * the tablenode presenting a 'hit' in the search result.
   **/ 
  TableNode           * tableNode;
  struct snode_t      * next;
} SearchResultNode;

/**
 * @brief iterator for a result.
 **/ 
struct LocalStoreResultIter_t {
  LocalStoreTable     * table;
  SearchResultNode    * currentResult;
};
/**
 * @brief iterator for a tables.
 **/ 
struct LocalStoreTableIter_t {
  int index;
};

static struct LocalStoreResultIter_t * empty_iter = NULL;
/**
 * Array of active LocalStoreTables. 
 **/
static LocalStoreTable ** tables = NULL;
static int tablesCount;
static int tablesArrayLength;

/**
 * Lock for access to mutable state of the module.
 * Mostly for accessing and changing the tables array.
 **/
static Mutex tablesLock;

/* the methods for handling nodes
 * links two table nodes first comes seconds prev and second comes firsts next
 * nodes can be NULL
 */
static void linkNodes(TableNode * first, TableNode * second) {
  if(first != NULL) {
    if(second != NULL) {
      second->prev = first;
    }
    first->next = second;
  } else { /* first == NULL */
    if(second != NULL) {
      second->prev = first;
    }
  }
}


/**
 * 
 * returns the index in the tables array
 * return -1 if an error occured or the id is not found
 **/ 
static int getTableForId(const DHT_TableId * tableId)
{
  int ind;

#if DEBUG_DHT_DATALAYER
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::getTableForId(tableId: %s)\n", hex.data);
#endif
  MUTEX_LOCK(&tablesLock);
  for (ind = 0; ind < tablesCount; ind++) {
    if (equalsHashCode160(&tables[ind]->tableDef->tableId, tableId)) {
      break;                                      /* break for */
    }
  }
  MUTEX_UNLOCK(&tablesLock);

  if (ind>=tablesCount) {
#if !DEBUG_DHT_DATALAYER
    HexName hex;
    hash2hex(tableId, &hex);
#endif
    LOG(LOG_WARNING, 
          "WARNING: datalayer::getTableForId() no such table "
          "(tableId: %s) \n",
           hex.data);
    return -1;
  } else {
      return ind;
  }

}

/**
 * 
 **/
static TableNode * new_TableNode(DHT_LocalStoreUnit * wrappedUnit) {
  TableNode         * tNode;

#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::new_TableNode() - operation started \n");
#endif
  tNode             = MALLOC(sizeof(TableNode));    /*malloc:TableNode*/
  tNode->storedUnit = wrappedUnit;
  tNode->next = tNode->prev = NULL;
#if DEBUG_DHT_DATALAYER
    HexName hex;
    hash2hex(&wrappedUnit->hash, &hex);
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::new_TableNode() - allocated node %s \n",hex.data);
#endif
  return tNode;
}
static void free_DataStoreUnit(DHT_LocalStoreUnit * theUnit) {
  FREE(theUnit->data->data);                       /*free:void*/
  FREE(theUnit->data);                               /*free:DHT_DataContainer*/
//  FREE(theUnit->hash);                               /*free:HashCode160*/
  FREE(theUnit);                                     /*free:DHT_LocalStoreUnit*/
}

static void free_TableNodeUnit(TableNode * tNode) {
  free_DataStoreUnit(tNode->storedUnit);
  FREE(tNode);                                       /*free:TableNode*/
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::free_TableNodeUnit() - operation ended \n");
#endif
}


/**
 * 
 **/
static SearchResultNode * new_SearchResultNode(TableNode * forTableNode) {
  SearchResultNode  * rNode;

#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::new_SearchResultNode()\n");
#endif
  rNode             = MALLOC(sizeof(SearchResultNode)); 
                      /* malloc:SearchResultNode*/
  rNode->tableNode  = forTableNode;
  rNode->next       = NULL;
  return rNode;
}


/**
 * creates a new iterator for a result in one table.
 * the iterator first points to a dummy search node, that indicates the 
 * beginning search. Dummy searchnode points to null table node.
 * 
 * @return LocalStoreResultIter
 **/
static LocalStoreResultIter * new_Iter(LocalStoreTable * table) {
  LocalStoreResultIter  * rIter;

#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::new_Iter()\n");
#endif
  rIter                 = MALLOC(sizeof(LocalStoreResultIter)); 
                            /* malloc:LocalStoreResultIter*/
  rIter->table          = table;
  /* put the dummy node at the beginning */
  rIter->currentResult  = new_SearchResultNode(NULL);
  
  return rIter;
}



/***************** helper functions for datastructures **********************/

/**
 * Constructor for DHT_LocalStoreUnit . Creates a copy of parameters so the 
 * memory allocation becomes entirely datalayers own consern. 
 * 
 * 
 * @param HashCode160 * hash - the hash code of the key
 * @param DHT_DataContainer * data - the key and value combined
 **/
DHT_LocalStoreUnit * new_DHT_LocalStoreUnit(
      HashCode160       * theHash, 
      DHT_DataContainer * theData) 
{
  DHT_LocalStoreUnit * newUnit;
  DHT_DataContainer * dataCopy;
  /* copy the datacontainer */
  dataCopy    = MALLOC(sizeof(DHT_DataContainer));  /*malloc:DHT_DataContainer*/
  dataCopy->dataLength = theData->dataLength;

  dataCopy->data    = MALLOC(theData->dataLength);    /*malloc:void*/
#if DEBUG_DHT_DATALAYER
  HexName hex;
  hash2hex(theHash, &hex);
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::new_DHT_LocalStoreUnit() - for hash #%s\n",hex.data);
#endif
  memcpy(dataCopy->data,theData->data,theData->dataLength);
  /* create the dataunit */
  newUnit           = MALLOC(sizeof(DHT_LocalStoreUnit));/*malloc:DHT_LocalStoreUnit*/
  memcpy(&newUnit->hash,theHash,sizeof(HashCode160));
  newUnit->data     = dataCopy;
  newUnit->flags    = 0;
  newUnit->created  = 0;
  newUnit->timeout  = 0;

  return newUnit;
}


/**
 * Iterating resultset. The iterating takes care of the inner datastructures by 
 * freeing them as it goes along. Ending returns NULL so results can be tested 
 * against NULL. Reaching end also frees the resultiterator. Example of usage
<pre>

</pre>
 * 
 * @param result returned by any of the search methods
 * @return the next DHT_LocalStoreUnit  in the set. NULL if there is no more 
 * results in the set.
 **/
DHT_LocalStoreUnit  * nextUnit(LocalStoreResultIter ** result) 
{
  SearchResultNode      * snode;
  LocalStoreResultIter  * iter = *result;
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::nextUnit()\n");
#endif

  if(iter == NULL) {                      /* iteration has passed the end */
    LOG(LOG_WARNING, 
      "WARNING: datalayer::nextUnit() - passed iterator was NULL \n");
    return NULL;
  }

  snode             = iter->currentResult;
  if (snode == NULL) { /* this shouldn't happen, but for some wierd insident it could */
    LOG(LOG_WARNING, 
      "WARNING: datalayer::nextUnit() - iterator was pointing to NULL\n");
    FREE(iter);                             /* free:LocalStoreResultIter    */
    *result        = empty_iter;            /* avoid segfaults */
    return NULL;
  } 
#if 0
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::nextUnit() - advancing iterator\n");
#endif
  iter->currentResult = snode->next;        /* advance iterator             */
  FREE(snode);                              /* free:SearchResultNode        */

  if (iter->currentResult == NULL) {        /* the end reached              */
#if DEBUG_DHT_DATALAYER
    LOG(LOG_DEBUG,
      "DEBUG: datalayer::nextUnit() - end was reached. freeing iterator\n");
#endif
    FREE(iter);                             /* free:LocalStoreResultIter    */
    *result        = empty_iter;            /* Set to NULL to indicate end  */
    return NULL;                            /* of search. also return NULL  */
  } else {
#if 0
    LOG(LOG_DEBUG,
      "DEBUG: datalayer::nextUnit() - return next result\n");
#endif
    return iter->currentResult->tableNode->storedUnit;
  }
}

/**
 * Iterating tableset. 
 * 
 * @param TableSet returned by ListTables
 * @return the next DHT_TableDef in the set. NULL if there is no more results 
 * in the set.
 **/
DHT_TableDef * nextTable(LocalStoreTableIter ** iterator) 
{
  LocalStoreTableIter * iter = *iterator;
  DHT_TableDef * result = NULL;
  
  if (iter == NULL) { /* error we have propably reached the end, but called next*/
    LOG(LOG_WARNING, 
        "WARNING: datalayer::nextTable() - table iterator was NULL "
        "(reached end or an error)\n");
    return NULL;
  }
  if (iter->index < tablesCount ) {
    result = tables[iter->index++]->tableDef; /* advance iterator ans return previous */
#if DEBUG_DHT_DATALAYER
    LOG(LOG_DEBUG, "DEBUG: datalayer::nextTable(tableId: %s, name: %s)\n",
      hash160ToString(&result->tableId), result->metadata.name);
#endif
    return result;
  } else {
    FREE(iter);                       /*free:LocalStoreTableIter*/
    *iterator = NULL;
    return NULL;
  }
}



int isMyData(DHT_LocalStoreUnit * data) {
  return (data->flags & OWN_DATA_FLAG) != 0;
}
void setMyData(DHT_LocalStoreUnit * data, int trueOrFalse) {
  if(trueOrFalse) {
    /* TODO finer datasructures (insert in own unit list) */
    data->flags = (data->flags | OWN_DATA_FLAG);
  } else {
    data->flags = (data->flags & ~OWN_DATA_FLAG);
  }
}
int isRepublishable(DHT_LocalStoreUnit * data) {
  return (data->flags & REPUBLISH_FLAG) != 0;
}
void setRepublishable(DHT_LocalStoreUnit * data, int trueOrFalse) {
  if(trueOrFalse) {
    data->flags = (data->flags | REPUBLISH_FLAG);
  } else {
    data->flags = (data->flags & ~REPUBLISH_FLAG);
  }
}


/****************** the actual interface functions *************************/


/**
 * Lists all tables this storage unit contains.
 * 
 * @return set of tables created to this unit.
 */
DHT_LocalTableSet localStoreListTables() 
{
  DHT_LocalTableSet     set;
  LocalStoreTableIter * iter;
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG, "DEBUG: datalayer::localStoreListTables()\n");
#endif
  iter            = MALLOC(sizeof(LocalStoreTableIter));
  iter->index     = 0;
  set.iterator    = iter;
  set.returnCode  = OK;
  return set;
}


/**
 * Function localStoreCreateTable creates a new table to the storage unit.
 * 
 * @param const DHT_TableId * table_id
 * @param const DHT_TableMetaDatan metadata
 * @param const DHT_TableConfig config
 *
 * @return int containing the successing information
 */
int localStoreCreateTable(
      const DHT_TableId       * tableId, 
      const DHT_TableMetaData * metadata,
      const DHT_TableConfig   * config) 
{
  LocalStoreTable   * newTable;
  DHT_TableDef      * tableDef;

  /* TODO: check if store has room for a new table */
  if(0) {
    return DHT_ERRORCODES__DATALAYER_STORAGE_FULL;
  }
  
#if DEBUG_DHT_DATALAYER
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: datalayer::localStoreCreateTable(tableId: %s)\n",hex.data);
#endif
  tableDef                 = MALLOC(sizeof(DHT_TableDef));      /*malloc:DHT_TableDef*/
//  tableDef->tableId        = MALLOC(sizeof(DHT_TableId));
//  tableDef->metadata       = MALLOC(sizeof(DHT_TableMetaData)); 
  tableDef->metadata.name = MALLOC(strlen(metadata->name)+1); 
//  tableDef->config         = MALLOC(sizeof(DHT_TableConfig)); 
  
  memcpy(&tableDef->tableId, tableId, sizeof(DHT_TableId)); 
  memcpy(&tableDef->config, config, sizeof(DHT_TableConfig)); 
  strcpy(tableDef->metadata.name, metadata->name); 

  /* TODO persistance (create) */
  newTable          = MALLOC(sizeof(LocalStoreTable));/*malloc:LocalStoreTable*/

  MUTEX_CREATE_RECURSIVE(&newTable->lock);
  newTable->tableDef   = tableDef;
  newTable->first   = NULL;
  /* TODO finer datastructures (init them here) */

  MUTEX_LOCK(&tablesLock);
  if (tablesCount == tablesArrayLength) {
#if DEBUG_DHT_DATALAYER
    LOG(LOG_DEBUG,
       "DEBUG: datalayer::localStoreListTables() - "
       "growing store capasity from %d to %d\n",
        tablesArrayLength,
        tablesArrayLength * 2);
#endif
    GROW(tables,
         tablesArrayLength,
         tablesArrayLength * 2);
    tablesArrayLength *= 2;
  }
  tables[tablesCount] = newTable;
  tablesCount++;
  MUTEX_UNLOCK(&tablesLock);

  return OK;
}

/***
 * Method localStorePut puts a constructed DHT_LocalStoreUnit to the specified table
 *
 * @param const DHT_TableId * table_id, 
 * @param DHT_LocalStoreUnit  * unit_to_store
 * 
 * @return int
 */
int localStorePut(const DHT_TableId * tableId, 
                  DHT_LocalStoreUnit  * unitToStore) 
{
  LocalStoreTable         * table;
  TableNode               * newNode;
  TableNode               * oldNode;
  DHT_LocalStoreUnit      * oldUnit;
  int ind;
  int keyCount;

  /* check if store has room for a new unit */
  if(0) {
    free_DataStoreUnit(unitToStore);
    LOG(LOG_FAILURE, 
      "WARNING: datalayer::localStorePut - local datastore full (some info)\n");
    
    return DHT_ERRORCODES__DATALAYER_STORAGE_FULL;
  }
  

  ind                     = getTableForId(tableId);
  if (ind == -1) {        /*getTableForId didn't find a matching table */
    return DHT_ERRORCODES__TABLE_NOT_FOUND;
  }
  table                   = tables[ind];
  /* this are needed for logging */
#if DEBUG_DHT_DATALAYER
  HexName hexTableId;
  hash2hex(&table->tableDef->tableId, &hexTableId);
  HexName hexUnitHash;
  hash2hex(&unitToStore->hash, &hexUnitHash);
  LOG(LOG_DEBUG, 
      "DEBUG: datalayer::localStorePut(tableId: %s, hash: %s)",
      hexTableId.data, hexUnitHash.data);
#endif
  newNode                 = new_TableNode(unitToStore);
  keyCount                = table->tableDef->config.maximumValuesPerKey;
  MUTEX_LOCK(&table->lock);
  oldNode = table->first;
  while (oldNode != NULL) {
    oldUnit = oldNode->storedUnit;
    /* compare keys if match */
    if (equalsHashCode160(&oldUnit->hash,&unitToStore->hash)) {
      /* if also content matches unit is not stored */
      if (equalsDataContainers(oldUnit->data, unitToStore->data)) {
#if !DEBUG_DHT_DATALAYER
        HexName hexTableId;
        hash2hex(&table->tableDef->tableId, &hexTableId);
        HexName hexUnitHash;
        hash2hex(&unitToStore->hash, &hexUnitHash);
#endif
        LOG(LOG_INFO, 
          "INFO: datalayer::localStorePut - in table (tableId: %s) a "
          "unit found with identical data. Unit was not stored!  #%s.\n",
          &hexTableId, &hexUnitHash);
        MUTEX_UNLOCK(&table->lock);
        free_DataStoreUnit(unitToStore);
        return DHT_ERRORCODES__DUPLICATE_DATA_UNIT;
      } else { /* only keys were the same */
        keyCount--;
        if(keyCount == 0) {
#if !DEBUG_DHT_DATALAYER
          HexName hexTableId;
          hash2hex(&table->tableDef->tableId, &hexTableId);
          HexName hexUnitHash;
          hash2hex(&unitToStore->hash, &hexUnitHash);
#endif
          LOG(LOG_INFO, 
            "INFO: datalayer::localStorePut - in table (tableId: %s) a "
            "keystorage exceeded for id #%s. (limit was %d).\n",
            hexTableId.data, hexUnitHash.data,
            table->tableDef->config.maximumValuesPerKey
          );
          MUTEX_UNLOCK(&table->lock);
          free_DataStoreUnit(unitToStore);
          return DHT_ERRORCODES__DATALAYER_KEYSTORAGE_FULL;
        }
      }
    }
    oldNode =  oldNode->next;
  }
  linkNodes(newNode, table->first);
  table->first            = newNode;
  /* TODO finer datastructures (update them here) */
  MUTEX_UNLOCK(&table->lock);
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG, 
     "Debug: datalayer::localStorePut - unlocked table lock for tableId: %s\n",
     hexTableId.data);
#endif
  /* TODO persistance (put) */

  return OK;
}

/**
 * Method localStoreDelete deletes a unit that has been locally stored from 
 * a table. The idea is that after a search result 
 * 
 * @param const DHT_TableId * table_id
 * @param LocalStoreResultIter * result_pointing_to_removable
 *
 * @return int containing the successing information
 *
 */
int localStoreDelete(const DHT_TableId * tableId, 
                     LocalStoreResultIter * iterator) 
{
  TableNode         * tnode;
  SearchResultNode  * snode;
  
#if DEBUG_DHT_DATALAYER
  HexName hexTableId;
  HexName hexUnitHash;
  hash2hex(tableId, &hexTableId);

  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDelete(tableId: %s)\n",
    hexTableId.data);
#endif
  if (iterator == NULL) {             /* when iterator had the end reached it */
    LOG(LOG_WARNING,                  /* was nullified */
      "FAILURE: result iterator was NULL \n");
    return SYSERR;  
	}
  
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDelete\n");
#endif
  snode = iterator->currentResult;
  if(snode == NULL) {                   /* the snode was propably allready */
    LOG(LOG_WARNING ,                   /* deleted thru the iterator       */
        "FAILURE: datalayer::localStoreDelete - result iterator was "
        "inconsistent (pointing to search result NULL) \n");
    /*cleanIterator (); TODO think */
    return SYSERR;  
  }
  
  tnode = snode->tableNode;
  if(tnode == NULL) {                    /* it was not iterated */
    LOG(LOG_FAILURE,
        "FAILURE: result iterator was pointing to a table unit that was NULL\n");
    return SYSERR;  
  }

#if DEBUG_DHT_DATALAYER
  hash2hex(&tnode->storedUnit->hash, &hexUnitHash);
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDelete() - unlink tablenode (hash: %s)\n",
    hexUnitHash.data
    );
#endif
  MUTEX_LOCK(&iterator->table->lock);      /* lock the whole table and        */
  if (tnode == iterator->table->first) {   /* unlink the tablenode            */
    iterator->table->first = tnode->next;
  }
  linkNodes(tnode->prev, tnode->next);
  MUTEX_UNLOCK(&iterator->table->lock);    /* unlock the table                */
  
  /* TODO persistance (delete) */
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDelete() - remove persistance \n");
#endif
  persistentRemove(tableId, tnode->storedUnit);
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDelete() - free memory \n");
#endif
  free_TableNodeUnit(tnode);
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDelete() - "
    "set search node to point to null \n");
#endif
  snode->tableNode = NULL;                /* set result_nodes pointer to NULL */
  return OK;
}

/**
 * Method localStoreGet
 * 
 * @param const DHT_TableId * table_id
 * @param HashCode160 key_to_get
 *  
 * @return DHT_LocalStoreResultSet containing the results
 */
DHT_LocalStoreResultSet localStoreGet(const DHT_TableId * tableId, 
                                      HashCode160 * keyToGet) 
{
  DHT_LocalStoreResultSet resultSet;
  LocalStoreTable         * table;
  TableNode               * tnode;
  SearchResultNode        * snode;
  LocalStoreResultIter    * resultIter;
  int ind;

#if DEBUG_DHT_DATALAYER
  HexName hexTableId;
  HexName hexUnitHash;
  hash2hex(tableId, &hexTableId);
  hash2hex(keyToGet, &hexUnitHash);

  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreGet(tableId: %s, hash: %s) \n",
      hexTableId.data, hexUnitHash.data);
#endif
  resultIter              = NULL;
  resultSet.iterator      = empty_iter;
  ind                     = getTableForId(tableId);
  if (ind == -1) {        /*getTableForId didn't find a matching table */
    resultSet.returnCode  = DHT_ERRORCODES__TABLE_NOT_FOUND;
    return resultSet;
  }
  table                   = tables[ind];
  /* initialise iterator and put the snode point to dummy node in iterator */
  resultIter              = new_Iter(table);
  snode                   = resultIter->currentResult;

#if 0
  /*DEBUG_DHT_DATALAYER*/
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreGet() - getLock \n");
#endif
  MUTEX_LOCK(&table->lock);
  /* TODO finer datasructures (hash table for faster lookups) */
  tnode                   = table->first;  
  while(tnode != NULL) {
    if (equalsHashCode160(&tnode->storedUnit->hash, keyToGet)) {
      snode->next         = new_SearchResultNode(tnode);
      snode               = snode->next;
    }
    tnode                 = tnode->next;
  }
  MUTEX_UNLOCK(&table->lock);

  if (snode == resultIter->currentResult) { /* no hits */
#if DEBUG_DHT_DATALAYER
      LOG(LOG_DEBUG,
        "DEBUG: datalayer::localStoreGet() - no hits\n");
#endif
    nextUnit(&resultIter);      /* nullifies and frees the iterator */
    resultSet.iterator      = empty_iter;   /* avoid segfaults */
    resultSet.returnCode    = DHT_ERRORCODES__EMPTY_RESULT;
  } else {
#if DEBUG_DHT_DATALAYER
      LOG(LOG_DEBUG,
        "DEBUG: datalayer::localStoreGet() - there was hits!!\n");
#endif
    resultSet.iterator      = resultIter;
    resultSet.returnCode    = OK;
  }
  return resultSet;
}

/**
 * Helper method for various 
 **/
static int freeLocalTable(int index) 
{
  LocalStoreTable         * table;
  TableNode               * tnode1,
                          * tnode2;

#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::freeLocalTable() - operation called \n");
#endif

  if (index == -1 || index >= tablesCount) {        /*getTableForId didn't find a matching table */
    return DHT_ERRORCODES__TABLE_NOT_FOUND;
  }
  table                   = tables[index];

  MUTEX_DESTROY(&table->lock);
  /* TODO finer datasructures (free them here) */
  tnode1                  = table->first;  
  while(tnode1 != NULL) {
    tnode2                = tnode1->next;
    free_TableNodeUnit(tnode1);
    tnode1                = tnode2;
  }
  MUTEX_LOCK(&tablesLock);
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::freeLocalTable() - unlinking table nro %d\n",index);
#endif
  if (tablesCount > 1) {
    tables[index]           = tables[tablesCount--];
  } else {
    tablesCount           = 0;
  }
  MUTEX_UNLOCK(&tablesLock);
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::freeLocalTable() - freeing table  \n");
#endif
  FREE(table);                            /*free:LocalStoreTable*/

  return OK;
}
/**
 * Method localStoreDropTable
 * 
 * @param const DHT_TableId * table_id
 * 
 * @return int containing the successing information
 **/
int localStoreDropTable(const DHT_TableId * tableId) 
{
  int res = OK;
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG,
    "DEBUG: datalayer::localStoreDropTable() - operation called \n");
#endif
  
  /* free memory */
  res = freeLocalTable(getTableForId(tableId));
  if (res == OK) {
    /* delete file*/
    res = persistentDeleteTable(tableId);
  }
  return res;
}

/**
 * Method localStoreGetExpired
 * 
 * @param const DHT_TableId * table_id, 
 * @param TIME_T now
 * 
 * @return DHT_LocalStoreResultSet containing the results
 **/
DHT_LocalStoreResultSet localStoreGetExpired(
      const DHT_TableId * tableId, 
      TIME_T            now) 
{
  DHT_LocalStoreResultSet resultSet;
  LocalStoreTable         * table;
  TableNode               * tnode;
  SearchResultNode        * snode;
  LocalStoreResultIter    * resultIter;
  int ind;

  resultIter              = NULL;
  resultSet.iterator      = empty_iter;
  ind                     = getTableForId(tableId);
  if (ind == -1) {        /*getTableForId didn't find a matching table */
    resultSet.returnCode  = DHT_ERRORCODES__TABLE_NOT_FOUND;
    return resultSet;
  }
  table                   = tables[ind];
  /* initialise iterator and put the search node point to dummy node in iterator */
  resultIter              = new_Iter(table);
  snode                   = resultIter->currentResult;

  MUTEX_LOCK(&table->lock);
  /* TODO finer datasructures (hash table for faster lookups) */
  tnode                   = table->first;  
  while(tnode != NULL) {
    if (tnode->storedUnit->timeout < now) {
      snode->next         = new_SearchResultNode(tnode);
      snode               = snode->next;
    }
    tnode                 = tnode->next;
  }
  MUTEX_UNLOCK(&table->lock);
  if (snode == resultIter->currentResult) { /* no hits */
    nextUnit(&resultIter);      /* nullifies and frees the iterator */
    resultSet.iterator      = empty_iter;            /* avoid segfaults */
    resultSet.returnCode    = DHT_ERRORCODES__EMPTY_RESULT;
  } else {
    resultSet.iterator      = resultIter;
    resultSet.returnCode    = OK;
  }
  return resultSet;
}


/**
 * Method localStoreGetCloserTo
 * 
 * @param const DHT_TableId * table_id, 
 * @param HashCode160 myID 
 * @param HashCode160 newID
 * 
 * @return DHT_LocalStoreResultSet containing the results
 **/
DHT_LocalStoreResultSet localStoreGetCloserTo(
          const DHT_TableId * tableId, 
          HashCode160       * myId, 
          HashCode160       * newId) 
{
  DHT_LocalStoreResultSet resultSet;
  LocalStoreTable         * table;
  TableNode               * tnode;
  SearchResultNode        * snode;
  LocalStoreResultIter    * resultIter;
  HashCode160             * myDistance;
  HashCode160             * otherDistance;
  int ind;

  resultSet.iterator      = NULL;
  ind                     = getTableForId(tableId);
  if (ind == -1) {        /*getTableForId didn't find a matching table */
    resultSet.returnCode  = DHT_ERRORCODES__TABLE_NOT_FOUND;
    return resultSet;
  }
  table                   = tables[ind];
  /* initialise iterator and put the search node point to dummy node in iterator */
  resultIter              = new_Iter(table);
  snode                   = resultIter->currentResult;

  /* malloc the new has codes*/
  myDistance              = MALLOC(sizeof(HashCode160));  /*malloc:HashCode160*/
  otherDistance           = MALLOC(sizeof(HashCode160));  /*malloc:HashCode160*/
  
  MUTEX_LOCK(&table->lock);
  tnode                   = table->first;  
  while(tnode != NULL) {
    /* compute my distance */
    xorMetricDistance(&tnode->storedUnit->hash, myId, myDistance);
    /* compute other nodes distance */
    xorMetricDistance(&tnode->storedUnit->hash, newId, otherDistance);
    if (dhtHashCodeCompare(myDistance, otherDistance) < 0) {
      snode->next         = new_SearchResultNode(tnode);
      snode               = snode->next;
    }
    tnode                 = tnode->next;
  }
  MUTEX_UNLOCK(&table->lock);
  
  FREE(myDistance);                                     /*free:HashCode160*/
  FREE(otherDistance);                                  /*free:HashCode160*/
  if (snode == resultIter->currentResult) { /* no hits */
    nextUnit(&resultIter);      /* nullifies and frees the iterator */
    resultSet.iterator      = empty_iter;            /* avoid segfaults */
    resultSet.returnCode    = DHT_ERRORCODES__EMPTY_RESULT;
  } else {
    resultSet.iterator      = resultIter;
    resultSet.returnCode    = OK;
  }
  return resultSet;
}


/**
 * Method localStoreGetOwnData. 
 * 
 * @param const DHT_TableId * tableId, 
 * 
 * @return DHT_LocalStoreResultSet containing the results
 **/
DHT_LocalStoreResultSet localStoreGetOwnData(const DHT_TableId * tableId) 
{
  DHT_LocalStoreResultSet resultSet;
  LocalStoreTable         * table;
  TableNode               * tnode;
  SearchResultNode        * snode;
  LocalStoreResultIter    * resultIter;
  int ind;

  resultSet.iterator      = NULL;
  ind                     = getTableForId(tableId);
  if (ind == -1) {        /*getTableForId didn't find a matching table */
    resultSet.returnCode  = DHT_ERRORCODES__TABLE_NOT_FOUND;
    return resultSet;
  }
  table                   = tables[ind];
  /* initialise iterator and put the search node point to dummy node in iterator */
  resultIter              = new_Iter(table);
  snode                   = resultIter->currentResult;

  MUTEX_LOCK(&table->lock);
  /* TODO finer datasructures (use own units list instead) */
  tnode                   = table->first;  
  while(tnode != NULL) {
    if (isMyData(tnode->storedUnit)) {
      snode->next         = new_SearchResultNode(tnode);
      snode               = snode->next;
    }
    tnode                 = tnode->next;
  }
  MUTEX_UNLOCK(&table->lock);

  if (snode == resultIter->currentResult) { /* no hits */
    nextUnit(&resultIter);      /* nullifies and frees the iterator */
    resultSet.iterator      = empty_iter;            /* avoid segfaults */
    resultSet.returnCode    = DHT_ERRORCODES__EMPTY_RESULT;
  } else {
    resultSet.iterator      = resultIter;
    resultSet.returnCode    = OK;
  }
  return resultSet;
}

int initLocalStore(CoreAPIForApplication * never_used) 
{
  DHT_TableDef        ** tablesFromFile;
  DHT_LocalStoreUnit   ** dataFromFile;
  const DHT_TableId    * tableId;
  const DHT_TableMetaData * metadata;
  const DHT_TableConfig * config;
  int                  result,
                       ti,    /* table index */
                       di;    /* data index */

#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG, "DEBUG: datalayer::initLocalStore()\n");
#endif

  /* init static variables */  
  tablesArrayLength   = INITIAL_TABLES_ARRAY_SIZE;
  tables              = MALLOC(sizeof(LocalStoreTable*) * tablesArrayLength);
  tablesCount         = 0;

  /* load from persistance */
  ti=0;
  tablesFromFile      = persistentListTables();
  while (tablesFromFile[ti] != NULL) {
    tableId           = &tablesFromFile[ti]->tableId;
    metadata          = &tablesFromFile[ti]->metadata;
    config            = &tablesFromFile[ti]->config;
    if((result = localStoreCreateTable(tableId,metadata,config)) != OK) 
    {
      /* what to do with the allready loaded tables ??? */
      return result;
    } else {
      /* load all data from the table also */
      di = 0;
      dataFromFile = persistentListData(tableId);
      while (dataFromFile[di] != NULL) {
        if((result = localStorePut(tableId, dataFromFile[di])) != OK) 
        {
          /* what to do with the allready loaded tables and values */
          return result;
        } else {
          /* the next data in the file */
          di++;
        }
      }
     /* the next table file */
      tablesFromFile++;
    }
  }
  
  return OK;
}



/**
 * frees everything we have allocated
 **/ 
int doneLocalStore(CoreAPIForApplication * capi) {
  int res=OK,
      ret=OK;
  int i;
  
#if DEBUG_DHT_DATALAYER
  LOG(LOG_DEBUG, "DEBUG: datalayer::doneLocalStore()\n");
#endif
  MUTEX_DESTROY(&tablesLock);
  for (i = 0; i < tablesCount; i++) {
    if((res = freeLocalTable(i)) != OK) {
      ret = res;
    }
  }
  tablesCount = 0;
  FREE(tables);
  return ret;
}

