/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */
 
/**
 * This file describes the interface for the localstorage of a DHT module.
 * 
 * 
 * @file module/datalayer.h
 * @brief Interface and datastructures for DHT's localstorage
 * @author Simo Viitanen
 * @version $Id: datalayer.h,v 1.22 2004/05/14 15:51:57 szviitan Exp $ 
 */

#ifndef DHT_DATALAYER_H
#define DHT_DATALAYER_H 1

#ifndef DEBUG_DHT_DATALAYER
#define DEBUG_DHT_DATALAYER 1
#endif

#include "gnunet_core.h"    /* for HashCode160               */
#include "api_structs.h"    /* for dht's structs             */
#include "api_errorcodes.h"

/**
 * error codes: moved to ../include/api_errorcodes.h
#define DHT_ERRORCODES__DATALAYER_STORAGE_FULL      2
#define DHT_ERRORCODES__DATALAYER_KEYSTORAGE_FULL   3
#define DHT_ERRORCODES__TABLE_NOT_FOUND             4
#define DHT_ERRORCODES__EMPTY_RESULT                5
#define DHT_ERRORCODES__DUPLICATE_DATA_UNIT         6
#define DHT_ERRORCODES__DUPLICATE_TABLE_ID          7
*/
/************************* datastructures ************************************/

/**
 * @brief DHT_LocalStoreUnit is the basic unit that is stored in the datalayer
 * 
 */ 
typedef struct DHT_LocalStoreUnit_t {
  /**
   * hash is the hashed value of the stored key
   */
  HashCode160 hash;
  /**
   * Created is the time this record was inserted  
   */
  cron_t created;
  
  /* the time of the next timeout for the record */
  cron_t timeout;
  
  /* flags containing information about the record */
  unsigned char flags;
  
  /* data contains the key and the value catenated */
  DHT_DataContainer * data;
  
} DHT_LocalStoreUnit;


/**
 * @brief abstract type of a search result
 **/
typedef struct LocalStoreResultIter_t LocalStoreResultIter;

/**
 * @brief struct for resultset from local storage
 **/
typedef struct DHT_LocalStoreResultSet_t {
  
  /**
   * the state of last operation
   **/
  unsigned int returnCode;

  /**
   * LocalStoreResultPointer to the actual result
   **/  
  LocalStoreResultIter * iterator;
  
} DHT_LocalStoreResultSet;


/**
 * @brief tabledef's are returned from tableset
 **/ 
typedef struct DHT_TableDef_t {
  
  /**
   * the identification of the table
   **/
  const DHT_TableId tableId;
  
  /**
   * the meta data of the table
   **/
  const DHT_TableMetaData metadata;
  
  /**
   * the config information of the table
   **/
  const DHT_TableConfig config;
  
} DHT_TableDef;

/**
 * @brief abstract type of a search result
 **/
typedef struct LocalStoreTableIter_t LocalStoreTableIter;

/**
 * abstract type for the tableset
 **/ 
typedef struct tableset_t {
  
  /**
   **/
  unsigned int returnCode;
  
  /**
   **/ 
  LocalStoreTableIter * iterator;
  
} DHT_LocalTableSet;

/***************** helper functions for datastructures **********************/
/**
 * checks if the mydata flag is on
 * 
 * @param DHT_LocalStoreUnit * dataunit to check
 * @return boolean value accordingly
 **/
int isMyData(DHT_LocalStoreUnit * data);
/**
 * Sets the mydata flag on and off according to the second param.
 * 
 * @param DHT_LocalStoreUnit * dataunit to set
 * @param int as boolean wether to set bit on or off
 **/
void setMyData(DHT_LocalStoreUnit * data, int trueOrFalse);
/**
 * Checks if the republishable flag is on
 * 
 * @param DHT_LocalStoreUnit * dataunit to check
 * @return boolean value accordingly
 **/
int isRepublishable(DHT_LocalStoreUnit * data);
/**
 * Sets the republishable flag on and off according to the second param.
 * 
 * @param DHT_LocalStoreUnit * dataunit to set
 * @param int as boolean wether to set bit on or off
 **/
void setRepublishable(DHT_LocalStoreUnit * data, int trueOrFalse);

/**
 * Iterating resultset. The iterating takes care of the inner datastructures by 
 * freeing them as it goes along. Ending returns NULL so results can be tested 
 * against NULL. Reaching end also frees memory allocation of the resultiterator. 
 * Example of usage
 * 
<pre>
DHT_LocalStoreResultSet search_result = search(table, ...);
if(isOkay(search_result.returnCode)) {
  DHT_LocalStoreUnit * unit = NULL;
  LocalStoreResultIter * iter = search_result->iterator
  while(unit = nextUnit(&iter) != NULL ) {
    doMyThing(unit);
  }
  // by this point all internal allocation for search_result->iterator is freed
}
</pre>
 * 
 * @param LocalStoreResultIterPointer returned by any of the search methods
 * @return the next DHT_LocalStoreUnit in the set. NULL if there is no more 
 * results in the set.
 **/
DHT_LocalStoreUnit * nextUnit(LocalStoreResultIter ** result);

/**
 * Iterating tableset. 
 * 
 * @param TableSet returned by ListTables
 * @return the next DHT_TableDef in the set. NULL if there is no more results 
 * in the set.
 **/
DHT_TableDef * nextTable(LocalStoreTableIter ** iterator);

/**
 * Constructor for DHT_LocalStoreUnit. Creates a copy of parameters so the 
 * memory management of DHT_LocalStoreUnit's becomes entirely datalayers own consern. 
 * 
 * 
 * @param HashCode160 * hash - the hash code of the key
 * @param DHT_DataContainer * data - the key and value combined
 **/
DHT_LocalStoreUnit * new_DHT_LocalStoreUnit(HashCode160 * hash, 
                                            DHT_DataContainer * data);



/****************** the actual interface functions *************************/
/**
 * Inits the localStore.
 * 
 * @return OK if initing was fine
 **/ 
int initLocalStore(CoreAPIForApplication * capi);
/**
 * Sets the localStore done. Frees memory and such.
 * 
 * @return OK if all is fine
 **/ 
int doneLocalStore(CoreAPIForApplication * capi);

/**
 * Lists all tables this storage unit contains.
 * 
 * @return set of tables created to this unit.
 */
DHT_LocalTableSet localStoreListTables();

/**
 * Function localStoreCreateTable creates a new table to the datalayer.
 * Returns the successing information.
 * 
 * @param const DHT_TableId * table_id
 * @param const DHT_TableMetaDatan metadata
 * @param const DHT_TableConfig config
 *
 * @return int containing the successing information
 */
int localStoreCreateTable(const DHT_TableId * tableId, 
                          const DHT_TableMetaData * metadata,
                          const DHT_TableConfig * config);

/***
 * Method localStorePut puts a constructed DHT_LocalStoreUnit to the specified 
 * table.
 * The unitToStore reference should not be used after this, because in case of 
 * error (DHT_ERRORCODES__DUPLICATE_DATA_UNIT tex.) the method frees that unit.
 * So implementation can change the referred unit.
 *
 * @param const DHT_TableId * tableId, 
 * @param DHT_LocalStoreUnit * unitToStore
 * 
 * @return int result of storing
 */
int localStorePut(const DHT_TableId * tableId, 
                  DHT_LocalStoreUnit * unitToStore);

/**
 * Method localStoreDelete deletes a unit that has been locally stored from 
 * a table.
<pre>
DHT_LocalStoreResultSet search_result = search(table, ...);
if(isOkay(search_result.returnCode)) {
  DHT_LocalStoreUnit * unit = NULL;
  LocalStoreResultIter * iter = search_result->iterator
  while(unit = nextUnit(&iter) != NULL ) {
    if (isToBeRemoved(unit)) {
      if (localStoreDelete(table, iter) == OK) {
        // unit is freed
      }
    }
  }
  // by this point all internal allocation for search_result->iterator is freed
}
</pre>
 * 
 * @param const DHT_TableId * tableId
 * @param LocalStoreResultIter * pointer
 *
 * @return int containing the successing information
 *
 */
int localStoreDelete(const DHT_TableId * tableId, 
                     LocalStoreResultIter * resultPointingToRemovable);

/**
 * Method localStoreGet gets all the units stored for one specific hash key.
 * 
 * @param const DHT_TableId * tableId
 * @param HashCode160 keyToGet
 *
 * @return DHT_LocalStoreResultSet containing the results
 */
DHT_LocalStoreResultSet localStoreGet(const DHT_TableId * tableId, 
                                      HashCode160 * keyToGet);

/**
 * Method localStoreGetOwnData returns all the units that are marked with myData
 * flag.
 * 
 * 
 * @see isMyData(DHT_LocalStoreUnit)
 * @see setMyData(DHT_LocalStoreUnit,int)
 * 
 * @param const DHT_TableId * tableId, 
 * @return DHT_LocalStoreResultSet containing the results
 **/
DHT_LocalStoreResultSet localStoreGetOwnData(const DHT_TableId * tableId);


/**
 * Method localStoreDropTable deletes the table from localStore so that it can't be used after this.
 * 
 * @param const DHT_TableId * tableId
 * 
 * @return int containing the successing information
 **/
int localStoreDropTable(const DHT_TableId * tableId);

/**
 * Method localStoreGetExpired
 * 
 * @param const DHT_TableId * tableId, 
 * @param TIME_T now
 * 
 * @return DHT_LocalStoreResultSet containing the results
 **/
DHT_LocalStoreResultSet localStoreGetExpired(const DHT_TableId * tableId, 
                                             TIME_T now);

/**
 * Method localStoreGetCloserTo
 * 
 * @param const DHT_TableId * tableId, 
 * @param HashCode160 myID 
 * @param HashCode160 newID
 * 
 * @return DHT_LocalStoreResultSet containing the results
 **/
DHT_LocalStoreResultSet localStoreGetCloserTo(const DHT_TableId * tableId, 
                                              HashCode160 * myID, 
                                              HashCode160 * newID);



#endif /* DHT_DATALAYER_H */
