 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */
 
/**
 * @file module/dht.c
 * @brief Core file of DHT-module for GNUnet p2p framework.
 * @author Marko Rih
 * @version $Id: dht.c,v 1.67 2004/05/23 16:48:30 mjraiha Exp $
 */

#include <gnunet_core.h>    /* for HashCode160               */
#include "dht.h"
#include "rpc.h"
#include "client-server.h"
#include "datalayer.h"
#include "kademlia.h"            
#include "dht_util.h"

#define DEBUG_DHT 1

/**
 * Global core API.
 **/
CoreAPIForApplication * coreAPI = NULL;

/*
 * Initialize the AFS module. This method name must match
 * the library name (libgnunet_XXX => initialize_XXX).
 * @param 
 * @return SYSERR on errors
 */
int initialize_dht_protocol(CoreAPIForApplication * capi) {
  int ok = OK;

  coreAPI = capi;

  LOG(LOG_DEBUG, "DEBUG: DHT registering handlers: "
    "%d %d %d %d %d %d %d %d %d %d and %d %d %d\n",
      DHT_CS_PROTO_REQUEST_API_ID,
      DHT_CS_PROTO_REQUEST_CREATE,
      DHT_CS_PROTO_REQUEST_JOIN,
      DHT_CS_PROTO_REQUEST_LEAVE,
      DHT_CS_PROTO_REQUEST_INSERT,
      DHT_CS_PROTO_REQUEST_FETCH,
      DHT_CS_PROTO_REQUEST_TABLES,
      DHT_CS_PROTO_REQUEST_INSERTED,
      DHT_CS_PROTO_REQUEST_DROP,
      DHT_CS_PROTO_REQUEST_STATUS,
      RPC_p2p_PROTO_REQ,
      RPC_p2p_PROTO_RES,
      RPC_p2p_PROTO_ACK);

  if (SYSERR == initClientServer(capi))
    ok = SYSERR;

  if (SYSERR == RPC_init(capi))
    ok = SYSERR;

  if (SYSERR == initLocalStore(capi))
    ok = SYSERR;

  if (SYSERR == Kademlia_init(capi))
    ok = SYSERR;

  return ok;
}

/*
 * Unregisters handlers, cleans memory structures etc when node exits.
 */
void done_dht_protocol() {

  doneClientServer(coreAPI);
  RPC_done(coreAPI);
  doneLocalStore(coreAPI);
  coreAPI = NULL;
}

/**
 * Create a new DHT-table.
 * @param tableConfig includes variables for finetuning table internals
 * @param metaData includes name for the new table
 * @param tableId of new created table for return value
 * @return success or failure of operation
 */
int create(DHT_TableConfig * tableConfig,
           DHT_TableMetaData * metaData,
           DHT_TableId ** tableId) {

  int res;
  /* invent a table id and hope its unique */
  *tableId = MALLOC(sizeof(DHT_TableId));
  makeRandomId(*tableId);

#if DEBUG_DHT
  HexName hex;
  hash2hex(*tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: dht::create(tableId: %s)\n",hex.data);
#endif
  /* create table for datalayer */
  res   = localStoreCreateTable(*tableId, metaData, tableConfig);
  if (res != OK) {
    /* error in creating table */
    /* free things here ??? */
    return res;
  } else {
    /* if there was room for a new table create it to kademlia */
    Kademlia_registerTable(
        *tableId,
        (DHT_NodeAddress *)(coreAPI->myIdentity),
        KADEMLIA_BUCKET_SIZE);
    return OK;
  }  
}

/**
 * Join an existing DHT-table.
 * @param nodeAddress of other node which we use to connect to DHT-table
 * @param tableId of table which we want to join
 * @return success or failure of operation
 */
int join(DHT_NodeAddress * nodeAddress,
         DHT_TableId * tableId) {
#if DEBUG_DHT
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: dht::join(tableId: %s)\n",hex.data);
#endif
  LOG(LOG_DEBUG, "Entering Kademlia_joinTable\n");
  Kademlia_joinTable(nodeAddress, tableId);
  
  /* LOCALSTORE: create a table accordingly */
  return OK;
}

/**
 * Leave from existing DHT-table.
 * @param tableId of table which we want to leave
 * @return success or failure of operation
 */
int leave(DHT_TableId * tableId) {
  int res = OK;
#if DEBUG_DHT
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: dht::leave(tableId: %s)\n",hex.data);
#endif
  /* KADEMLIA: drop the routing table */
  
  /* if it was okay drop also from localstore */
  if(res == OK) {
    res = localStoreDropTable(tableId);
  }
  return res;
}

/**
 * Insert <key,value>-pair into DHT-table.
 * @param tableId of table which we want insert pair
 * @param key of pair
 * @param value of pair
 * @return success or failure of operation
 */
int insert(DHT_TableId * tableId,
           DHT_DataContainer * key,
           DHT_DataContainer * value) {
  int result;
  HashCode160       * asHash;
  DHT_DataContainer * asData;

#if DEBUG_DHT
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: dht::insert(tableId: %s)\n",hex.data);
#endif
  /* insert in local store as my data */
  convertKeyValueToHashNData(key,value,&asHash,&asData);

  result = dhtStoreValue(
    tableId,
    asHash,
    asData);

  FREE(asHash);
  FREE(asData);
  return result;
}

/**
 * Fetch resultset of <key,value>-pairs from DHT-table.
 * @param tableId of table we want to fetch from
 * @param key search this
 * @param result set for return value
 * @return success or failure of operation
 */
int fetch(DHT_TableId * tableId,
          DHT_DataContainer * key,
          DHT_ResultSet ** result) {
 /* KADEMLIA: this propably should be in kademlia, beacause 
  * kademlia uses the same kind of functionality when dealing with the 
  * handleFetch() or ...
  */
  DHT_LocalStoreResultSet localLookup;

  DHT_ResultSet           * resultSet;
  DHT_ResultSetItem       * resultItem;
  HashCode160             * keyH;
  
  DHT_LocalStoreUnit      * localData;
  DHT_DataContainer       * itemKey;
  DHT_DataContainer       * itemValue;

#if DEBUG_DHT
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: dht::fetch(tableId: %s) keyLength %d\n",
      hex.data,
      key->dataLength);
#endif

  resultItem = NULL;
  keyH = MALLOC(sizeof(HashCode160));        /*malloc:HashCode160*/
  hash(key->data, key->dataLength, keyH);
  resultSet = MALLOC(sizeof(DHT_ResultSet));  /*malloc:DHT_ResultSet*/
  resultSet->firstItem = NULL;

  /* do a local lookup */
  localLookup = localStoreGet(tableId, keyH);
  if (localLookup.returnCode == OK) {
    resultSet->errorCode = OK;
    /* populate resultSet */
    int count = 0;
    while ((localData = nextUnit(&localLookup.iterator)) != NULL) {
      /* extract key and value from localData->data
       * if the key is match then set the value to the resultItem->data*/
      itemKey=itemValue=NULL;
      if (OK != convertDataToKeyNValue(localData->data, &itemKey, &itemValue)) 
      {
        LOG(LOG_FAILURE, 
          "FAILURE: dht::fetch() - malformed data from datalayer !!\n");
        freeResultSet(result);
        return SYSERR;
      }
      resultItem = MALLOC(sizeof(DHT_ResultSetItem)); /*malloc:DHT_ResultSetItem*/
      resultItem->data = itemValue;
      resultItem->nextItem = resultSet->firstItem;
      resultSet->firstItem = resultItem;
      count++;

      FREE(itemKey->data); /* we don't need it any more*/ /*free:void*/
      FREE(itemKey); /*free:DHT_DataContainer*/
    }
    *result = resultSet;
  } else {
    /* KADEMLIA: do the fetch over rpc ??? 
     * fetch(table,key);
     * or should all of this function be in kademlia, because it also needs
     * convertDataToKeyNValue and equlsDataContainers ???
     */
    resultSet->errorCode = localLookup.returnCode;
  }
  FREE(keyH);                       /*free:HashCode160*/
  return OK;
}


/**
 * Fetch a list of tables from other node.
 * @param nodeAddress of other node
 * @param tableset set for return value
 * @return success or failure of operation
 */
int tables(DHT_NodeAddress * nodeAddress,
           DHT_TableSet ** tableSet) {

  LOG(LOG_DEBUG,
      "Starting table list from node %s.\n",
      hash160ToString(getNodeAddressHash(nodeAddress)));
  DHT_NodeAddress *myIdentity; 

  myIdentity = (DHT_NodeAddress *)(coreAPI->myIdentity); 

  (*tableSet) = MALLOC(sizeof(DHT_TableSet)); 
  (*tableSet)->errorCode = OK; 
  (*tableSet)->firstItem = NULL; 
  /*
  if (contactsEqual(nodeAddress, myIdentity)) {
    return dhtListLocalTables(*tableSet);
  }
  else {
  */
    LOG(LOG_DEBUG,
        "Listing tables using RPC.\n");
    return dhtListRemoteTables(nodeAddress, *tableSet);
  /*
  }
  */
}

/**
 * List <key,value>-pairs inserted by this node.
 * @param tableId of table we want to list inserted pairs of
 * @param result set for return value
 * @return success or failure of operation
 */
int inserted(DHT_TableId * tableId,
             DHT_DataList ** result) {

  DHT_LocalStoreResultSet localResult;
  DHT_LocalStoreUnit      * unit;
  DHT_DataListItem        * item,
                          * last;
  DHT_DataContainer       * itemKey;
  DHT_DataContainer       * itemValue;
 /* DHT_StoredDataReference * itemUniq;*/

#if DEBUG_DHT
  HexName hex;
  hash2hex(tableId, &hex);
  LOG(LOG_DEBUG, "DEBUG: dht::inserted(tableId: %s)\n",hex.data);
#endif
  
  int count = 0;

  item=last=NULL;
  unit=NULL;
  localResult = localStoreGetOwnData(tableId);
  if (localResult.returnCode == OK) {
    *result = MALLOC(sizeof(DHT_DataList)); /*malloc:DHT_DataList*/
    while ((unit = nextUnit(&localResult.iterator)) != NULL) {
      item       = MALLOC(sizeof(DHT_DataListItem)); /*malloc:DHT_DataListItem*/
      itemKey=itemValue=NULL;
      if (OK != convertDataToKeyNValue(unit->data, &itemKey, &itemValue)) 
      {
        LOG(LOG_FAILURE, 
          "FAILURE: dht::inserted() - malformed data from datalayer !!\n");
        freeDataList(result);
        return SYSERR;
      }
      item->key   = itemKey;
      item->value = itemValue;
      memcpy(
        &item->uniqueReference.hashCode,
        &unit->hash,
        sizeof(HashCode160)
        );
      item->uniqueReference.insertionTime = unit->created;
      item->nextItem=NULL;
      if(count++ == 0) {
        (*result)->firstItem = item;
      } else {
        last->nextItem = item;
      }
      last = item;
    }
  }
  return OK;
}

/**
 * Remove <key,value>-pairs inserted by this node.
 * @param tableId of table we want to drop inserted pair of
 * @param dataReference to pair we want to drop
 * @return success or failure of operation
 */
int drop(DHT_TableId * tableId,
         DHT_StoredDataReference * dataReference) {
  DHT_LocalStoreResultSet localResult;
  DHT_LocalStoreUnit      * unit;
  
#if DEBUG_DHT
  HexName hex;
  HexName hex2;
  hash2hex(tableId, &hex);
  hash2hex(&dataReference->hashCode, &hex2);
  LOG(LOG_DEBUG, "DEBUG: dht::drop(tableId: %s, %s)\n",hex.data,hex2.data);
#endif
  localResult = localStoreGet(tableId, &dataReference->hashCode);
  if (localResult.returnCode == OK) {
    int res = OK;
    while ((unit = nextUnit(&localResult.iterator)) != NULL) {
      if(isMyData(unit) && unit->created == dataReference->insertionTime) {
        res = localStoreDelete(tableId, localResult.iterator);
        /* continue iteration till the end */
      }
    }
    return res;
  } else {
    return localResult.returnCode;
  }
}

/* utility functions for */
int freeResultSet(DHT_ResultSet ** result) {
  DHT_ResultSetItem * item, * prevItem;
  item=prevItem=NULL;
  if (*result != NULL) {
    item = (*result)->firstItem;
    while(item != NULL) {
      prevItem = item;
      item = prevItem->nextItem;
      FREE(prevItem->data);
      FREE(prevItem);
    }
    FREE(*result);
    *result = NULL;
  }
  return OK;
}

int freeDataList(DHT_DataList **result) {
  DHT_DataListItem  * item, 
                    * prevItem;
  item=prevItem=NULL;
  if (*result != NULL) {
    item = (*result)->firstItem;
    while(item != NULL) {
      prevItem = item;
      item = prevItem->nextItem;
      freeDataContainer(prevItem->key);
      freeDataContainer(prevItem->value);
      FREE(prevItem);
    }
    FREE(*result);
    *result = NULL;
  }
  return OK;
}

int freeTableSet(DHT_TableSet ** result) {
  DHT_TableSetItem  * item, 
                    * prevItem;
  item=prevItem=NULL;
  if (*result != NULL) {
    item = (*result)->firstItem;
    while(item != NULL) {
      prevItem = item;
      item = prevItem->nextItem;
      FREE(prevItem);
    }
    FREE(*result);
    *result = NULL;
  }
  return OK;
}
