/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file module/dht.h
 * @brief Core file of DHT-module for GNUnet p2p framework.
 * @author Marko Rih
 * @version $Id: dht.h,v 1.20 2004/05/10 22:28:53 mjraiha Exp $
 **/

#ifndef DHT_H
#define DHT_H

#include "gnunet_core.h"
#include "api_structs.h"

extern CoreAPIForApplication * coreAPI;

/*
 * Initialize the AFS module. This method name must match
 * the library name (libgnunet_XXX => initialize_XXX).
 * @param capi 
 * @return SYSERR on errors
 */
int initialize_dht_protocol(CoreAPIForApplication * capi);

/*
 * Unregisters handlers, cleans memory structures etc when node exits.
 */
void done_dht_protocol();

/**
 * Create a new DHT-table.
 * @param tableConfig includes variables for finetuning table internals
 * @param metaData includes name for the new table
 * @param tableId of new created table for return value
 * @return success or failure of operation
 */
int create(DHT_TableConfig * tableConfig,
           DHT_TableMetaData * metaData,
           DHT_TableId ** tableId);

/**
 * Join an existing DHT-table.
 * @param nodeAddress of other node which we use to connect to DHT-table
 * @param tableId of table which we want to join
 * @return success or failure of operation
 */
int join(DHT_NodeAddress * nodeAddress,
         DHT_TableId * tableId);

/**
 * Leave from existing DHT-table.
 * @param tableId of table which we want to leave
 * @return success or failure of operation
 */
int leave(DHT_TableId * tableId);

/**
 * Insert <key,value>-pair into DHT-table.
 * @param tableId of table which we want insert pair
 * @param key of pair
 * @param value of pair
 * @return success or failure of operation
 */
int insert(DHT_TableId * tableId,
           DHT_DataContainer * key,
           DHT_DataContainer * value);

/**
 * Fetch resultset of <key,value>-pairs from DHT-table.
 * @param tableId of table we want to fetch from
 * @param key search this
 * @param result set for return value
 * @return success or failure of operation
 */
int fetch(DHT_TableId * tableId,
          DHT_DataContainer * key,
          DHT_ResultSet ** result);

/**
 * Fetch a list of tables from other node.
 * @param nodeAddress of other node
 * @param tableset set for return value
 * @return success or failure of operation
 */
int tables(DHT_NodeAddress * nodeAddress,
           DHT_TableSet ** tableSet);

/**
 * List <key,value>-pairs inserted by this node.
 * @param tableId of table we want to list inserted pairs of
 * @param result set for return value
 * @return success or failure of operation
 */
int inserted(DHT_TableId * tableId,
             DHT_DataList ** result);

/**
 * Remove <key,value>-pairs inserted by this node.
 * @param tableId of table we want to drop inserted pair of
 * @param dataReference to pair we want to drop
 * @return success or failure of operation
 */
int drop(DHT_TableId * tableId,
         DHT_StoredDataReference * dataReference);


/**
 * Frees the allocated resultset and all the items in it. 
 * Also frees all data in items.
 * Sets the result pointer to NULL.
 * @param result
 * @return successing  
 */
int freeResultSet(DHT_ResultSet ** result);

/**
 * Frees the allocated datalist and all the items in it. 
 * Also frees all keys and values in items.
 * Sets the result pointer to NULL.
 * @param result
 * @return successing  
 */
int freeDataList(DHT_DataList **result);

/**
 * Frees the allocated tablelist and all the items in it. 
 * Also frees all data in items.
 * Sets the result pointer to NULL.
 * @param result
 * @return successing  
 */
int freeTableSet(DHT_TableSet ** result);

#endif

