/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */

/**
 * @file module/dht_api.c
 * @brief DHT-module's core API's implementation. 
 * @author Tomi Tukiainen
 * @version $Id: dht_api.c,v 1.32 2004/05/24 11:06:36 mjraiha Exp $ 
 */

#include <string.h>
#include <dht_api.h>
#include "api_constants.h"
#include "client-server-api.h"
#include "../util/networking.h"

#define timeOut(startTime,millis) (cronTime(NULL)>startTime+millis)
#define CONTINUE -5115714

/* ************************* GLOBALS ********************************** */
/* these globals are used by API as it's internal state                 */

/** used to lock/unlock API operations */
static Semaphore * apiSemaphore; 

/** used to lock/unlock API's threads internally */
static Semaphore * intSemaphore; 

/** socket for receiving TCP messages */
static GNUNET_TCP_SOCKET * sock; 

/** used as a unique identifier for current request, starts from 0 */
static unsigned short currentRequestId;

/** unique identifier of the API, received from gnunetd */
static unsigned short apiId; 

/**
 * Used to exchange current-operation-status-info between threads. 
 * In general, API functions's implementations change and receiver 
 * thread read this. 
 *
 * If currentOperationStatus is some WAITFOR-state, receiver thread
 * will try to receive data.  
 *
 * Use intSemaphore for reading/writing. 
 */
static unsigned short currentOperationStatus;

/**
 * Used to exchange data-received-info between threads. 
 * In general, API function sets this to 0 and changes
 * currentOperationStatus to *WAITFOR*. Then receiver
 * thread knows that API waits for some data. 
 *
 * Receiver thread discards data where 'apiId' doesn't
 * match this api's id OR 'requestId' doesn't match 
 * 'currentRequestId'. Otherwise data is received and
 * it's up to the method to check what type of message
 * received data is. 
 *
 * When any qualified data is received, receiver thread
 * changes dataReceived to 1, indicating that some new data
 * is available.
 *
 * Use intSemaphore for reading/writing. 
 */
static unsigned short dataReceived; 

/**
 * Pointer to a 64k memory area that is used to store
 * received message (no matter what is received). 
 *
 * API function knows what to cast when using this
 * pointer. 
 *
 * Use intSemaphore for reading/writing. 
 */
static CS_HEADER * receivedData;

/**
 * Global variable that always contains size (HBO) of last
 * received message (no matter what is received). 
 *
 * Last received message = last received message that
 * caused dataReceived-flag become true. 
 */
static unsigned short receivedDataSize; 

/**
 * Global variable that always contains tcpType (HBO) of last
 * received message (no matter what is received). 
 *
 * Last received message = last received message that
 * caused dataReceived-flag become true. 
 */
static unsigned short receivedDataTcpType; 

/**
 * Pointers to the receivedData for various message types
 * (use pointer when you know which one to use). 
 */
static DHT_CS_REPLY_STANDARD * receivedStandardReplyMsg;
static DHT_CS_REPLY_FAILURE * receivedFailureMsg; 
static DHT_CS_REPLY_RESULTS * receivedResultsMsg;
static DHT_CS_REPLY_STATUS * receivedStatusMsg; 

/**
 * Used to indicate that destroy-method has been called 
 *
 * Use intSemaphore for reading/writing. 
 */
static unsigned short destroyCalled; 

/**
 * Used to store handle that's needed for joining receiver 
 * thread when destroy()-method is called. 
 */
static pthread_t receiveThreadHandle; 

/* ************************* API RECEIVER THREAD ********************************** */
/* Implementation for RECEIVER THREAD and helpers for its direct use                */

/**
 * Handles standard reply message from gnunetd.
 */
void handleStandardReply(DHT_CS_REPLY_STANDARD * message) {
  unsigned short messageApiId;
  unsigned short messageRequestId; 
  
  messageApiId     = ntohs(message->dhtCSHeader.apiId);
  messageRequestId = ntohs(message->dhtCSHeader.requestId); 
  
  if(/* messageApiId == apiId && <- removed because messages don't have apiId set correctly */
     messageRequestId == currentRequestId) {
    
    dataReceived = 1; 
    receivedDataSize    = ntohs(receivedData->size); 
    receivedDataTcpType = ntohs(receivedData->tcpType); 
    
  }
  else {
    LOG(LOG_DEBUG, "dht_api::handleStandardReply() - Ignored msg apiId r %d a %d, reqId r %d a %d.\n",
        messageApiId, apiId, messageRequestId, currentRequestId);
  }
}

/**
 * Handles failure reply message from gnunetd.
 */
void handleFailureReply(DHT_CS_REPLY_FAILURE * message) {
  unsigned short messageApiId;
  unsigned short messageRequestId; 

  messageApiId     = ntohs(message->dhtCSHeader.apiId);
  messageRequestId = ntohs(message->dhtCSHeader.requestId); 
    
  if(/* messageApiId == apiId && <- removed because messages don't have apiId set correctly */
     messageRequestId == currentRequestId) {
    
    dataReceived = 1; 
    receivedDataSize = ntohs(receivedData->size); 
    receivedDataTcpType = ntohs(receivedData->tcpType); 

  }
  else {
    LOG(LOG_DEBUG, "dht_api::handleFailureReply() - Ignored msg apiId r %d a %d, reqId r %d a %d.\n",
        messageApiId, apiId, messageRequestId, currentRequestId);
  }
}

/**
 * Handles results reply message from gnunetd.
 */
void handleResultsReply(DHT_CS_REPLY_RESULTS * message) {
  unsigned short messageApiId;
  unsigned short messageRequestId; 

  messageApiId     = ntohs(message->dhtCSHeader.apiId);
  messageRequestId = ntohs(message->dhtCSHeader.requestId); 
    
  if(/* messageApiId == apiId && <- removed because messages don't have apiId set correctly */
     messageRequestId == currentRequestId) {
    
    dataReceived = 1; 
    receivedDataSize = ntohs(receivedData->size); 
    receivedDataTcpType = ntohs(receivedData->tcpType); 
    
  }
  else {
    LOG(LOG_DEBUG, "dht_api::handleResultsReply() - Ignored msg apiId r %d a %d, reqId r %d a %d.\n",
        messageApiId, apiId, messageRequestId, currentRequestId);
  }
}

/**
 * Handles status reply message from gnunetd.
 */
void handleStatusReply(DHT_CS_REPLY_STATUS * message) {
  unsigned short messageApiId;
  unsigned short messageRequestId; 

  messageApiId     = ntohs(message->dhtCSHeader.apiId);
  messageRequestId = ntohs(message->dhtCSHeader.requestId); 
    
  if(/* messageApiId == apiId && <- removed because messages don't have apiId set correctly */
     messageRequestId == currentRequestId) {

    dataReceived = 1; 
    receivedDataSize = ntohs(receivedData->size); 
    receivedDataTcpType = ntohs(receivedData->tcpType); 

  }
  else {
    LOG(LOG_DEBUG, "dht_api::handleStatusReply() - Ignored msg apiId r %d a %d, reqId r %d a %d.\n",
        messageApiId, apiId, messageRequestId, currentRequestId);
  }
}

/**
 * Receiver thread for the TCP-connection from gnunetd. 
 */
static void * receiveThread(GNUNET_TCP_SOCKET * sock) {
  unsigned short tcpType; 
  
  while(!destroyCalled) {
    LOG(LOG_DEBUG, "dht_api::receiveThread() -> SEMAPHORE_DOWN(intSemaphore)\n"); 
    SEMAPHORE_DOWN(intSemaphore); 
    LOG(LOG_DEBUG, "dht_api::receiveThread() - continue execution (check if should receive)\n"); 

    /* 
       try to receive new data if api function is waiting for some and 
       data has not yet been received. 
     */
    if(currentOperationStatus != OPERATION_STATUS_OTHER &&
       !dataReceived) {

      SEMAPHORE_UP(intSemaphore); 

      LOG(LOG_DEBUG, "dht_api::receiveThread() -> readFromSocket(). \n"); 

      if(!(OK == readFromSocket(sock, (CS_HEADER **)&receivedData))) {
	LOG(LOG_FAILURE, "dht_api::receiveThread() - Failed to read from socket -> break.\n"); 
	break; 
      }
      LOG(LOG_DEBUG, "dht_api::receiveThread() - received something\n"); 

      tcpType = ntohs(receivedData->tcpType);

      LOG(LOG_DEBUG, "dht_api::receiveThread() -> SEMAPHORE_DOWN(intSemaphore)\n"); 
      SEMAPHORE_DOWN(intSemaphore); 
      LOG(LOG_DEBUG, "dht_api::receiveThread() - continue execution (something received)\n"); 

      if(currentOperationStatus != OPERATION_STATUS_OTHER) {

	LOG(LOG_DEBUG, "dht_api::receiveThread() - receive message of type %d.\n",tcpType); 

	switch(tcpType) {
	case DHT_CS_PROTO_REPLY_STANDARD:
	  handleStandardReply(receivedStandardReplyMsg); 
	  break;
	case DHT_CS_PROTO_REPLY_FAILURE:
	  handleFailureReply(receivedFailureMsg);
	  break;
	case DHT_CS_PROTO_REPLY_RESULTS:
	  handleResultsReply(receivedResultsMsg);
	  break;
	case DHT_CS_PROTO_REPLY_STATUS:
	  handleStatusReply(receivedStatusMsg);
	  break; 
	default:
	  break; /* silently ignore */
	}
      }
      else {
	/* silently ignore everything, because no operation is in progress */
	LOG(LOG_DEBUG, "dht_api::receiveThread() - ignoring message of type %d.\n",tcpType); 
      }

      SEMAPHORE_UP(intSemaphore); 
    }
    else {
      SEMAPHORE_UP(intSemaphore);
      gnunet_util_sleep(cronMILLIS * RECEIVE_THREAD_SLEEP_MILLIS);
    }
  }

  LOG(LOG_DEBUG, "dht_api::receiveThread() - returning.\n"); 

  return NULL; 
}


/* ************************* API FUNCTIONS ********************************** */
/* Implementations for API functions and helpers for their direct use         */


/**
 * DOWNs apiSemaphore and increments currentRequestId. Should be called 
 * in the beginning of every api operation. 
 */
void apiOperationStart() {
  SEMAPHORE_DOWN(apiSemaphore); /* Blocks subsequent API-function calls */
  SEMAPHORE_DOWN(intSemaphore);
  currentRequestId ++; 
  SEMAPHORE_UP(intSemaphore);
}


/**
 * UPs apiSemaphore and returns errorCode. Note that intSemaphore must 
 * be UP'd and apiSemaphore DOWN'd when this function is called. 
 *
 * @param errorCode errorcode to return
 * @return errorCode
 */
int apiOperationEnd(int errorCode) {
  SEMAPHORE_UP(apiSemaphore);
  return errorCode;
}

/**
 * Causes API operation to sleep a while (wait for a message from TCP). 
 * UP's intSemaphore for sleeping period and DOWN's it immediately after
 * that. 
 */
void apiOperationWaitMessage() {
  SEMAPHORE_UP(intSemaphore);
  gnunet_util_sleep(cronMILLIS * API_SLEEP_MILLIS); 
  SEMAPHORE_DOWN(intSemaphore);
}

/**
 * Causes API to receive message or timeout. 
 * 
 * @param operationStatus operation status that indicates what is waited for
 * @param timeOut timeout for receiving
 */
void apiOperationReceive(int operationStatus,
			 int timeOut) {
  cron_t receiveStartTime; 

  cronTime(&receiveStartTime);

  SEMAPHORE_DOWN(intSemaphore); 
  dataReceived = 0;    /* global flag */
  currentOperationStatus = operationStatus;

  while(!dataReceived &&
	!timeOut(receiveStartTime,timeOut)) {
    apiOperationWaitMessage(); 
  }

  currentOperationStatus = OPERATION_STATUS_OTHER; 
  SEMAPHORE_UP(intSemaphore); 
}

/**
 * Handles standard reply message by returning CONTINUE or DHT_ERRORCODES__OP_REQUEST_REJECTED. 
 */
int handleStandardMsg(void *unUsed) {
  if(ntohs(receivedStandardReplyMsg->requestStatusNumber)==REQUEST_STATUS_REJECTED) {
    LOG(LOG_DEBUG, "dht_api::handleStandardMsg() - Request was rejected.\n"); 
    return DHT_ERRORCODES__OP_REQUEST_REJECTED; 
  }
  LOG(LOG_DEBUG, "dht_api::handleStandardMsg() - Request was not rejected.\n"); 
  return CONTINUE; 
}

/**
 * Handles failure message by returning its failureReasonNumber. 
 */
int handleFailureMsg(void * unUsed) {
   LOG(LOG_DEBUG, "dht_api::handleFailureMsg() - Received failure message from gnunetd, returning failure to app.\n"); 
   return ntohs(receivedFailureMsg->failureReasonNumber); 
}

/**
 * Handles api-id result message by setting apiId-variable and returning OK. 
 */
int handleApiIdMsg(void * unUsed) {
  apiId = ntohs(receivedResultsMsg->dhtCSHeader.apiId); 
  LOG(LOG_DEBUG, "dht_api::handleApiIdMsg() - got apiId, returning.\n");
  return OK; 
}

/**
 * Handles a create result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be DHT_TableId (in Network Byte Order, of course).  
 *
 * @param tableHandle dht table handle where received table id is to be written. 
 * @return errorCode for the execution
 */

int handleCreateResultMsg(void * vpTableHandle) {
  DHT_TableHandle * tableHandle; 
  DHT_TableId * src; 
  DHT_TableId * dst; 

  tableHandle = vpTableHandle; 

  if(receivedDataSize != 
     sizeof(DHT_CS_REPLY_RESULTS) + sizeof(DHT_TableId)) {

    LOG(LOG_FAILURE, "dht_api::handleCreateResultMsg() - message size is invalid: %d != %d ",
	receivedResultsMsg->header.size, sizeof(DHT_CS_REPLY_RESULTS) + sizeof(DHT_TableId)); 

    tableHandle->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  
  src = (DHT_TableId *)receivedResultsMsg->data; 
  dst = &(tableHandle->tableId); 
  tableHandle->errorCode = OK; 
  ntohHashCode160Copy(dst,src); 

  LOG(LOG_DEBUG, "dht_api::handleCreateResultMsg() - create ok, tableHandle's fields set.\n"); 
  return OK; 
}

/**
 * Handles an insert result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be nonexistent. 
 *
 * @return errorCode for the execution
 */
int handleInsertResultMsg(void * unUsed) {

  if(receivedDataSize != sizeof(DHT_CS_REPLY_RESULTS)) {

    LOG(LOG_FAILURE, "dht_api::handleInsertResultMsg() - message size is invalid: %d != %d ",
	receivedDataSize, sizeof(DHT_CS_REPLY_RESULTS)); 

    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  LOG(LOG_DEBUG, "dht_api::handleInsertResultMsg() - insert ok.\n"); 
  return OK; 
}

/**
 * Handles a fetch result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be DHT_CSFetchResult (variable size). 
 *
 * @param vpResultSet resultset where received data is to be written. 
 * @return errorCode for the execution. 
 */

int handleFetchResultMsg(void * vpResultSet) {
  DHT_ResultSet * resultSet; 
  unsigned short * lengthPtr; 
  char * valuePtr; 
  unsigned short valueCount;
  unsigned short cumulativeLengthOfValues = 0; 
  unsigned short expectedSize; 
  unsigned short i; 

  DHT_ResultSetItem * storedResultSetItem; 
  DHT_ResultSetItem * prevStoredResultSetItem = 0; 

  resultSet = vpResultSet; 

  /*
    Check that it's safe to read valueCount from message. 
   */
  expectedSize = sizeof(DHT_CS_REPLY_RESULTS)+sizeof(unsigned short); 
  if(receivedDataSize < expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleFetchResultMsg() - message size is invalid: %d < %d ",
	receivedDataSize, expectedSize); 

    resultSet->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  
  lengthPtr = (unsigned short*)receivedResultsMsg->data; 
  valueCount = ntohs(*lengthPtr); 

  /*
    Check that it's safe to count cumulative length of values from the message
   */

  expectedSize += valueCount * sizeof(unsigned short); 
  if(receivedDataSize < expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleFetchResultMsg() - message size is invalid: %d < %d ",
	receivedDataSize, expectedSize); 

    resultSet->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Read in the cumulative length of values (to determine final message size)
  */

  lengthPtr ++; 
  for(i=0;i<valueCount;i++) {
    cumulativeLengthOfValues += ntohs(*lengthPtr); 
    lengthPtr++; 
  }
  
  /*
    Check that it's safe to read values from the message
  */

  expectedSize += cumulativeLengthOfValues; 
  if(receivedDataSize != expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleFetchResultMsg() - message size is invalid: %d != %d ",
	receivedDataSize, expectedSize); 

    resultSet->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Read the values and build up the DHT_ResultSet
   */

  valuePtr  = (char *)lengthPtr;
  lengthPtr = (unsigned short *)receivedResultsMsg->data; 
  lengthPtr++; 

  for(i=0;i<valueCount;i++) {
    storedResultSetItem = MALLOC(sizeof(DHT_ResultSetItem)); 
    storedResultSetItem->nextItem         = NULL;
    storedResultSetItem->data = MALLOC(sizeof(DHT_DataContainer)); 
    storedResultSetItem->data->dataLength = ntohs(*lengthPtr);
    storedResultSetItem->data->data       = MALLOC(ntohs(*lengthPtr)); 
    memcpy(storedResultSetItem->data->data,valuePtr,ntohs(*lengthPtr)); 
    if(i==0) {
      resultSet->firstItem = storedResultSetItem; 
    }
    else {
      prevStoredResultSetItem->nextItem = storedResultSetItem; 
    }
    prevStoredResultSetItem = storedResultSetItem; 
    valuePtr += ntohs(*lengthPtr); 
    lengthPtr ++; 
  }

  LOG(LOG_DEBUG, "dht_api::handleFetchResultMsg() - resultset populated, returning.\n"); 
  return OK; 
}

/**
 * Handles leave result message by checking size. 
 *
 * @return errorcode for the execution
 */
int handleLeaveResultMsg(void *unUsed) {
  if(receivedDataSize != sizeof(DHT_CS_REPLY_RESULTS)) {
    LOG(LOG_FAILURE, 
	"dht_api::handleLeaveResultMsg() - Size was wrong: %d != %d\n",
	receivedDataSize, 
	sizeof(DHT_CS_REPLY_RESULTS)); 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  LOG(LOG_DEBUG, "dht_api::handleLeaveResultMsg() - Result message was ok, returning.\n"); 
  return OK; 
}

/**
 * Handles a drop result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be nonexistent. 
 *
 * @return errorCode for the execution
 */
int handleDropResultMsg(void * unUsed) {

  if(receivedDataSize != sizeof(DHT_CS_REPLY_RESULTS)) {

    LOG(LOG_FAILURE, "dht_api::handleDropResultMsg() - message size is invalid: %d != %d ",
	receivedDataSize, sizeof(DHT_CS_REPLY_RESULTS)); 

    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  LOG(LOG_DEBUG, "dht_api::handleDropResultMsg() - Result message was ok, returning.\n"); 
  return OK; 
}

/**
 * Handles data received from DHT-module by dispatching control to
 * correct handler function. If NULL-pointer is given as a handler
 * for specific message type, failure will be logged if that type
 * is received and DHT_ERRORCODES__CS_PROTO_UNEXPECTED_MESSAGE_RECEIVED
 * will be returned. 
 *
 * @param standardMsgHandler function to handle standard reply message
 * @param failureMsgHandler function to handle failure message
 * @param resultsMsgHandler function to handle results message
 * @param statusMsgHandler function to handle status message
 * @param handlerParam parameter that is passed to 
 * @return errorCode for the execution of handler function
 */
int handleReceivedData(int (*standardMsgHandler)(void *),
		       int (*failureMsgHandler)(void *),
		       int (*resultsMsgHandler)(void *),
		       int (*statusMsgHandler)(void *),
		       void *standardHandlerParam,
		       void *failureHandlerParam, 
		       void *resultsHandlerParam,
		       void *statusHandlerParam) {

  switch(receivedDataTcpType) {
  case DHT_CS_PROTO_REPLY_STANDARD:
    if(standardMsgHandler == NULL) {
      LOG(LOG_FAILURE, 
	  "dht_api::handleReceivedData() - Received unexpected standard reply message from gnunetd.\n"); 
      return DHT_ERRORCODES__CS_PROTO_UNEXPECTED_MESSAGE_RECEIVED; 
    }
    else return standardMsgHandler(standardHandlerParam);
  case DHT_CS_PROTO_REPLY_FAILURE:
    if(failureMsgHandler == NULL) {
      LOG(LOG_FAILURE, 
	  "dht_api::handleReceivedData() - Received unexpected failure message from gnunetd.\n"); 
      return DHT_ERRORCODES__CS_PROTO_UNEXPECTED_MESSAGE_RECEIVED; 
    }
    else return failureMsgHandler(failureHandlerParam); 
  case DHT_CS_PROTO_REPLY_RESULTS:
    if(resultsMsgHandler == NULL) {
      LOG(LOG_FAILURE, 
	  "dht_api::handleReceivedData() - Received unexpected results message from gnunetd.\n"); 
      return DHT_ERRORCODES__CS_PROTO_UNEXPECTED_MESSAGE_RECEIVED; 
    }
    else return resultsMsgHandler(resultsHandlerParam); 
  case DHT_CS_PROTO_REPLY_STATUS:
    if(statusMsgHandler == NULL) {
      LOG(LOG_FAILURE, 
	  "dht_api::handleReceivedData() - Received unexpected status message from gnunetd.\n"); 
      return DHT_ERRORCODES__CS_PROTO_UNEXPECTED_MESSAGE_RECEIVED; 
    }
    else return statusMsgHandler(statusHandlerParam); 
  }

  LOG(LOG_FAILURE, 
      "dht_api::handleReceivedData() - Received message was not DHT CS message.\n"); 
  return DHT_ERRORCODES__API_ERROR_UNKNOWN; 
}

/**
 * Handles a join result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be DHT_TableId (in Network Byte Order, of course).  
 *
 * @param vpTableHandle dht table handle where received table id is to be written. 
 * @return errorcode for the execution. 
 */
int handleJoinResultMsg(void * vpTableHandle) {
  DHT_TableHandle * tableHandle; 
  DHT_TableId * src; 
  DHT_TableId * dst; 

  tableHandle = vpTableHandle; 
  
  if(receivedDataSize != 
     sizeof(DHT_CS_REPLY_RESULTS) + sizeof(DHT_TableId)) {

    LOG(LOG_FAILURE, "dht_api::handleJoinResultMsg() - message size is invalid: %d != %d ",
	receivedDataSize, sizeof(DHT_CS_REPLY_RESULTS) + sizeof(DHT_TableId)); 

    tableHandle->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  
  src = (DHT_TableId *)receivedResultsMsg->data; 
  dst = &(tableHandle->tableId); 
  tableHandle->errorCode = OK; 
  ntohHashCode160Copy(dst,src); 

  LOG(LOG_DEBUG, "dht_api::handleJoinResultMsg() - join ok, tableHandle's fields set, returning.\n"); 
  return OK; 
}

/**
 * Handles a tables result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be nicely packed (variable size). 
 *
 * @param vpResultSet tableset where received data is to be written. 
 * @return errorCode for the execution. 
 */
int handleTableResultMsg(void * vpTableSet) {
  DHT_TableSet * tableSet; 
  unsigned short * lengthPtr; 
  char * valuePtr; 
  unsigned short valueCount;
  unsigned short cumulativeLengthOfValues = 0; 
  unsigned short expectedSize; 
  unsigned short i; 

  DHT_TableSetItem * storedTableSetItem; 
  DHT_TableSetItem * prevStoredTableSetItem = 0; 

  tableSet = vpTableSet; 

  /*
    Check that it's safe to read valueCount from message
   */
  LOG(LOG_DEBUG, "dht_api::handleTableResultMsg() - check size for reading valueCount..\n"); 

  expectedSize = sizeof(DHT_CS_REPLY_RESULTS)+sizeof(unsigned short); 
  if(receivedDataSize < expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleTableResultMsg() - message size is invalid: %d < %d \n",
	receivedDataSize, expectedSize); 

    tableSet->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  
  lengthPtr = (unsigned short *)receivedResultsMsg->data;
  valueCount = ntohs(*lengthPtr); 

  LOG(LOG_DEBUG, "dht_api::handleTableResultMsg() - valuecount is %d..\n",
      valueCount); 

  /*
    Check that there is non-pairless number of values received because received data
    should contain pairs. 
  */

  if(valueCount%2 != 0) {
    LOG(LOG_FAILURE, 
	"dht_api::handleTableResultMsg() - Invalid results: pairless count of values when expecting pairs.\n"); 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Check that it's safe to count cumulative length of values from the message
   */

  expectedSize += valueCount * sizeof(unsigned short); 

  LOG(LOG_DEBUG, "dht_api::handleTableResultMsg() - number of values known, check min size.\n"); 

  if(receivedDataSize < expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleTableResultMsg() - message size is invalid: %d < %d \n",
	receivedDataSize, expectedSize); 

    tableSet->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Read in the cumulative length of values (to determine final message size)
  */

  lengthPtr++; /* following valueCount is an array of value lengths. */
  for(i=0;i<valueCount;i++) {
    cumulativeLengthOfValues += ntohs(*lengthPtr); 
    lengthPtr++; 
  }
  
  /*
    Check that it's safe to read values from the message
  */

  LOG(LOG_DEBUG, "dht_api::handleTableResultMsg() - cumulative length is %d.\n",
      cumulativeLengthOfValues); 

  expectedSize += cumulativeLengthOfValues; 
  if(receivedDataSize != expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleTableResultMsg() - message size is invalid: %d != %d \n",
	receivedDataSize, expectedSize); 

    tableSet->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Read the values and build up the DHT_TableSet
   */

  valuePtr  = (char *)lengthPtr;
  lengthPtr = (unsigned short *)receivedResultsMsg->data; 
  lengthPtr++; /* first dataitem is valueCount, we want lengths here */

  for(i=0;i<valueCount;i+=2) {
    storedTableSetItem           = MALLOC(sizeof(DHT_TableSetItem)); 
    storedTableSetItem->nextItem = NULL;
 
    if(i == 0) 
      tableSet->firstItem = storedTableSetItem; 
    else
      prevStoredTableSetItem->nextItem = storedTableSetItem; 

    ntohHashCode160Copy(&(storedTableSetItem->tableId),(DHT_TableId *)valuePtr);
    valuePtr += sizeof(DHT_TableId); 
    lengthPtr++; 
    storedTableSetItem->tableMetaData.name = (MALLOC(ntohs(*lengthPtr))); 
    memcpy(storedTableSetItem->tableMetaData.name, valuePtr, ntohs(*lengthPtr)); 

    prevStoredTableSetItem = storedTableSetItem; 
    valuePtr += ntohs(*lengthPtr); 
    lengthPtr ++; 
  }
  LOG(LOG_DEBUG, "dht_api::handleTableResultMsg() - tableset populated, returning.\n"); 
  return OK; 
}

/**
 * Handles a list inserted result-message, that is, a DHT_CS_REPLY_RESULTS-message where
 * data will be DHT_CSFetchResult (variable size). 
 *
 * @param vpDataList datalist where received data is to be written. 
 * @return errorCode for the execution. 
 */
int handleListInsertedResultMsg(void * vpDataList) {
  DHT_DataList * dataList; 
  DHT_CSFetchResult * csFetchResult; 
  unsigned short * lengthPtr; 
  char * valuePtr; 
  unsigned short valueCount;
  unsigned short cumulativeLengthOfValues = 0; 
  unsigned short expectedSize; 
  unsigned short i; 

  DHT_DataListItem * storedDataListItem; 
  DHT_DataListItem * prevStoredDataListItem = 0; 

  dataList = vpDataList; 

  /*
    Check that it's safe to cast DHT_CSFetchResult from message
   */
  expectedSize = sizeof(DHT_CS_REPLY_RESULTS)+sizeof(DHT_CSFetchResult); 
  if(receivedDataSize < expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleListInsertedResultMsg() - message size is invalid: %d < %d ",
	receivedDataSize, expectedSize); 

    dataList->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }
  
  csFetchResult = (DHT_CSFetchResult *)receivedResultsMsg->data; 
  valueCount = ntohs(csFetchResult->valueCount); 

  /*
    Check that there is n*4 values received because received data
    should contain groups of 4 items. 
  */

  if(valueCount%4 != 0) {
    LOG(LOG_FAILURE, 
	"dht_api::handleListInsertedResultMsg() - Invalid results: count of values %d is not n*4.",
	valueCount); 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Check that it's safe to count cumulative length of values from the message
   */

  expectedSize += valueCount*sizeof(short); 
  if(receivedDataSize < expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleListInsertedResultMsg() - message size is invalid: %d < %d ",
	receivedDataSize, expectedSize); 

    dataList->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Read in the cumulative length of values (to determine final message size)
  */

  lengthPtr = (short *)csFetchResult->data; 
  for(i=0;i<valueCount;i++) {
    cumulativeLengthOfValues += ntohs(*lengthPtr); 
    lengthPtr++; 
  }
  
  /*
    Check that it's safe to read values from the message
  */

  expectedSize += cumulativeLengthOfValues; 
  if(receivedDataSize != expectedSize) {

    LOG(LOG_FAILURE, "dht_api::handleListInsertedResultMsg() - message size is invalid: %d != %d ",
	receivedDataSize, expectedSize); 

    dataList->errorCode = DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
    return DHT_ERRORCODES__CS_PROTO_INVALID_RESULTS; 
  }

  /*
    Read the values and build up the DHT_DataList
   */

  valuePtr  = (char *)(lengthPtr+1);
  lengthPtr = (short *)csFetchResult->data; 

  for(i=0;i<valueCount;i+=4) {
    storedDataListItem           = MALLOC(sizeof(DHT_DataListItem)); 
    storedDataListItem->nextItem = NULL;
 
    if(i == 0) 
      dataList->firstItem = storedDataListItem; 
    else
      prevStoredDataListItem->nextItem = storedDataListItem; 

    /* get key */
    storedDataListItem->key = MALLOC(sizeof(DHT_DataContainer)); 
    storedDataListItem->key->dataLength = ntohs(*lengthPtr);
    storedDataListItem->key->data = MALLOC(ntohs(*lengthPtr));
    memcpy(storedDataListItem->key->data,valuePtr,ntohs(*lengthPtr)); 
    valuePtr += ntohs(*lengthPtr);
    lengthPtr ++;

    /* get value */
    storedDataListItem->value = MALLOC(sizeof(DHT_DataContainer)); 
    storedDataListItem->value->dataLength = ntohs(*lengthPtr);
    storedDataListItem->value->data = MALLOC(ntohs(*lengthPtr));
    memcpy(storedDataListItem->value->data,valuePtr,ntohs(*lengthPtr)); 
    valuePtr += ntohs(*lengthPtr);
    lengthPtr ++;

    /* get hashcode */
    ntohHashCode160Copy(&(storedDataListItem->uniqueReference.hashCode),
			(HashCode160 *)valuePtr);
    valuePtr += ntohs(*lengthPtr);
    lengthPtr ++; 

    /* get insertion time */
    memcpy(&(storedDataListItem->uniqueReference.insertionTime),
	   valuePtr,
	   ntohs(*lengthPtr));
    valuePtr += ntohs(*lengthPtr);
    lengthPtr++;
    
    prevStoredDataListItem = storedDataListItem; 
  }
  LOG(LOG_DEBUG, "dht_api::handleInsertedResultMsg() - dataset populated, returning.\n"); 
  return OK; 
}

/**
 * Initializes API for application use. 
 *
 * @return errorcode for the execution, see error_handling.h
 */
int initializeApi() {
  DHT_CS_REQUEST_API_ID msg; 
  
  apiSemaphore           = SEMAPHORE_NEW(1);

  SEMAPHORE_DOWN(apiSemaphore); /* Blocks subsequent API-function calls */

  LOG(LOG_DEBUG, "dht_api::initializeApi() - starting.\n");

  destroyCalled          = 0; 
  currentRequestId       = 0; 
  receiveThreadHandle    = 0; 
  intSemaphore           = SEMAPHORE_NEW(1); 

  LOG(LOG_DEBUG, "dht_api::initializeApi() - Getting client socket.\n"); 

  sock                   = getClientSocket();
  currentOperationStatus = OPERATION_STATUS_OTHER; 

  LOG(LOG_DEBUG, "dht_api::initializeApi() - Allocating space for receive buffer.\n"); 

  receivedData = MALLOC(64 * 1024);
  if(receivedData == NULL) {
    LOG(LOG_FAILURE, "dht_api::initializeApi() - Failed to allocate message receive buffer.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED);
  }

  receivedStandardReplyMsg = (DHT_CS_REPLY_STANDARD *) receivedData;
  receivedFailureMsg = (DHT_CS_REPLY_FAILURE *) receivedData; 
  receivedResultsMsg = (DHT_CS_REPLY_RESULTS *) receivedData;
  receivedStatusMsg  = (DHT_CS_REPLY_STATUS *) receivedData; 

  SEMAPHORE_DOWN(intSemaphore);
  apiId = 0; 
  currentRequestId = 0; 
  currentRequestId ++; 
  SEMAPHORE_UP(intSemaphore);

  LOG(LOG_DEBUG, "dht_api::initializeApi() - Creating receiveThread.\n"); 

  if (0 != PTHREAD_CREATE(&receiveThreadHandle,
			  (void * (*)()) &receiveThread, 
			  sock,
			  16*1024)) {
    /* TODO: ADD LOG-MESSAGE */
    LOG(LOG_FAILURE, "dht_api::initializeApi() - Failed to create receiver thread.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__API_CS_PROTO_RECEIVER_THREAD_CREATE_FAILED); 
  }
  
  LOG(LOG_DEBUG, "dht_api::initializeApi() - Creating apiId-request message.\n"); 

  memset(&msg,0,sizeof(DHT_CS_REQUEST_API_ID));
  msg.header.size 
    = htons(sizeof(DHT_CS_REQUEST_API_ID));
  msg.header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_API_ID); 
  msg.dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  LOG(LOG_DEBUG, "dht_api::initializeApi() - Writing message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, &msg.header)) {
    LOG(LOG_FAILURE, "dht_api::initializeApi() - Writing message to socket failed.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED);
  }

  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS, 
		      API_INITIALIZE_RESULTS_TIMEOUT); 
  
  if(dataReceived) {
    return apiOperationEnd(handleReceivedData(NULL,
					      handleFailureMsg,
					      handleApiIdMsg,
					      NULL,
					      NULL,NULL,NULL,NULL)); 
  }
  else if(!dataReceived) {
    LOG(LOG_FAILURE, "dht_api::initializeApi() - timeout when waiting for apiId result.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT);
  }

  SEMAPHORE_UP(apiSemaphore);
  return DHT_ERRORCODES__API_ERROR_UNKNOWN; 
}

int destroyApi() {
  void * unused;

  LOG(LOG_DEBUG, "dht_api::destroyApi() - starting.\n");
  SEMAPHORE_DOWN(apiSemaphore);
  SEMAPHORE_DOWN(intSemaphore);
  destroyCalled=1;
  SEMAPHORE_UP(intSemaphore);
  closeSocketTemporarily(sock); /* This will wake receiverThread up. */
  pthread_join(receiveThreadHandle, &unused);
  releaseClientSocket(sock);
  SEMAPHORE_FREE(intSemaphore);
  SEMAPHORE_FREE(apiSemaphore);
  LOG(LOG_DEBUG, "dht_api::destroyApi() -> doneUtil().\n");
  doneUtil();
  LOG(LOG_DEBUG, "dht_api::destroyApi() -> returning.\n");
  return OK; 
}

/**
 * Creates a new DHT-table. 
 *
 * @param meta desciption for the table
 * @param config configuration for the table
 * @return handle to the created table
 */
DHT_TableHandle *create(DHT_TableMetaData * meta,
			DHT_TableConfig * config) {

  DHT_CS_REQUEST_CREATE * createMessage; 
  DHT_TableHandle * tableHandle; 
  unsigned short createMessageSize; /* Used to store create message's (variable) size. */
  int errorCode; 

  LOG(LOG_DEBUG, "dht_api::create() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  tableHandle = MALLOC(sizeof(DHT_TableHandle)); 
  if(tableHandle == NULL) {
    LOG(LOG_FAILURE, "dht_api::create() - Failed to allocate space for table handle.\n"); 
    SEMAPHORE_UP(apiSemaphore); 
    return NULL; 
  }

  LOG(LOG_DEBUG, "dht_api::create() - Creating create-request message.\n"); 

  createMessageSize = sizeof(DHT_CS_REQUEST_CREATE)+strlen(meta->name)+1; 
  createMessage     = MALLOC(createMessageSize);
  if(createMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::create() - Failed to allocate space for create request-message.\n"); 
    tableHandle->errorCode=DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 
    SEMAPHORE_UP(apiSemaphore); 
    return tableHandle;
  }

  memset(createMessage,0,createMessageSize);
  
  createMessage->header.size 
    = htons(createMessageSize);
  createMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_CREATE); 
  createMessage->dhtCSHeader.apiId
    = htons(apiId); 
  createMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  createMessage->tableConfig.replicationCount      = htonl(config->replicationCount);
  createMessage->tableConfig.parallelismCount      = htonl(config->parallelismCount);
  createMessage->tableConfig.maximumValuesPerKey   = htonl(config->maximumValuesPerKey); 
  createMessage->tableConfig.expirationTimeSeconds = htonl(config->expirationTimeSeconds); 
  createMessage->tableConfig.flags                 = htonl(config->flags); 
  createMessage->tableConfig.cacheTimeMultiplier   = htonl(config->cacheTimeMultiplier); 
  strcpy(createMessage->name, meta->name); 
  
  LOG(LOG_DEBUG, "dht_api::create() - Writing create request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)createMessage)) {
    LOG(LOG_FAILURE, "dht_api::create() - Writing create request-message to socket failed.\n"); 
    tableHandle->errorCode = DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED; 
    FREE(createMessage); 
    SEMAPHORE_UP(apiSemaphore); 
    return tableHandle; 
  }

  LOG(LOG_DEBUG, "dht_api::create() - Freeing create request-message.\n"); 
  FREE(createMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK, API_CREATE_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleCreateResultMsg,
				   NULL,
				   NULL,NULL,tableHandle,NULL); 

    if(errorCode != CONTINUE) {
      SEMAPHORE_UP(apiSemaphore);
      tableHandle->errorCode = errorCode; 
      return tableHandle; 
    }

  }
  else {
    LOG(LOG_FAILURE, "dht_api::create() - timeout when waiting for standard reply.\n"); 
    tableHandle->errorCode = DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return tableHandle; 
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_CREATE_RESULTS_TIMEOUT); 
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleCreateResultMsg,
				   NULL,
				   NULL,NULL,tableHandle,NULL); 

    SEMAPHORE_UP(apiSemaphore);
    tableHandle->errorCode = errorCode; 
    return tableHandle; 

  }
  else {
    LOG(LOG_FAILURE, "dht_api::create() - timeout when waiting for create results.\n"); 
    tableHandle->errorCode = DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return tableHandle; 
  }
  tableHandle->errorCode=DHT_ERRORCODES__API_ERROR_UNKNOWN; 
  SEMAPHORE_UP(apiSemaphore); 
  return tableHandle; 
} // create(...)

/**
 * Inserts <key,value>-pair to a DHT-table. 
 *
 * @param tableHandle handle of the table where <key,value>-pair is to be inserted
 * @param key key of the <key,value>-pair
 * @param value value of the <key,value>-pair
 * @return errorcode for the execution, see error_handling.h
 */
int insert(DHT_TableHandle * tableHandle,
	   DHT_DataContainer * key, 
	   DHT_DataContainer * value) {

  DHT_CS_REQUEST_INSERT * insertMessage; 
  int errorCode; 
  unsigned short insertMessageSize; /* Used to store insert message's (variable) size. */

  LOG(LOG_DEBUG, "dht_api::insert() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  LOG(LOG_DEBUG, "dht_api::insert() - Creating insert-request message.\n"); 

  if( sizeof(DHT_CS_REQUEST_INSERT)+key->dataLength+value->dataLength > 65535 ) {
    LOG(LOG_FAILURE, "dht_api::insert() - Too much payload, %d bytes.",
	sizeof(DHT_CS_REQUEST_INSERT)+key->dataLength+value->dataLength); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_PAYLOAD_EXCEEDED); 
  }
    
  insertMessageSize = sizeof(DHT_CS_REQUEST_INSERT)+key->dataLength+value->dataLength; 
  insertMessage     = MALLOC(insertMessageSize);

  if(insertMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::insert() - Failed to allocate space for insert request-message.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED);
  }

  memset(insertMessage,0,insertMessageSize);
  
  insertMessage->header.size 
    = htons(insertMessageSize);
  insertMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_INSERT); 
  insertMessage->dhtCSHeader.apiId
    = htons(apiId); 
  insertMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 

  htonHashCode160Copy(&(insertMessage->targetTableId),&(tableHandle->tableId)); 
  
  insertMessage->keyLength   = htons(key->dataLength);
  insertMessage->valueLength = htons(value->dataLength); 

  memcpy(insertMessage->keyAndValue, key->data, key->dataLength); 
  memcpy(insertMessage->keyAndValue+key->dataLength, value->data, value->dataLength); 
  
  LOG(LOG_DEBUG, "dht_api::insert() - Writing insert request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)insertMessage)) {
    LOG(LOG_FAILURE, "dht_api::insert() - Writing insert request-message to socket failed.\n"); 
    FREE(insertMessage);
    return apiOperationEnd(DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED); 
  }

  LOG(LOG_DEBUG, "dht_api::insert() - Freeing insert request-message.\n"); 
  FREE(insertMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK, 
		      API_INSERT_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleInsertResultMsg,
				   NULL,
				   NULL,NULL,NULL,NULL); 
    
    if(errorCode != CONTINUE) {
      return apiOperationEnd(errorCode); 
    }
  }
  else {
    LOG(LOG_FAILURE, "dht_api::insert() - timeout when waiting for standard reply.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT);
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_INSERT_RESULTS_TIMEOUT);
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleInsertResultMsg,
				   NULL,
				   NULL,NULL,NULL,NULL); 
    
    return apiOperationEnd(errorCode); 

  }
  else {
    LOG(LOG_FAILURE, "dht_api::insert() - timeout when waiting for insert results.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT);
  }
  return apiOperationEnd(DHT_ERRORCODES__API_ERROR_UNKNOWN); 
} // insert(...)


/**
 * Retrieves values that are mapped to given key in a DHT-table.
 *
 * @param table handle of the table where retrieval is to be done
 * @param key key that is used when searching mappings
 * @return set that contains values that are mapped to the key
 */
DHT_ResultSet *fetch(DHT_TableHandle * tableHandle,
		     DHT_DataContainer * key) {

  DHT_CS_REQUEST_FETCH * fetchMessage; 
  DHT_ResultSet * resultSet; 
  int errorCode; 
  unsigned short fetchMessageSize; /* Used to store fetch message's (variable) size. */

  LOG(LOG_DEBUG, "dht_api::fetch() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  resultSet = MALLOC(sizeof(DHT_ResultSet)); 
  if(resultSet == NULL) {
    LOG(LOG_FAILURE, "dht_api::fetch() - Failed to allocate space for resultset.\n"); 
    SEMAPHORE_UP(apiSemaphore); 
    return NULL; 
  }

  LOG(LOG_DEBUG, "dht_api::fetch() - Creating fetch-request message."
   "key.size=%d & key = %s\n", key->dataLength, key->data); 

  fetchMessageSize = sizeof(DHT_CS_REQUEST_FETCH)+key->dataLength; 
  fetchMessage     = MALLOC(fetchMessageSize);
  if(fetchMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::fetch() - Failed to allocate space for fetch request-message.\n"); 
    resultSet->errorCode = DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 
    SEMAPHORE_UP(apiSemaphore); 
    return resultSet;
  }
  memset(fetchMessage,0,fetchMessageSize);
  
  fetchMessage->header.size 
    = htons(fetchMessageSize);
  fetchMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_FETCH); 
  fetchMessage->dhtCSHeader.apiId
    = htons(apiId); 
  fetchMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  htonHashCode160Copy(&fetchMessage->targetTableId, &tableHandle->tableId);
  
  memcpy(fetchMessage->key, key->data, key->dataLength); 
  fetchMessage->keyLength = htonl(key->dataLength); 
  
  LOG(LOG_DEBUG, "dht_api::fetch() - Writing fetch request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)fetchMessage)) {
    LOG(LOG_FAILURE, "dht_api::fetch() - Writing fetch request-message to socket failed.\n"); 
    resultSet->errorCode = DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED; 
    FREE(fetchMessage);
    SEMAPHORE_UP(apiSemaphore); 
    return resultSet; 
  }

  LOG(LOG_DEBUG, "dht_api::fetch() - Freeing fetch request-message.\n"); 
  FREE(fetchMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK, 
		      API_FETCH_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleFetchResultMsg,
				   NULL,
				   NULL,NULL,resultSet,NULL); 
    
    if(errorCode != CONTINUE) {
      resultSet->errorCode = errorCode; 
      SEMAPHORE_UP(apiSemaphore);
      return resultSet;
    }
  }
  else {
    LOG(LOG_FAILURE, "dht_api::fetch() - timeout when waiting for standard reply.\n"); 
    resultSet->errorCode = DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return resultSet; 
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_FETCH_RESULTS_TIMEOUT); 
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleFetchResultMsg,
				   NULL,
				   NULL,NULL,resultSet,NULL); 
    
    resultSet->errorCode = errorCode; 
    SEMAPHORE_UP(apiSemaphore);
    return resultSet;

  }
  else {
    LOG(LOG_FAILURE, "dht_api::fetch() - timeout when waiting for fetch results.\n"); 
    resultSet->errorCode = DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return resultSet; 
  }
  resultSet->errorCode=DHT_ERRORCODES__API_ERROR_UNKNOWN; 
  SEMAPHORE_UP(apiSemaphore); 
  return resultSet; 
} // fetch(...)

/**
 * Joins a DHT-table through other node. 
 *
 * @param nodeAddress address of the other node
 * @param tableId id of the to be joined table
 * @return handle to the joined table
 */
DHT_TableHandle *join(DHT_NodeAddress * nodeAddress,
		      DHT_TableId * tableId) {

  DHT_CS_REQUEST_JOIN * joinMessage; 
  DHT_TableHandle * tableHandle; 
  int errorCode; 
  unsigned short joinMessageSize; /* Used to store join message's (constant) size. */

  LOG(LOG_DEBUG, "dht_api::join() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  tableHandle = MALLOC(sizeof(DHT_TableHandle)); 
  if(tableHandle == NULL) {
    LOG(LOG_FAILURE, "dht_api::join() - Failed to allocate space for table handle.\n"); 
    SEMAPHORE_UP(apiSemaphore); 
    return NULL; 
  }

  LOG(LOG_DEBUG, "dht_api::join() - Creating join-request message.\n"); 

  joinMessageSize = sizeof(DHT_CS_REQUEST_JOIN); 
  joinMessage     = MALLOC(joinMessageSize);
  if(joinMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::join() - Failed to allocate space for join request-message.\n"); 
    tableHandle->errorCode=DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 
    SEMAPHORE_UP(apiSemaphore); 
    return tableHandle;
  }

  memset(joinMessage,0,joinMessageSize);
  
  joinMessage->header.size 
    = htons(joinMessageSize);
  joinMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_JOIN); 
  joinMessage->dhtCSHeader.apiId
    = htons(apiId); 
  joinMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  htonHashCode160Copy(&(joinMessage->helperNodeAddress.hashPubKey),&(nodeAddress->hashPubKey)); 
  htonHashCode160Copy(&(joinMessage->joinedTableId),tableId); 
  
  LOG(LOG_DEBUG, "dht_api::join() - Writing join request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)joinMessage)) {
    LOG(LOG_FAILURE, "dht_api::join() - Writing join request-message to socket failed.\n"); 
    tableHandle->errorCode = DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED; 
    FREE(joinMessage);
    SEMAPHORE_UP(apiSemaphore); 
    return tableHandle; 
  }

  LOG(LOG_DEBUG, "dht_api::join() - Freeing join request-message.\n"); 
  FREE(joinMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK,
		      API_JOIN_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleJoinResultMsg,
				   NULL,
				   NULL,NULL,tableHandle,NULL); 
    
    if(errorCode != CONTINUE) {
      tableHandle->errorCode = errorCode; 
      SEMAPHORE_UP(apiSemaphore);
      return tableHandle;
    }

  }
  else {
    LOG(LOG_FAILURE, "dht_api::join() - timeout when waiting for standard reply.\n"); 
    tableHandle->errorCode = DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return tableHandle; 
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_JOIN_RESULTS_TIMEOUT);
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleJoinResultMsg,
				   NULL,
				   NULL,NULL,tableHandle,NULL); 
    
    tableHandle->errorCode = errorCode; 
    SEMAPHORE_UP(apiSemaphore);
    return tableHandle;
  }
  else {
    LOG(LOG_FAILURE, "dht_api::join() - timeout when waiting for join results.\n"); 
    tableHandle->errorCode = DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return tableHandle; 
  }
  tableHandle->errorCode=DHT_ERRORCODES__API_ERROR_UNKNOWN; 
  SEMAPHORE_UP(apiSemaphore); 
  return tableHandle; 
} // join(...)

/**
 * Leaves a joined DHT-table. 
 *
 * @param tableId id of the to be leaved table
 * @return errorcode for the execution
 */
int leave(DHT_TableHandle * tableHandle) {

  DHT_CS_REQUEST_LEAVE * leaveMessage; 
  int errorCode; 
  unsigned short leaveMessageSize; /* Used to store leave message's (constant) size. */

  LOG(LOG_DEBUG, "dht_api::leave() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  LOG(LOG_DEBUG, "dht_api::leave() - Creating leave-request message.\n"); 

  leaveMessageSize = sizeof(DHT_CS_REQUEST_LEAVE); 
  leaveMessage     = MALLOC(leaveMessageSize);
  if(leaveMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::leave() - Failed to allocate space for leave request-message.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED);
  }

  memset(leaveMessage,0,leaveMessageSize);
  
  leaveMessage->header.size 
    = htons(leaveMessageSize);
  leaveMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_LEAVE); 
  leaveMessage->dhtCSHeader.apiId
    = htons(apiId); 
  leaveMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  htonHashCode160Copy(&(leaveMessage->leavedTableId),&(tableHandle->tableId)); 
  
  LOG(LOG_DEBUG, "dht_api::leave() - Writing leave request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)leaveMessage)) {
    LOG(LOG_FAILURE, "dht_api::leave() - Writing leave request-message to socket failed.\n"); 
    FREE(leaveMessage);
    return apiOperationEnd(DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED); 
  }

  LOG(LOG_DEBUG, "dht_api::leave() - Freeing leave request-message.\n"); 
  FREE(leaveMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK,
		      API_LEAVE_ACK_TIMEOUT); 
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleLeaveResultMsg,
				   NULL,
				   NULL,NULL,NULL,NULL); 
    
    if(errorCode != CONTINUE) {
      return apiOperationEnd(errorCode);
    }

  }
  else {
    LOG(LOG_FAILURE, "dht_api::leave() - timeout when waiting for standard reply.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT);
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_LEAVE_RESULTS_TIMEOUT);
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleLeaveResultMsg,
				   NULL,
				   NULL,NULL,NULL,NULL); 
    
    return apiOperationEnd(errorCode);
  }
  else {
    LOG(LOG_FAILURE, "dht_api::leave() - timeout when waiting for leave results.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT);
  }
  return apiOperationEnd(DHT_ERRORCODES__API_ERROR_UNKNOWN); 
} // leave(...)


/**
 * Retrieves a listing of tables to whom a DHT-node is joined. Information is 
 * received only about public tables.
 *
 * @param address address of the DHT-node whose joined tables are to be listed.
 * @return set that contains information about tables to whom DHT-node is joined.
 */
DHT_TableSet *listTables(DHT_NodeAddress * address) {

  DHT_CS_REQUEST_TABLES * tablesMessage; 
  DHT_TableSet * tableSet; 
  int errorCode; 
  unsigned short tablesMessageSize; /* Used to store tables message's (constant) size. */

  LOG(LOG_DEBUG, "dht_api::tables() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  tableSet = MALLOC(sizeof(DHT_TableSet)); 
  if(tableSet == NULL) {
    LOG(LOG_FAILURE, "dht_api::tables() - Failed to allocate space for tableset.\n"); 
    SEMAPHORE_UP(apiSemaphore); 
    return NULL; 
  }

  LOG(LOG_DEBUG, "dht_api::tables() - Creating tables-request message.\n"); 

  tablesMessageSize = sizeof(DHT_CS_REQUEST_TABLES); 
  tablesMessage     = MALLOC(tablesMessageSize);
  if(tablesMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::tables() - Failed to allocate space for tables request-message.\n"); 
    tableSet->errorCode = DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 
    SEMAPHORE_UP(apiSemaphore); 
    return tableSet;
  }
  memset(tablesMessage,0,tablesMessageSize);
  
  tablesMessage->header.size 
    = htons(tablesMessageSize);
  tablesMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_TABLES); 
  tablesMessage->dhtCSHeader.apiId
    = htons(apiId); 
  tablesMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 

  htonHashCode160Copy(&(tablesMessage->nodeAddress.hashPubKey),
		      &(address->hashPubKey)); 
  
  LOG(LOG_DEBUG, "dht_api::tables() - Writing tables request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)tablesMessage)) {
    LOG(LOG_FAILURE, "dht_api::tables() - Writing tables request-message to socket failed.\n"); 
    tableSet->errorCode = DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED; 
    FREE(tablesMessage);
    SEMAPHORE_UP(apiSemaphore); 
    return tableSet; 
  }

  LOG(LOG_DEBUG, "dht_api::tables() - Freeing tables request-message.\n"); 
  FREE(tablesMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK, 
		      API_TABLES_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleTableResultMsg,
				   NULL,
				   NULL,NULL,tableSet,NULL); 
    
    if(errorCode != CONTINUE) {
      tableSet->errorCode = errorCode; 
      SEMAPHORE_UP(apiSemaphore);
      return tableSet;
    }
  }
  else {
    LOG(LOG_FAILURE, "dht_api::tables() - timeout when waiting for standard reply.\n"); 
    tableSet->errorCode = DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return tableSet; 
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_TABLES_RESULTS_TIMEOUT); 
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleTableResultMsg,
				   NULL,
				   NULL,NULL,tableSet,NULL); 
    
    tableSet->errorCode = errorCode; 
    SEMAPHORE_UP(apiSemaphore);
    return tableSet;

  }
  else {
    LOG(LOG_FAILURE, "dht_api::tables() - timeout when waiting for tables results.\n"); 
    tableSet->errorCode = DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return tableSet; 
  }
  tableSet->errorCode=DHT_ERRORCODES__API_ERROR_UNKNOWN; 
  SEMAPHORE_UP(apiSemaphore); 
  return tableSet; 
} // listTables(...)
 
/**
 * Retrieves a listing of data that DHT-node has inserted into a table. 
 *
 * @param handle of the table where data is inserted
 * @return set that contains information about all inserted keys. 
 */
DHT_DataList *listInsertedData(DHT_TableHandle * tableHandle) {

  DHT_CS_REQUEST_INSERTED * insertedMessage; 
  DHT_DataList * dataList; 
  int errorCode; 
  unsigned short insertedMessageSize; /* Used to store inserted message's (constant) size. */

  LOG(LOG_DEBUG, "dht_api::inserted() - starting.\n");

  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */

  dataList = MALLOC(sizeof(DHT_DataList)); 
  if(dataList == NULL) {
    LOG(LOG_FAILURE, "dht_api::inserted() - Failed to allocate space for datalist.\n"); 
    SEMAPHORE_UP(apiSemaphore); 
    return NULL; 
  }

  LOG(LOG_DEBUG, "dht_api::inserted() - Creating inserted-request message.\n"); 

  insertedMessageSize = sizeof(DHT_CS_REQUEST_INSERTED); 
  insertedMessage     = MALLOC(insertedMessageSize);
  if(insertedMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::inserted() - Failed to allocate space for inserted request-message.\n"); 
    dataList->errorCode = DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 
    SEMAPHORE_UP(apiSemaphore); 
    return dataList;
  }
  memset(insertedMessage,0,insertedMessageSize);
  
  insertedMessage->header.size 
    = htons(insertedMessageSize);
  insertedMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_INSERTED); 
  insertedMessage->dhtCSHeader.apiId
    = htons(apiId); 
  insertedMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  htonHashCode160Copy(&(insertedMessage->insertedTableId),
		      &(tableHandle->tableId)); 
  
  LOG(LOG_DEBUG, "dht_api::inserted() - Writing inserted request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)insertedMessage)) {
    LOG(LOG_FAILURE, "dht_api::inserted() - Writing inserted request-message to socket failed.\n"); 
    dataList->errorCode = DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED; 
    FREE(insertedMessage);
    SEMAPHORE_UP(apiSemaphore); 
    return dataList; 
  }

  LOG(LOG_DEBUG, "dht_api::inserted() - Freeing inserted request-message.\n"); 
  FREE(insertedMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK, 
		      API_INSERTED_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleListInsertedResultMsg,
				   NULL,
				   NULL,NULL,dataList,NULL); 
    
    if(errorCode != CONTINUE) {
      dataList->errorCode = errorCode; 
      SEMAPHORE_UP(apiSemaphore);
      return dataList;
    }
  }
  else {
    LOG(LOG_FAILURE, "dht_api::inserted() - timeout when waiting for standard reply.\n"); 
    dataList->errorCode = DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return dataList; 
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_INSERTED_RESULTS_TIMEOUT); 
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleListInsertedResultMsg,
				   NULL,
				   NULL,NULL,dataList,NULL); 
    
    dataList->errorCode = errorCode; 
    SEMAPHORE_UP(apiSemaphore);
    return dataList;

  }
  else {
    LOG(LOG_FAILURE, "dht_api::inserted() - timeout when waiting for inserted results.\n"); 
    dataList->errorCode = DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT;
    SEMAPHORE_UP(apiSemaphore);
    return dataList; 
  }
  dataList->errorCode=DHT_ERRORCODES__API_ERROR_UNKNOWN; 
  SEMAPHORE_UP(apiSemaphore); 
  return dataList; 
} // listInserted(...)

/**
 * Drops data that is inserted to a DHT-table by DHT-node.
 *
 * @param table table where <key,value>-pair is inserted
 * @param uniqueReference reference to data to drop
 * @return errorcode for the execution, see error_handling.h
 */
int dropInsertedData(DHT_TableHandle * table, 
		     DHT_StoredDataReference * uniqueReference) {

  DHT_CS_REQUEST_DROP * dropMessage; 
  int errorCode; 
  unsigned short dropMessageSize; /* Used to store drop message's (constant) size. */

  LOG(LOG_DEBUG, "dht_api::drop() - starting.\n");
  apiOperationStart(); /* block subsequent api-operations & increment currentRequestId */
  LOG(LOG_DEBUG, "dht_api::drop() - Creating drop-request message.\n"); 

  dropMessageSize = sizeof(DHT_CS_REQUEST_DROP); 
  dropMessage     = MALLOC(dropMessageSize);
  if(dropMessage == NULL) {
    LOG(LOG_FAILURE, "dht_api::drop() - Failed to allocate space for drop request-message.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED); 
  }
  memset(dropMessage,0,dropMessageSize);
  
  dropMessage->header.size 
    = htons(dropMessageSize);
  dropMessage->header.tcpType
    = htons(DHT_CS_PROTO_REQUEST_DROP); 
  dropMessage->dhtCSHeader.apiId
    = htons(apiId); 
  dropMessage->dhtCSHeader.requestId
    = htons(currentRequestId); 
  
  htonHashCode160Copy(&(dropMessage->insertedTableId),
		      &(table->tableId)); 
  
  htonHashCode160Copy(&(dropMessage->droppedDataReference.hashCode),
		      &(uniqueReference->hashCode));
  
  dropMessage->droppedDataReference.insertionTime
    = htonl(uniqueReference->insertionTime);  

  LOG(LOG_DEBUG, "dht_api::drop() - Writing drop request-message to socket.\n"); 

  if (SYSERR == writeToSocket(sock, (CS_HEADER *)dropMessage)) {
    LOG(LOG_FAILURE, "dht_api::drop() - Writing drop request-message to socket failed.\n"); 
    FREE(dropMessage);
    return apiOperationEnd(DHT_ERRORCODES__API_CS_PROTO_WRITE_TO_SOCKET_FAILED); 
  }

  LOG(LOG_DEBUG, "dht_api::drop() - Freeing drop request-message.\n"); 
  FREE(dropMessage);

  apiOperationReceive(OPERATION_STATUS_WAITFOR_ACK, 
		      API_DROP_ACK_TIMEOUT);
  
  if(dataReceived) {
    
    errorCode = handleReceivedData(handleStandardMsg,
				   handleFailureMsg,
				   handleDropResultMsg,
				   NULL,
				   NULL,NULL,NULL,NULL); 
    
    if(errorCode != CONTINUE) {
      return apiOperationEnd(errorCode); 
    }
  }
  else {
    LOG(LOG_FAILURE, "dht_api::drop() - timeout when waiting for standard reply.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_ACK_TIMEOUT);
  }
  
  /*
    ACK received -> start waiting for results. 
  */
  
  apiOperationReceive(OPERATION_STATUS_WAITFOR_RESULTS,
		      API_DROP_RESULTS_TIMEOUT); 
  
  if(dataReceived) {

    errorCode = handleReceivedData(NULL,
				   handleFailureMsg,
				   handleDropResultMsg,
				   NULL,
				   NULL,NULL,NULL,NULL); 
    
    return apiOperationEnd(errorCode);

  }
  else {
    LOG(LOG_FAILURE, "dht_api::drop() - timeout when waiting for drop results.\n"); 
    return apiOperationEnd(DHT_ERRORCODES__CS_PROTO_RESULTS_TIMEOUT);
  }

  return apiOperationEnd(DHT_ERRORCODES__API_ERROR_UNKNOWN); 

} // dropInsertedData(...)


/* ************************* UTIL FUNCTIONS ********************************** */
/* Implementations for API's util functions                                    */

int isError(int errorCode) {
  return (errorCode != OK); 
} 

int resultSetNext(DHT_ResultSet * resultSet, DHT_DataContainer ** target) {
  DHT_ResultSetItem * newFirstItem; 

  (*target) = NULL;

  if(resultSet->firstItem == NULL)
    return DHT_ERRORCODES__API_NO_SUCH_ELEMENT; 

  (*target) = MALLOC(sizeof(DHT_DataContainer)); 
  
  if((*target) == NULL)
    return DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 

  (*target)->dataLength = resultSet->firstItem->data->dataLength;
  (*target)->data       = resultSet->firstItem->data->data; 

  newFirstItem = resultSet->firstItem->nextItem; 
  FREE(resultSet->firstItem);
  resultSet->firstItem = newFirstItem; 

  return OK; 
} 


int tableSetNext(DHT_TableSet * tableSet, DHT_TableSetItem ** target) {
  DHT_TableSetItem * newFirstItem; 

  (*target) = NULL;

  if(tableSet->firstItem == NULL)
    return DHT_ERRORCODES__API_NO_SUCH_ELEMENT; 

  (*target) = MALLOC(sizeof(DHT_TableSetItem)); 
  
  if((*target) == NULL)
    return DHT_ERRORCODES__API_MEMORY_ALLOCATION_FAILED; 

  memcpy(&((*target)->tableId), 
	 &(tableSet->firstItem->tableId), 
	 sizeof(DHT_TableId));

  (*target)->tableMetaData.name
    = tableSet->firstItem->tableMetaData.name; 

  newFirstItem = tableSet->firstItem->nextItem; 
  FREE(tableSet->firstItem);
  tableSet->firstItem = newFirstItem; 

  return OK; 
} 

