 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */

/**
 * @file module/dht_util.c
 * @brief Common utility functions for the GNUnet/DHT module
 * @author Risto Saarelma
 * @version $Id: dht_util.c,v 1.18 2004/05/18 19:01:31 rsaarelm Exp $
 */

#include "dht_util.h"

HashCode160 * getNodeAddressHash(
    DHT_NodeAddress * contact) {
  return &contact->hashPubKey;
}

int contactsEqual(
    DHT_NodeAddress * c1,
    DHT_NodeAddress * c2) {
  return equalsHashCode160(getNodeAddressHash(c1), getNodeAddressHash(c2));
}

void xorMetricDistance(
    HashCode160 * h1,
    HashCode160 * h2,
    HashCode160 * result) {
  xorHashCodes(h1, h2, result);
}

int getHashCodeBit(HashCode160 * code, int bit) {
  /* The following code depends on the internal structure of 
   * GNUnet's HashCode160.  The order of bits can be pretty much
   * arbitrary as long as this is the only place where the
   * hash code's innards are being poked.
   */
  int bits_per_int = 32;

  switch (bit / bits_per_int) {
    case 0:
      if (code->e & (1 << bit%bits_per_int))
        return bit;
      break;
    case 1:
      if (code->d & (1 << bit%bits_per_int))
        return bit;
      break;
    case 2:
      if (code->c & (1 << bit%bits_per_int))
        return bit;
      break;
    case 3:
      if (code->b & (1 << bit%bits_per_int))
        return bit;
      break;
    case 4:
      if (code->a & (1 << bit%bits_per_int))
        return bit;
      break;
    default:
      LOG(LOG_WARNING, "Invalid bit index %d.\n", bit);
  }
  return 0;
}

int dhtHashCodeCompare(
    HashCode160 * h1,
    HashCode160 * h2) {
  int i;
  int diff;
  for (i = 160 - 1; i >= 0; --i) {
    diff = getHashCodeBit(h2, i) - getHashCodeBit(h1, i);
    if (diff < 0) 
      return -1;
    else if (diff > 0)
      return 1;
  }
  return 0;
}

int dhtHashCodeCompareDistance(
    HashCode160 * h1,
    HashCode160 * h2,
    HashCode160 * target) {
  int i;
  int diff;
  int b1, b2, bt;
  for (i = 160 - 1; i >= 0; --i) {
    b1 = getHashCodeBit(h1, i);
    b2 = getHashCodeBit(h2, i);
    bt = getHashCodeBit(target, i);
    /* Check XOR distance. */
    diff = (b2 ^ bt) - (b1 ^ bt);
    if (diff < 0) 
      return -1;
    else if (diff > 0)
      return 1;
  }
  return 0;
}

/**
 * memory layout of the keyValue->data is
 * &lt;headerKey,headerData,valueKey,valueData&gt;
 * */
int convertKeyValueToHashNData(
      DHT_DataContainer * key, 
      DHT_DataContainer * value,
      HashCode160       ** hashedKey,
      DHT_DataContainer ** keyValue) 
{
  HashCode160 * theHash;
  DHT_DataContainer * theData;
  int offset;

  theHash   = MALLOC(sizeof(HashCode160));
  hash(key->data, key->dataLength, theHash);
  theData   = MALLOC(sizeof(DHT_DataContainer));
  offset    = sizeof(DHT_DataContainer) /* key header */
              + key->dataLength
              + sizeof(DHT_DataContainer) /* value header */
              + value->dataLength;
  theData->dataLength = offset;
  theData->data = MALLOC(offset);
  
  offset = 0;
  /*printf("offset == %d\n",offset);*/
  /* first key header */
  memcpy( theData->data,
          key,
          sizeof(DHT_DataContainer));
  offset +=  sizeof(DHT_DataContainer);
  /*printf("offset == %d\n",offset);*/
  /* then key data */
  memcpy( theData->data+offset,
          key->data,
          key->dataLength);
  offset +=  key->dataLength;
  /*printf("offset == %d\n",offset);*/
  /* then value header */
  memcpy( theData->data+offset,
          value,
          sizeof(DHT_DataContainer));
  offset +=  sizeof(DHT_DataContainer);
  /*printf("offset == %d\n",offset);*/
  /* last value data */
  memcpy( theData->data+offset,
          value->data,
          value->dataLength);
  offset +=  value->dataLength;
  
  /*printf("offset == %d\n",offset);*/
  *hashedKey = theHash;
  *keyValue = theData;
  
  return OK;
}
      
/* */
int convertDataToKeyNValue(
      DHT_DataContainer * keyValue,
      DHT_DataContainer ** key, 
      DHT_DataContainer ** value) 
{
  /* We could do it the hack way and use the knowledge that 
     keyvalue is formed like this [KeyHeader,KeyData,ValueKey,ValueData]
     The problem is that it segfaults if someone frees anything else 
     than KeyHeader.
     So lets do it the safe way and use a lot of malloc and memcpy
   */
  DHT_DataContainer * theKey = NULL;
  DHT_DataContainer * theValue = NULL;
  int offset;
  int size;

  size = keyValue->dataLength;  
  /* first of all we should have more than both headers */
  if (size < 2*sizeof(DHT_DataContainer)) {
    LOG(LOG_FAILURE, 
      "FAILURE: dht_util::convertDataToKeyValue() - "
      "malformed keyValue in conversion datalength was %d bytes while "
      "conversion needed at least %d bytes\n",
      size,
      (2*sizeof(DHT_DataContainer)));
    *key    = NULL;
    *value  = NULL;
    return SYSERR;
  }
  /* KEY */
  theKey    = MALLOC(sizeof(DHT_DataContainer)); /*malloc:DHT_DataContainer*/
  offset = 0;

  /*printf("offset == %d\n",offset);*/
  /* key header */
  memcpy( theKey,
          keyValue->data,
          sizeof(DHT_DataContainer));
  /* now we should have more than the datalength*/
  offset +=  sizeof(DHT_DataContainer);
  /*printf("offset == %d\n",offset);*/
  if (size < 2*sizeof(DHT_DataContainer) + theKey->dataLength) {
    LOG(LOG_FAILURE, 
      "FAILURE: dht_util::convertDataToKeyValue() - "
      "malformed keyValue in conversion while getting key->data. "
      "Datalength was %d bytes while conversion needed at least %d bytes\n",
      size,
      (2*sizeof(DHT_DataContainer) + theKey->dataLength));
    FREE(theKey);
    *key    = NULL;
    *value  = NULL;
    return SYSERR;
  }
  if(theKey->dataLength == 0) {
    theKey->data = NULL;
  } else {
    theKey->data = MALLOC(theKey->dataLength); /*malloc:void*/
    /* key data */
    memcpy( theKey->data,
           keyValue->data+offset,
            theKey->dataLength);
  }
  offset +=  theKey->dataLength;
  /*printf("offset == %d\n",offset);*/

  /* VALUE */
  theValue  = MALLOC(sizeof(DHT_DataContainer)); /*malloc:DHT_DataContainer*/
  /* value header */
  memcpy( theValue,
          keyValue->data+offset,
          sizeof(DHT_DataContainer));
  offset +=  sizeof(DHT_DataContainer);
  /*printf("offset == %d\n",offset);*/
  /* value data */
  if (size != offset + theValue->dataLength) {
    LOG(LOG_FAILURE, 
      "FAILURE: dht_util::convertDataToKeyValue() - "
      "malformed keyValue in conversion while checking result. "
      "Conversions datalength was %d bytes while "
      "conversion used %d byte\n",
      size,
      offset + theValue->dataLength
      );
    /* clean up and return error */
    FREE(theKey->data);     /*free:void*/
    FREE(theValue);         /*free:DHT_DataContainer*/
    FREE(theKey);           /*free:DHT_DataContainer*/
    *key    = NULL;
    *value  = NULL;
    return SYSERR;
  } else {
    /*printf("offset == %d\n",offset+theValue->dataLength);*/
    if(theValue->dataLength == 0) {
      theValue->data = NULL;
    } else {
      theValue->data = MALLOC(theValue->dataLength); /*malloc:void*/
      memcpy( theValue->data,
              keyValue->data+offset,
              theValue->dataLength);
    }
    *key    = theKey;
    *value  = theValue;
    return OK;
  }
}


int equalsDataContainers(DHT_DataContainer * data1, DHT_DataContainer * data2) {
  if (data1->dataLength == data2->dataLength) {
    return (0 == memcmp(data1->data,data2->data,sizeof(data1->dataLength)));
  } else {
    return 0;
  }
}

DHT_DataContainer * newDataContainer(
    unsigned int dataLength,
    void * data) {
  DHT_DataContainer * result = MALLOC(sizeof(DHT_DataContainer));
  result->dataLength = dataLength;
  result->data = MALLOC(dataLength);
  memcpy(result->data, data, dataLength);
  return result;
}

DHT_DataContainer * cloneDataContainer(
    DHT_DataContainer * source) {
  DHT_DataContainer * result = MALLOC(sizeof(DHT_DataContainer));
  result->dataLength = source->dataLength;
  result->data = MALLOC(result->dataLength);
  memcpy(result->data, source->data, result->dataLength);
  return result;
}


void freeDataContainer(
    DHT_DataContainer * container) {
  FREE(container->data);
  FREE(container);
}

char * hash160ToString(const HashCode160 * hash) {
  static HexName hex;
  if (hash != NULL) {
    hash2hex(hash, &hex);
  } else {
    return "NULL";
  }
  return hex.data;
}

void serializeTableData(
    const DHT_TableId * tableId,
    const DHT_TableMetaData * metaData,
    void ** data,
    size_t * dataLength) {
  *dataLength = 
    sizeof(DHT_TableId) + strlen(metaData->name) + 1;
  *data = MALLOC(*dataLength);

  memcpy(*data, tableId, sizeof(DHT_TableId));
  strcpy(*data + sizeof(DHT_TableId), metaData->name);
}

void deserializeTableData(
    DHT_TableId * tableId,
    DHT_TableMetaData * metaData,
    const void * data,
    const size_t dataLength) {
  memcpy(tableId, data, sizeof(DHT_TableId));
  char * name = (char *)(data + sizeof(DHT_TableId));
  int stringLength = strlen(name);
  metaData->name = MALLOC(stringLength + 1);
  strcpy(metaData->name, name);
}

void addToTableSet(
    DHT_TableSet * tableSet,
    const DHT_TableId * tableId,
    const DHT_TableMetaData * metaData) {
  DHT_TableSetItem * item = MALLOC(sizeof(DHT_TableSetItem));
  memcpy(&item->tableId, tableId, sizeof(DHT_TableId));
  item->tableMetaData.name = STRDUP(metaData->name);
  item->nextItem = tableSet->firstItem;
  tableSet->firstItem = item;
}
