 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */

/**
 * @file module/dht_util.h
 * @brief Common utility functions for the GNUnet/DHT module
 * @author Risto Saarelma
 * @version $Id: dht_util.h,v 1.9 2004/05/18 15:10:21 rsaarelm Exp $
 */

#ifndef DHT_UTIL_H
#define DHT_UTIL_H

#include "api_structs.h"

struct Vector;

/**
 * @return A hashcode of a DHT node address (from the address struct; do not
 * free the returned pointer)
 */
HashCode160 * getNodeAddressHash(
    DHT_NodeAddress * contact);

/**
 * @return Whether two DHT node addresses are equal
 */
int contactsEqual(
    DHT_NodeAddress * c1,
    DHT_NodeAddress * c2);

/**
 * Writes the XOR-metric distance between hashcodes \a h1 and \a h2, defined
 * as a1 XOR a2 in \a result.
 */
void xorMetricDistance(
    HashCode160 * h1,
    HashCode160 * h2,
    HashCode160 * result);

/**
 * @return Bit \a bit from hashcode \a code.
 */
int getHashCodeBit(
    HashCode160 * code,
    int bit);

/**
 * @return -1 if \a h1 is greater than \a h2, 0 if \a h1 and \a h2 are
 * equal and 1 if \a h2 is greater than \a h1.
 */
int dhtHashCodeCompare(
    HashCode160 * h1,
    HashCode160 * h2);

/**
 * @return -1 if \a h1 is closer to \a target than \a h2, 0 if \a h1 and \a
 * h2 are equal and 1 if \a h2 is closer to \a target than \a h1.
 */

int dhtHashCodeCompareDistance(
    HashCode160 * h1,
    HashCode160 * h2,
    HashCode160 * target);

/**
 * Function convertKeyValueToHashNData() converts Key,Value pairs to Hash,Data 
 * pairs. The hash is formed from the key and the data contains the key and
 * value catenated.
 * @param key
 * @param value
 * @param hash
 * @param data
 * @return OK
 **/
int convertKeyValueToHashNData(
      DHT_DataContainer * key, 
      DHT_DataContainer * value,
      HashCode160       ** hash,
      DHT_DataContainer ** data);

/**
 * Function convertDataToKeyNValue() converts Data back to Key,Value pair.
 * 
 * @param data
 * @param key
 * @param value
 * @param hash
 * @return OK if succesfull SYSERR if the data was malformed
 **/
int convertDataToKeyNValue(
      DHT_DataContainer * data,
      DHT_DataContainer ** key, 
      DHT_DataContainer ** value);

/**
 * Function equalsDataContainers() tests two DHT_DataContainers for equality.
 * 
 * @param data1
 * @param data2
 * @return int accordingly
 */
int equalsDataContainers(DHT_DataContainer * data1, DHT_DataContainer * data2);

/**
 * Creates a new DHT_DataContainer with \a dataLength bytes of \a data copied
 * into its internal data.
 *
 * The pair of newDataContainer and freeDataContainer can be handy with RPC
 * calls.
 * 
 * @return the newly allocated container
 */

DHT_DataContainer * newDataContainer(
    unsigned int dataLength,
    void * data);

/**
 * @return a newly allocated data container with data copied from \a source.
 */
DHT_DataContainer * cloneDataContainer(
    DHT_DataContainer * source);

/**
 * Frees a DHT_DataContainer and the memory it holds.  Use together with
 * newDataContainer.
 */
void freeDataContainer(
    DHT_DataContainer * container);

/**
 * Creates a printable representation of the HashCode160 into an internal
 * buffer and returns pointer to it.  Not thread safe.  Previous return
 * values become invalid every time the function is called.
 */

char * hash160ToString(
    const HashCode160 * hash);

void serializeTableData(
    const DHT_TableId * tableId,
    const DHT_TableMetaData * metaData,
    void ** data,
    size_t * dataLength);

void deserializeTableData(
    DHT_TableId * tableId,
    DHT_TableMetaData * metaData,
    const void * data,
    const size_t dataLength);

void addToTableSet(
    DHT_TableSet * tableSet,
    const DHT_TableId * tableId,
    const DHT_TableMetaData * metaData);

#endif /* DHT_UTIL_H */
