 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */

/**
 * @file module/kademlia.c
 * @brief Implementation of the Kademlia algorithm for GNUnet
 * @author Risto Saarelma
 * @version $Id: kademlia.c,v 1.27 2004/05/23 17:09:44 rsaarelm Exp $
 */

#include "kademlia.h"
#include "vector.h"
#include "dht_util.h"
#include "datalayer.h"
#include "api_structs.h"
#include "gnunet_util.h"

/* Hack to get stdout logs during unit tests. */
/*
#define KADEMLIA_LOG printf( 
 */
#define KADEMLIA_LOG LOG(LOG_WARNING,

/* To make the editor not go crazy over mismatched parentheses. */
#define KADEMLIA_GOL );
   
/* Note: making an assumption here that an instance of the program
 * will only manage one node, therefore can use a global collection
 * of routing tables instead of a node-specific structure.
 */
static Vector * routingTables = NULL; 

static CoreAPIForApplication * capi = NULL;

/************************* General *************************/

int Kademlia_init(CoreAPIForApplication * coreApi) {
  int ok = 0;

  capi = coreApi;
  routingTables = vectorNew();
  ok = Kademlia_registerRpcCalls();

  KADEMLIA_LOG
      "Kademlia_init: Done, RPC call registering status: %d (0 == ok).\n",
      ok
  KADEMLIA_GOL
  
  return ok;
}

/************************* Routing Tables *************************/


Kademlia_RoutingTable * Kademlia_tableLookup(
    DHT_TableId * tableId) {
  if (routingTables == NULL) {
    KADEMLIA_LOG
      "Kademlia_tableLookup: Routing table container not initialized.\n"
    KADEMLIA_GOL

    return NULL;
  }
  Kademlia_RoutingTable * result = NULL;
  vectorLock(routingTables);

  for (result = (Kademlia_RoutingTable *)vectorGetFirst(routingTables); 
      result != NULL;
      result = (Kademlia_RoutingTable *)vectorGetNext(routingTables)) {
    if (equalsHashCode160(&result->tableId, tableId)) 
      break;
  }
  vectorUnlock(routingTables);
  KADEMLIA_LOG 
    (result 
    ? "Kademlia_tableLookup: Found table %s.\n"
    : "Kademlia_tableLookup: Did not find table %s.\n"),
    hash160ToString(tableId)
  KADEMLIA_GOL
  return result;
}

Kademlia_RoutingTable * Kademlia_registerTable(
    DHT_TableId * tableId,
    DHT_NodeAddress * localId,
    int bucketSize) {
  Kademlia_RoutingTable * newRoutingTable;
  /* If this is the first routing table created, will create the
   * routing table container as well.
   */
  if (routingTables == NULL) {
    routingTables = vectorNew();
  }

  /* Does the tableId already have a table? */
  newRoutingTable = Kademlia_tableLookup(tableId);
  if (newRoutingTable != NULL) {
    KADEMLIA_LOG
        "Kademlia_registerTable: An existing routing table is being "
        "re-registered.\nFreeing existing table first.\n"
    KADEMLIA_GOL
    Kademlia_freeRoutingTable(newRoutingTable);
  } else {
    newRoutingTable = MALLOC(sizeof(Kademlia_RoutingTable));
    vectorInsertLast(routingTables, newRoutingTable);
  }
  Kademlia_initRoutingTable(
      newRoutingTable,
      tableId,
      localId,
      bucketSize);
  KADEMLIA_LOG
      "Kademlia_registerTable: registered new table %s of %d bucket size ",
      hash160ToString(tableId),
      bucketSize
  KADEMLIA_GOL
  KADEMLIA_LOG
      "in node %s.\n",
      hash160ToString(getNodeAddressHash(localId))
  KADEMLIA_GOL
  return newRoutingTable;
}

void Kademlia_joinTable(
    DHT_NodeAddress * tableNode,
    DHT_TableId * tableId) {
  if (capi == NULL) {
    KADEMLIA_LOG
        "Kademlia_joinTable: Failure: Kademlia is not initialized.\n"
    KADEMLIA_GOL
    return;
  }

  KADEMLIA_LOG
      "Kademlia_joinTable: Node %s ",
      hash160ToString(getNodeAddressHash(capi->myIdentity))
  KADEMLIA_GOL

  KADEMLIA_LOG
      "tries to join table %s.\n",
      hash160ToString(tableId)
  KADEMLIA_GOL

  Kademlia_RoutingTable * routingTable = 
    Kademlia_tableLookup(tableId);

  if (routingTable == NULL) {
    /* Joining a new table, do initialization stuff.
     */
    routingTable = Kademlia_registerTable(
        tableId,
        capi->myIdentity,
        KADEMLIA_BUCKET_SIZE);

    /* Get some contacts in the new table.
     */
    Kademlia_newContact(
        routingTable,
        tableNode);

    /* Node lookup causes contacts to be registered in
     * the routing table.  Lookup on local address is an
     * initiation procedure specified by the Kademlia
     * algorithm paper.
     */
    Vector * dummyVec = vectorNew();
    nodeLookup(
        tableId,
        getNodeAddressHash(capi->myIdentity),
        dummyVec);
    vectorFree(dummyVec);

    /* TODO: Make the values more sensible. */
    DHT_TableMetaData metaData;
    metaData.name = "dummy_name";
    DHT_TableConfig tableConfig;
    tableConfig.replicationCount = KADEMLIA_BUCKET_SIZE;
    tableConfig.parallelismCount = 1;
    tableConfig.maximumValuesPerKey = 10;
    tableConfig.expirationTimeSeconds = 3600;
    tableConfig.flags = 0;
    tableConfig.cacheTimeMultiplier = 1.0f;

    localStoreCreateTable(tableId, &metaData, &tableConfig);

  }

}

void Kademlia_clearTables() {
  if (routingTables == NULL) 
    return;
  vectorLock(routingTables);
  while (vectorSize(routingTables) > 0) {
    Kademlia_freeRoutingTable(
        (Kademlia_RoutingTable *)vectorGetLast(routingTables));
    vectorRemoveLast(routingTables);
  }

  vectorUnlock(routingTables);
}

void Kademlia_initRoutingTable(
    Kademlia_RoutingTable * routingTable,
    DHT_TableId * tableId,
    DHT_NodeAddress * localId,
    int bucketSize) {
  int i;
  for (i = 0; i < KADEMLIA_N_BITS; ++i) {
    routingTable->kBuckets[i] = vectorNew();
  }
  memcpy(&routingTable->localId, localId, sizeof(DHT_NodeAddress));
  memcpy(&routingTable->tableId, tableId, sizeof(DHT_TableId));
  routingTable->bucketSize = bucketSize;
}

void Kademlia_freeRoutingTable(
    Kademlia_RoutingTable * routingTable) {
  int i;
  if (routingTable == NULL) {
    LOG(LOG_WARNING, "Kademlia_initRoutingTable: Tried to free a null routing table.\n");
    return;
  }
  for (i = 0; i < KADEMLIA_N_BITS; ++i) {
    vectorFree(routingTable->kBuckets[i]);
  }
}

/************************* Node Contacts *************************/

void Kademlia_newContact(
    Kademlia_RoutingTable * routingTable,
    DHT_NodeAddress * contact) {
  KADEMLIA_LOG
      "Kademlia_newContact: New contact %s ",
      hash160ToString(getNodeAddressHash(contact))
  KADEMLIA_GOL
  KADEMLIA_LOG
      " for routing table %s",
      hash160ToString(&routingTable->tableId)
  KADEMLIA_GOL
  
  if (dhtHashCodeCompare(
        getNodeAddressHash(contact), 
        getNodeAddressHash(&routingTable->localId)) == 0) {
    KADEMLIA_LOG
      " is the local node.  Abandoning it.\n"
    KADEMLIA_GOL
    return;
  }   

  Vector * kBucket = routingTable->kBuckets[kBucketIndex(
      routingTable,
      getNodeAddressHash(contact))];
  DHT_NodeAddress * newContact = contact;
  int oldIndex = vectorIndexOf(kBucket, contact);

  /* Contacts are ordered in the bucket from the least recently seen
   * to the most recently seen.  The following operations must maintain
   * this ordering.
   */
  
  /* If contact is already in the bucket, move it to the end of the bucket.
   * (First remove the contact from its original place, it will be added
   * again at the end of the function.
   */
  if (oldIndex != -1) {
    vectorRemoveAt(kBucket, oldIndex);
    KADEMLIA_LOG
      " is already in the routing table, mark as newly seen.\n"
    KADEMLIA_GOL
  } else if (vectorSize(kBucket) < routingTable->bucketSize) {
  /* Add automatically to a non-full bucket.
   * (Since newContact already is contact, nothing needs to be
   * done here.
   */
    KADEMLIA_LOG
      " added to routing table.\n"
    KADEMLIA_GOL
  } else {
    DHT_NodeAddress * leastRecentlySeen = 
      (DHT_NodeAddress *)vectorRemoveAt(kBucket, 0);
    /* If the least recently seen node does not respond, discard
     * it and insert the new contact.  If it does respond, it is now the
     * new contact and is inserted at the end of the bucket.
     */
    
    if (dhtPingNode(
          &routingTable->tableId,
          leastRecentlySeen)) {
      KADEMLIA_LOG
          " abandoned in favor of existing contact %s.\n",
          hash160ToString(getNodeAddressHash(leastRecentlySeen))
      KADEMLIA_GOL
      newContact = leastRecentlySeen;
    } else {
      KADEMLIA_LOG
        " replaced an unresponsive existing contact %s.\n",
          hash160ToString(getNodeAddressHash(leastRecentlySeen))
      KADEMLIA_GOL
    }
  }

  if (newContact != NULL) {
    vectorInsertLast(kBucket, newContact);
  } else {
    LOG(LOG_WARNING,
        "Kademlia_newContact: NULL node address given as new contact.\n");
  }
}

int kBucketIndex(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code) {
  HashCode160 d;
  int i;
  xorMetricDistance(getNodeAddressHash(&routingTable->localId), code, &d);
  for (i = KADEMLIA_N_BITS - 1; i >= 0; --i) {
    if (getHashCodeBit(&d, i) != 0)
      return i;
  }
  return 0;
}

int nextClosestKBucketIndex(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code,
    int currentIndex) {
  /* TODO: This code needs testing, the logic is somewhat complex.
   */
  /* A bit inefficient repeating kBucketIndex computation every call.  If this
   * troubles you too much, just make a version of this function which also
   * takes closestIndex as a parameter.
   */
  int closest = kBucketIndex(
      routingTable,
      code);
  int i = currentIndex;

  /* In the buckets below index closest, if the most significant bit different
   * from localId in the bucket's contacts is the same in code, the bucket's
   * contents are closer to code than localId.  Otherwise they are further from
   * code than localId.
   */

  /* Iterating buckets closer than local to target. 
   */
  if (i <= closest &&
      getHashCodeBit(getNodeAddressHash(&routingTable->localId), i)
      != getHashCodeBit(code, i)) {
    --i;
    while (i >= 0) {
      if (getHashCodeBit(getNodeAddressHash(&routingTable->localId), i) !=
          getHashCodeBit(code, i)) {
        return i;
      } else {
        --i;
      }
    }
  } else while (i < KADEMLIA_N_BITS) {
  /* Going upwards
   */
      ++i;
      if (i > closest || getHashCodeBit(
            getNodeAddressHash(&routingTable->localId),
            i) ==
          getHashCodeBit(code, i)) {
        return i < KADEMLIA_N_BITS ? i : -1;
      } else {
        ++i;
      }
    }
    return -1;
}

int getNClosestNodes(
    Kademlia_RoutingTable * routingTable,
    int n,
    HashCode160 * target,
    Vector * buffer) {
  int amount = 0;
  int i;
  DHT_NodeAddress * nextAddress;
  Vector * kBucket;

  
  for (i = kBucketIndex(routingTable, target);
      i != -1 && amount < n;
      i = nextClosestKBucketIndex(routingTable, target, i)) {
    kBucket = routingTable->kBuckets[i];

    vectorLock(kBucket);
    for (nextAddress = (DHT_NodeAddress *)vectorGetFirst(kBucket);
        nextAddress != NULL && amount < n;
        nextAddress = (DHT_NodeAddress *)vectorGetNext(kBucket)) {
      vectorInsertLast(buffer, nextAddress);
      amount++;
    }
    vectorUnlock(kBucket);

    /* Add self. */
    if (i == 0 && amount < n) {
      vectorInsertLast(buffer, &routingTable->localId);
      amount++;
    }
  }
  return amount;
}

int numLocalNodesCloserToCode(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code) {
  int bucketIndex = kBucketIndex(routingTable, code);
  int nodeAmount = 0;
  int i;

  /* All contacts in the code's own bucket are by definition closer to the code
   * than to localId.
   */

  nodeAmount += vectorSize(routingTable->kBuckets[bucketIndex]);

  /* The buckets above bucketIndex have contacts by definition further from
   * code than localId.  The can be ignored.
   */

  /* In the buckets below bucketIndex, if the most significant bit different
   * from localId in the bucket's contacts is the same in code, the bucket's
   * contents are closer to code than localId.  Otherwise they are further from
   * code than localId.
   */

  for (i = 0; i < bucketIndex; ++i) {
    if (getHashCodeBit(getNodeAddressHash(&routingTable->localId), i) !=
        getHashCodeBit(code, i))
      nodeAmount += vectorSize(routingTable->kBuckets[i]);
  }

  return nodeAmount;
}

/****************************** RPC stuff ******************************/

int Kademlia_registerRpcCalls() {
  RPC_register(
      "DHT_ping",
      rpc_DHT_ping);
  RPC_register(
      "DHT_store",
      rpc_DHT_store);
  RPC_register(
      "DHT_findNode",
      rpc_DHT_findNode);
  RPC_register(
      "DHT_findValue",
      rpc_DHT_findValue);
  RPC_register(
      "DHT_listTables",
      rpc_DHT_listTables);
  
  return 0;
}

DHT_DataContainer * makeDataContainerCopy(
    const DHT_DataContainer * source) {
  DHT_DataContainer * result = MALLOC(sizeof(DHT_DataContainer));
  result->dataLength = source->dataLength;
  result->data = MALLOC(result->dataLength);
  memcpy(result->data, source->data, result->dataLength);
  return result;
}

int dhtPingNode(
    DHT_TableId * table,
    DHT_NodeAddress * target) {
  RPC_Param * request = RPC_paramNew();
  RPC_Param * result = RPC_paramNew();

  /* TODO: Make ping return tableId, ping should not only tell whether
     the other node is running DHT but also whether the other node maintains
     the table we are interested in.
   */
  int error = RPC_execute(
      target,
      "DHT_ping",
      request,
      result,
      KADEMLIA_RPC_TIMEOUT);

  int pingStatus = 0;

  if (!error && strcmp(RPC_paramName(result, 0), "DHT_pong") == 0)
    pingStatus = 1;

  if (error) {
    LOG(LOG_WARNING, "WARNING: RPC error in dhtPingNode.");
  }

  RPC_paramFree(request);
  RPC_paramFree(result);
  return pingStatus;
}

/**
 * Sort the values in range from beginIndex to endIndex - 1 in 
 * ascending order by distance from \a target.
 */
void sortVectorRangeByDistance(
    Vector * v,
    HashCode160 * target,
    int beginIndex,
    int endIndex) {
  /*
  KADEMLIA_LOG
    "Sorting vector range [%d, %d] in size %d vector.\n",
    beginIndex,
    endIndex,
    vectorSize(v)
  KADEMLIA_GOL
  */
  if (beginIndex >= endIndex - 1) 
    return;
  /*
  if (beginIndex == endIndex - 2) {
    if (dhtHashCodeCompareDistance(
          getNodeAddressHash(
            (DHT_NodeAddress *)vectorGetAt(v, beginIndex)),
          getNodeAddressHash(
            (DHT_NodeAddress *)vectorGetAt(v, beginIndex + 1)),
          target) > 0) 
      vectorSwap(v, beginIndex, beginIndex + 1);
    return;
  }
  */

  /* Do a quicksort */
  int pos = beginIndex + 1;
  int pivotIndex = beginIndex;
  int end = endIndex - 1;
  int diff;
    
  while (pos < end) {
    diff = dhtHashCodeCompareDistance(
        getNodeAddressHash(
          (DHT_NodeAddress *)vectorGetAt(v, pivotIndex)),
        getNodeAddressHash(
          (DHT_NodeAddress *)vectorGetAt(v, pos)),
        target);
    /* diff =
     * -1  : The node is further than the pivot, move to end
     * 0   : The node is equal to the pivot
     * 1   : The node is closer than the pivot, move to beginning
     */

    if (diff < 0) {
      vectorSwap(v, pos, end);
      --end;
    } else if (diff >= 0) {
      vectorSwap(v, pos, pivotIndex);
      ++pos;
      ++pivotIndex;
    }
  }

  sortVectorRangeByDistance(
      v,
      target,
      beginIndex,
      pivotIndex - 1);

  sortVectorRangeByDistance(
      v,
      target,
      pivotIndex + 1,
      endIndex);
}

void sortVectorByDistance(
    Vector * v,
    HashCode160 * target) {
  sortVectorRangeByDistance(v, target, 0, vectorSize(v));
}

int isNodeInVector(
    DHT_NodeAddress * node,
    Vector * v) {
  DHT_NodeAddress * a;
  vectorLock(v);
  for (a = (DHT_NodeAddress *)vectorGetFirst(v);
      a != NULL;
      a = (DHT_NodeAddress *)vectorGetNext(v)) {
    if (0 == dhtHashCodeCompare(
          getNodeAddressHash(node),
          getNodeAddressHash(a))) {
      vectorUnlock(v);
      return 1;
    }
  }
  vectorUnlock(v);
  return 0;
}

/**
 * Merges Vectors \a original and \newItems into \original to contain \a n
 * DHT_Address pointer values closest to \a target.
 *
 * @return The number of changes made to original vector
 */
int mergeNNodesClosestToTarget(
    int n,
    Vector * originalSorted,
    Vector * newItemsSorted,
    HashCode160 * target) {
  int changes = 0;
  int origIndex = 0;
  int newIndex = 0;
  int diff;
  while (origIndex < n && origIndex < vectorSize(originalSorted) &&
      newIndex < vectorSize(newItemsSorted)) {
    diff = dhtHashCodeCompareDistance(
        getNodeAddressHash(
          (DHT_NodeAddress *)vectorGetAt(originalSorted, origIndex)),
        getNodeAddressHash(
          (DHT_NodeAddress *)vectorGetAt(newItemsSorted, newIndex)),
        target);
    if (diff > 0) {
      vectorInsertAt(
          originalSorted,
          vectorGetAt(newItemsSorted, newIndex),
          origIndex);
      ++newIndex;
      ++changes;
    } else {
      ++origIndex;
    }
  }
  return changes;
}

void generalLookup(
    DHT_TableId * table,
    HashCode160 * key,
    const char * rpcFunc,
    Vector * values,
    Vector * nodes) {

  KADEMLIA_LOG
    "generalLookup: Starting lookup for key %s ",
    hash160ToString(key)
  KADEMLIA_GOL
  KADEMLIA_LOG
    "in table %s, RPC: %s.\n",
    hash160ToString(table),
    rpcFunc
  KADEMLIA_GOL

  Kademlia_RoutingTable * routingTable = Kademlia_tableLookup(table);
  if (!routingTable) {
    LOG(LOG_FAILURE,
        "generalLookup: Tried to look up from an unregistered table.\n");
    return;
  }

  /* No parallel search with alpha param yet. */
  Vector * unqueriedNodes = vectorNew();
  Vector * queriedNodes = vectorNew();
  /* Free the stuff in here when finished. */
  Vector * freeMe = vectorNew();

  RPC_Param * request;
  RPC_Param * result;


  int valueFound = 0;
  int error;
  /* Safety precation, only do a finite number of iterations even if the
   * lookups just keep converging without getting a definite closest set.
   */
  int maxIterations = KADEMLIA_N_BITS;
  int i;
  int numNodesToFind = routingTable->bucketSize;
  DHT_NodeAddress * newContact;
  DHT_NodeAddress * rpcContact;

  /* Need to find bucketSize nodes closest to key.  Start with locally known
   * closest nodes.
   */

  /*
  KADEMLIA_LOG
    "generalLookup: Collecting local nodes.\n"
  KADEMLIA_GOL
  */
  getNClosestNodes(
      routingTable,
      routingTable->bucketSize,
      key,
      unqueriedNodes);

  /*
  KADEMLIA_LOG
    "generalLookup: Starting lookup loop.\n"
  KADEMLIA_GOL
  */

  while (vectorSize(unqueriedNodes) > 0 && maxIterations-- > 0) {
    KADEMLIA_LOG
      "generalLookup: Iterating: Number of unqueried nodes: %d, Number of " 
      "queried nodes: %d\n",
      vectorSize(unqueriedNodes), vectorSize(queriedNodes)
    KADEMLIA_GOL
    request = RPC_paramNew();
    result = RPC_paramNew();

    RPC_paramAdd(request, "sender", sizeof(DHT_NodeAddress), capi->myIdentity);
    RPC_paramAdd(request, "code", sizeof(HashCode160), key);
    RPC_paramAdd(request, "tableId", sizeof(DHT_TableId), table);


    /* Call the first unqueried node and move it to queried nodes' set.
     */
    newContact = vectorRemoveAt(unqueriedNodes, 0);
    /*
    KADEMLIA_LOG
      "unqueriedNodes after removal: %d\n", vectorSize(unqueriedNodes)
    KADEMLIA_GOL
    */

    vectorInsertLast(queriedNodes, newContact);

    /*
    KADEMLIA_LOG
      "Trying to rpc.\n"
    KADEMLIA_GOL
    */

    error = RPC_execute(
        newContact,
        rpcFunc,
        request,
        result,
        KADEMLIA_RPC_TIMEOUT);

    /*
    KADEMLIA_LOG
      "RPC result: %d\n", error
    KADEMLIA_GOL
    */

    if (error) {
      LOG(
          LOG_WARNING,
          "WARNING: RPC error in dhtGeneralLookup calling %s from node %s.\n",
          rpcFunc,
          hash160ToString(
            getNodeAddressHash(
              (DHT_NodeAddress *)vectorGetLast(queriedNodes))));
    
    } else {
      Kademlia_newContact(routingTable, newContact);
      for (i = 0; i < RPC_paramCount(result); ++i) {
        /* Everything that comes from RPC must be freed by the caller (this
         * func).
         */
        vectorInsertLast(
            freeMe,
            RPC_paramValueByPosition(result, i));

        /*
        KADEMLIA_LOG
          "generalLookup: Queued %d to be freed.\n",
          RPC_paramValueByPosition(result, i)
        KADEMLIA_GOL

        KADEMLIA_LOG
          "generalLookup: Processing RPC result param: %s\n",
          RPC_paramName(result, i)
        KADEMLIA_GOL
        */

        if (!strcmp(RPC_paramName(result, i), "nodeId")) {
          rpcContact = 
            (DHT_NodeAddress *)RPC_paramValueByPosition(result, i);

          KADEMLIA_LOG
            "generalLookup: Received node contact %s via RPC.\n",
            hash160ToString(getNodeAddressHash(rpcContact))
          KADEMLIA_GOL

          if (!isNodeInVector(rpcContact, queriedNodes) &&
              !isNodeInVector(rpcContact, unqueriedNodes)) {
            /*
            KADEMLIA_LOG
              "Received node data: %s\n",
              hash160ToString(getNodeAddressHash(rpcContact))
            KADEMLIA_GOL
            */
            vectorInsertLast(unqueriedNodes, rpcContact);
          }
        } else if(!strcmp(RPC_paramName(result, i), "value")) {
          if (values == NULL) {
            LOG(
                LOG_WARNING,
                "WARNING: dhtGeneralLookup received value data "\
                "but was called without a value vector.\n");
          } else {
            valueFound = 1;
            vectorInsertLast(
                values,
                newDataContainer(
                  RPC_paramValueLengthByPosition(result, i),
                  RPC_paramValueByPosition(result, i)));
          }
        }
      }
    }


    RPC_paramFree(result);
    RPC_paramFree(request);

    if (valueFound) {
      /* TODO: cache values in local node.
       */
      break;
    }
  } 

  /* If the caller wanted nodes (Vector \a nodes is not NULL), add the
   * bucketSize closest nodes into the \a nodes Vector.
   */
  if (nodes != NULL) {
    KADEMLIA_LOG 
      "generalLookup: Storing nodes\n"
    KADEMLIA_GOL
    sortVectorByDistance(queriedNodes, key);
    KADEMLIA_LOG
      "Done sorting vector.\n"
    KADEMLIA_GOL
    for (i = 0; 
        i < numNodesToFind && i < vectorSize(queriedNodes); 
        ++i) 
      vectorInsertLast(
          nodes,
          newDataContainer(
            sizeof(DHT_NodeAddress),
            vectorGetAt(queriedNodes, i)));
  }

  KADEMLIA_LOG
    "Kademlia: General lookup done.  Cleaning up.\n"
  KADEMLIA_GOL

  while (vectorSize(freeMe)) {
    void * ptr = vectorRemoveLast(freeMe);
    KADEMLIA_LOG
      "Freeing %d\n", ptr
    KADEMLIA_GOL
    /*
    if (ptr) 
      FREE(ptr);
      */
  }

  vectorFree(freeMe);
  vectorFree(queriedNodes);
  vectorFree(unqueriedNodes);

  KADEMLIA_LOG
    "Kademlia: General lookup cleanup done.\n"
  KADEMLIA_GOL
}

void nodeLookup(
    DHT_TableId * table,
    HashCode160 * key,
    Vector * result) {
  generalLookup(
      table,
      key,
      "DHT_findNode",
      NULL,
      result);
}

void valueLookup(
    DHT_TableId * table,
    HashCode160 * key,
    Vector * result) {

  generalLookup(
      table,
      key,
      "DHT_findValue",
      result,
      NULL);

  /* TODO: put stuff in result.*/
  /* TODO: compare result keyData with parameter keyData. */
}

int localStoreValue(
    DHT_TableId * table,
    HashCode160 * key,
    DHT_DataContainer * value) {
  int cacheTime;
  Kademlia_RoutingTable * routingTable = Kademlia_tableLookup(table);
  int dist = numLocalNodesCloserToCode(routingTable, key);
  
  DHT_LocalStoreUnit * unitToStore;
  unitToStore = new_DHT_LocalStoreUnit(key, value);

  /* Compute cacheTime based on approximated distance from key.
   */
  if (dist <= routingTable->bucketSize) {
    cacheTime = 0;
  } else {
    /* TODO: Calculate approppriate delay if further from code.
     */
    cacheTime = 0;
  }
  
  /* cacheTime 0 means that this node should store the data permanently.
   */
  if (cacheTime == 0) {
    setMyData(unitToStore, 1);
    setRepublishable(unitToStore, 1);
  }

  unitToStore->created = cronTime(NULL);
  unitToStore->timeout = cronTime(NULL);

  if(localStorePut(table, unitToStore) == OK) {
    return OK;
  } else {
    return SYSERR;
  }
}

int storeValueLocal(
    DHT_TableId * tableId,
    HashCode160 * key,
    DHT_DataContainer * value) {
  DHT_LocalStoreUnit * unitToStore;

  unitToStore = new_DHT_LocalStoreUnit(key, value);
  setMyData(unitToStore,1);
  setRepublishable(unitToStore,1);
  unitToStore->created = cronTime(NULL);
  unitToStore->timeout = cronTime(NULL);

  return localStorePut(tableId, unitToStore);
}

int dhtStoreValue(
    DHT_TableId * tableId,
    HashCode160 * key,
    DHT_DataContainer * value) {
  RPC_Param * request;
  RPC_Param * result;
  int stores = 0;
  int error;
  
  Vector * nodesClosestToKey = vectorNew();

  DHT_NodeAddress * node;
  
  valueLookup(tableId, key, nodesClosestToKey);

  for (node = (DHT_NodeAddress *)vectorGetFirst(nodesClosestToKey);
      node != NULL;
      node = (DHT_NodeAddress *)vectorGetNext(nodesClosestToKey)) {
    request = RPC_paramNew();
    result = RPC_paramNew();

    RPC_paramAdd(request, "tableId", sizeof(DHT_TableId), tableId);
    RPC_paramAdd(request, "key", sizeof(HashCode160), key);
    RPC_paramAdd(request, "value", value->dataLength, value->data);

    error = RPC_execute(
        node,
        "DHT_store",
        request,
        result,
        KADEMLIA_RPC_TIMEOUT);

    if (error) {
      LOG(LOG_WARNING, "WARNING: RPC error in dhtStoreValue");
    }
    
    if (!error && result) {
      /* Check store result. */
      if (!strcmp(RPC_paramName(result, 0), "ok")) {
        ++stores;
      } else if (!strcmp(RPC_paramName(result, 0), "error")) {
        KADEMLIA_LOG
            "dhtStoreValue: RPC store of value failed.\n"
        KADEMLIA_GOL
      } else {
        LOG(LOG_WARNING, "dhtStoreValue: Malformed RPC response: unknown " \
          "parameter '$s'\n", RPC_paramName(result, 0));
      }
    }

    RPC_paramFree(result);
    RPC_paramFree(request);
  }

  int n_external_nodes = vectorSize(nodesClosestToKey);
  KADEMLIA_LOG
    "dhtStoreValue: Stored value in %d out of %d remote nodes.\n",
    stores,
    n_external_nodes
  KADEMLIA_GOL

  vectorFree(nodesClosestToKey);
  if (stores > 0 || n_external_nodes == 0) {
    /* Save to network was successful or no network nodes found, store locally
     * as well.
     */
    storeValueLocal(tableId, key, value);
    return OK;
  } else {
     return SYSERR;
  }
}

int dhtListLocalTables(
    DHT_TableSet * tableSet) {

  DHT_LocalTableSet localSet = localStoreListTables();
  DHT_TableDef * tableDef;
  if (localSet.returnCode == OK) {
    tableSet->errorCode = OK; 

    while ((tableDef = nextTable(&localSet.iterator)) != NULL) {
      addToTableSet(tableSet, &tableDef->tableId, &tableDef->metadata);
    }
  }

  return localSet.returnCode;
}

int dhtListRemoteTables(
    DHT_NodeAddress * nodeAddress,
    DHT_TableSet * tableSet) {
  RPC_Param * req = RPC_paramNew();
  RPC_Param * res = RPC_paramNew();

  int error = RPC_execute(
      nodeAddress,
      "DHT_listTables",
      req,
      res,
      KADEMLIA_RPC_TIMEOUT);

  KADEMLIA_LOG
    "Called listTables RPC: %d\n",
    error
  KADEMLIA_GOL

  if (error) {
    LOG(LOG_DEBUG,
        "dhtListRemoteTables() - RPC error getting tables over RPC.\n");
    RPC_paramFree(res);
    RPC_paramFree(req);
    return SYSERR;
  } 

  DHT_TableId tableId;
  DHT_TableMetaData metaData;
  int i;
  for (i = 0; i < RPC_paramCount(res); ++i) {
    if (!strcmp(RPC_paramName(res, i), "table")) {
      KADEMLIA_LOG
        "Deserializing received table data.\n"
      KADEMLIA_GOL
      deserializeTableData(
          &tableId,
          &metaData,
          RPC_paramValueByPosition(res, i),
          RPC_paramValueLengthByPosition(res, i));
      KADEMLIA_LOG
        "Done: table name: %s\n",
        metaData.name
      KADEMLIA_GOL
      addToTableSet(tableSet, &tableId, &metaData);
    } else {
      LOG(LOG_WARNING,
          "dht::tables() - Unknown RPC return parameter: %s\n",
          RPC_paramName(res, i));
    }
  }
  RPC_paramFree(res);
  RPC_paramFree(req);
  return OK;
}

/**
 * Used to query whether a node is running DHT.
 * @param No arguments
 * @param Return value DHT_pong: NULL
 */
void rpc_DHT_ping(
    const RPC_Param * arg,
    RPC_Param * res) {
  RPC_paramAdd(res, "DHT_pong", 0, NULL);
}

/**
 * Store data in node.
 *
 * @param RPC_Param * arg 
 *    Fields: 
 *      tableId: DHT_TableId Table to store data
 *      key: HashCode160 Hash key of data
 *      value: binary data to store in node
 *
 * @param RPC_Param * res
 *    Fields:
 *      ok: (no data) if data successfully stored in node
 *      error: (no data) if data was not stored in node
 */

void rpc_DHT_store(
    const RPC_Param * arg,
    RPC_Param * res) {

  DHT_TableId * table = (DHT_TableId *)RPC_paramValueByName(arg, "tableId");
  HashCode160 * key = (HashCode160 *)RPC_paramValueByName(arg, "key");
  void * value = RPC_paramValueByName(arg, "value");
  int valueLength = RPC_paramValueLengthByName(arg, "value");
  DHT_DataContainer data;

  /* Wait for RPC api change... */

  if (table == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_store: Malformed RPC, parameter 'tableId' node found.\n");
    return;
  }
  if (key == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_store: Malformed RPC, parameter 'key' node found.\n");
    return;
  }
  if (value == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_store: Malformed RPC, parameter 'value' node found.\n");
    return;
  }

  data.data = value;
  data.dataLength = valueLength;

  if (localStoreValue(table, key, &data)) {
    RPC_paramAdd(res, "ok", 0, NULL);
  } else {
    RPC_paramAdd(res, "error", 0, NULL);
  }
}



/**
 * Find k nodes closest to target from the local routing table.
 *
 * @param RPC_Param * arg Target hashcode
 *    Fields: code: HashCode160, tableId: DHT_TableId, 
 *      sender: DHT_NodeAddress
 *  
 * @param RPC_Param * res Result of search. 
 *    Fields: nodeID: DHT_NodeID
 *            nodeID: DHT_NodeID
 *               ....
 *             (up to K nodeIDs)
 */

void rpc_DHT_findNode(
    const RPC_Param * arg,
    RPC_Param * res) {
  KADEMLIA_LOG
    "rpc_DHT_findNode called.\n"
  KADEMLIA_GOL

  HashCode160 * code = (HashCode160 *)RPC_paramValueByName(arg, "code");  
  DHT_TableId * tableId = (DHT_TableId *)RPC_paramValueByName(arg, "tableId");  
  DHT_NodeAddress * sender = (DHT_NodeAddress *)RPC_paramValueByName(
    arg, "sender");

  if (code == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_findNode: Malformed RPC, parameter 'code' not found.\n");
    return;
  }

  if (tableId == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_findNode: Malformed RPC, parameter 'tableId' not found.\n");
    return;
  }

  if (sender == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_findNode: Malformed RPC, parameter 'sender' not found.\n");
    return;
  }

  Kademlia_RoutingTable * table = Kademlia_tableLookup(tableId);

  Kademlia_newContact(table, sender);

  DHT_NodeAddress * node;

  if (table == NULL) {
    LOG(LOG_WARNING,
        "rpc_DHT_findNode: RPC requested unknown tableId");
    return;
  }
  
  Vector * resultNodes = vectorNew();
  getNClosestNodes(
      table,
      table->bucketSize,
      code,
      resultNodes);

  for (node = (DHT_NodeAddress *)vectorGetFirst(resultNodes);
      node != NULL;
      node = (DHT_NodeAddress *)vectorGetNext(resultNodes)) {
    RPC_paramAdd(res, "nodeId", sizeof(DHT_NodeAddress), node);
  }

  vectorFree(resultNodes);
  
}


/**
 * Finds either the value needed or k nodes closer to the value
 *
 * @param RPC_Param * arg Key hashcode
 *    Fields: code: HashCode160, tableId: DHT_TableId
 *      sender: DHT_NodeAddress
 *  
 * @param RPC_Param * res Result of search. 
 *    Fields: nodeID: DHT_NodeID
 *            nodeID: DHT_NodeID
 *               ....
 *             (up to K nodeIDs)
 *      OR    value: DHT_ResultSet  
 *            value: DHT_ResultSet  
 *               ....
 *             (as many values as are found)
 */

void rpc_DHT_findValue(
    const RPC_Param * arg,
    RPC_Param * res) {
  DHT_DataContainer * keyData = RPC_paramValueByName(arg, "code");  
  DHT_DataContainer * tableIdData = RPC_paramValueByName(arg, "tableId");  
  DHT_NodeAddress * sender = (DHT_NodeAddress *)RPC_paramValueByName(
    arg, "sender");

  if (keyData == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_findValue: Malformed RPC, parameter 'code' not found.\n");
    return;
  }

  if (tableIdData == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_findValue: Malformed RPC, parameter 'tableId' not found.\n");
    return;
  }

  if (sender == NULL) {
    LOG(LOG_FAILURE,
        "rpc_DHT_findNode: Malformed RPC, parameter 'sender' not found.\n");
    return;
  }

  HashCode160 * key = (HashCode160 *)keyData->data;

  const Kademlia_RoutingTable * table =
    Kademlia_tableLookup((DHT_TableId *)tableIdData->data);

  Kademlia_newContact(table, sender);

  if (table == NULL) {
    LOG(LOG_WARNING,
        "rpc_DHT_findNode: RPC requested unknown tableId");
    return;
  }

  DHT_LocalStoreResultSet searchResult = localStoreGet(
      (DHT_TableId *)tableIdData->data,
      key);

  LocalStoreResultIter * iter = NULL;
  DHT_LocalStoreUnit * unit = NULL;

  int nResults = 0;

  if (1/*isOkay(searchResult.returnCode)*/) {
    iter = searchResult.iterator;
    while ((unit = (DHT_LocalStoreUnit *)nextUnit(&iter)) != NULL) {
      ++nResults;
      RPC_paramAdd(res, "value", unit->data->dataLength, unit->data->data);
    }
  }
  if (nResults == 0) {
    /* No values here, do a findNode instead */
    rpc_DHT_findNode(arg, res);
  }
}

/**
 * List all tables of this node in result.
 * Each datum in the result is named "table" and is composed of a 20 byte
 * tableID followed by int length of table name followed by table name string.
 * 
 */
void rpc_DHT_listTables(
    const RPC_Param * arg,
    RPC_Param * res) {
  DHT_LocalTableSet tables = localStoreListTables();
  DHT_TableDef * tableDef;
  void * data;
  size_t dataLength;

  tableDef = nextTable(&tables.iterator);
  /*
  while ((tableDef = nextTable(&tables.iterator)) != NULL) {
  */
  while (tableDef != NULL) {
    /*
    KADEMLIA_LOG
      "Serializing table %d: %s\n",
      tableDef,
      hash160ToString(&tableDef->tableId)
    KADEMLIA_GOL
    */

    serializeTableData(
        &tableDef->tableId,
        &tableDef->metadata,
        &data,
        &dataLength);

    RPC_paramAdd(
        res,
        "table",
        dataLength,
        data);
    tableDef = nextTable(&tables.iterator);
  }
}
