 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */

/**
 * @file module/kademlia.h
 * @brief The Kademlia-algorithm for distributed hashing
 * @author Risto Saarelma
 * @version $Id: kademlia.h,v 1.14 2004/05/18 15:10:21 rsaarelm Exp $
 */

#ifndef DHT_KADEMLIA_H
#define DHT_KADEMLIA_H

#include "api_structs.h"
#include "rpc.h"

/************************* *************************/

/************************* Constants *************************/

/**
 * The size K in Kademlia's K-buckets.
 */
#define KADEMLIA_BUCKET_SIZE 20

/**
 * The number of bits in the key hashes and node IDs Kademlia uses.
 */
#define KADEMLIA_N_BITS 160

/**
 * The amount of parallel queries Kademlia will make.
 */
#define KADEMLIA_ALPHA 1

#define KADEMLIA_RPC_TIMEOUT 100

/************************* Structs *************************/

struct Vector;

typedef struct {
  DHT_NodeAddress localId;
  DHT_TableId tableId;
  Vector * kBuckets[KADEMLIA_N_BITS];
  int bucketSize;
} Kademlia_RoutingTable;

/************************* General *************************/

int Kademlia_init(CoreAPIForApplication * coreApi);

/************************* Routing Tables *************************/

/**
 * Return the local routing table corresponding to tableId if one is found.
 */
Kademlia_RoutingTable * Kademlia_tableLookup(
  DHT_TableId * tableId);

/**
 * Register a new table to kademlia.
 * @return pointer to the new routing table
 */
Kademlia_RoutingTable * Kademlia_registerTable(
  DHT_TableId * tableId,
  DHT_NodeAddress * localId,
  int bucketSize);

/**
 * Join a table on another node.
 */

void Kademlia_joinTable(
    DHT_NodeAddress * tableNode,
    DHT_TableId * tableId);

/**
 * Remove all local routing tables.
 */
void Kademlia_clearTables();

void Kademlia_initRoutingTable(
    Kademlia_RoutingTable * routingTable,
    DHT_TableId * tableId,
    DHT_NodeAddress * localId,
    int bucketSize);

/**
 * Free whatver structures have been allocated in the routing table.
 */
void Kademlia_freeRoutingTable(
    Kademlia_RoutingTable * routingTable);

/************************* Node Contacts *************************/

/**
 * Notify a routing table of a node.
 */

void Kademlia_newContact(
    Kademlia_RoutingTable * routingTable,
    DHT_NodeAddress * contact);

/**
 * @return Index of the k-bucket of \a localNode into which \a code belongs.
 */
int kBucketIndex(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code);

/**
 * @return Index of the k-bucket just behind k-bucket \a currentIndex in
 * the list of k-buckets sorted by distance to \a code or -1 if no next index.
 */
int nextClosestKBucketIndex(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code,
    int currentIndex);

/**
 * @return Pointer to node or NULL if node was not in local routing table.
 */
DHT_NodeAddress * findLocalNode(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code);

/**
 * Write at most n local nodes approximately closest to target code in buffer.
 * @Param buffer The Vector into which the found values are appended
 *
 * @return The number of nodes found
 */

int getNClosestNodes(
    Kademlia_RoutingTable * routingTable,
    int n,
    HashCode160 * target,
    Vector * buffer);


/**
 * @return the number of nodes in the routing table which are XOR-closer
 * to \a code than \a localId.
 */
int numLocalNodesCloserToCode(
    Kademlia_RoutingTable * routingTable,
    HashCode160 * code);


/************************* RPC Related Functions *************************/

/**
 * Register Kademlia's rpc functions to the rpc module.
 */
int Kademlia_registerRpcCalls();

void generalLookup(
    DHT_TableId * table,
    HashCode160 * key,
    const char * rpcFunc,
    Vector * values,
    Vector * nodes);

void nodeLookup(
    DHT_TableId * table,
    HashCode160 * key,
    Vector * result);

void valueLookup(
    DHT_TableId * table,
    HashCode160 * key,
    Vector * result);

int dhtStoreValue(
    DHT_TableId * table,
    HashCode160 * key,
    DHT_DataContainer * value);

int dhtPingNode(
    DHT_TableId * table,
    DHT_NodeAddress * target);

int dhtListLocalTables(
    DHT_TableSet * tableSet);

int dhtListRemoteTables(
    DHT_NodeAddress * nodeAddress,
    DHT_TableSet * tableSet);

void rpc_DHT_ping(
    const RPC_Param * arg,
    RPC_Param * res);

void rpc_DHT_store(
    const RPC_Param * arg,
    RPC_Param * res);

void rpc_DHT_findNode(
    const RPC_Param * arg,
    RPC_Param * res);

void rpc_DHT_findValue(
    const RPC_Param * arg,
    RPC_Param * res);

void rpc_DHT_listTables(
    const RPC_Param * arg,
    RPC_Param * res);

#endif /* DHT_KADEMLIA_H */
