/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
*/

/**
 * This file provides the implementation of remote procedure calls over GNUnet.
 *
 * @file module/rpc.c
 * @brief Implementation of RPC's
 * @author Antti Salonen
 * @version $Id: rpc.c,v 1.35 2004/05/18 16:01:46 aksalone Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <connection.h>
#include <gnunet_core.h> /* HostIdentity */
#include <gnunet_util.h>
#include "rpc.h"


#define RPC_DATA_FIELD_SIZE 1000
#define RPC_ACK_COUNT_MAX 500

/**
 * Values for CallInstance.localRole
 */
#define ROLE_CALLER 1
#define ROLE_TARGET 2


/**
 * A REQ message. The data field is used as if it was of variable size, the 
 * exact being given in the header.size member before the message is handed
 * to GNUnet. The dataLength field is technically unnecessary, but it is
 * used for more straight-forward handling.
 */
typedef struct {
  p2p_HEADER header; 
  unsigned short sequenceNumber;
  unsigned long timestamp;
  unsigned long rpcIdentifier;
  unsigned long segmentNumber;
  unsigned long segmentCount;
  unsigned short dataLength;
  char data[RPC_DATA_FIELD_SIZE];
} RPC_P2PMessageReq;

/**
 * A RES message. The data field is used as if it was of variable size, the 
 * exact being given in the header.size member before the message is handed
 * to GNUnet. The dataLength field is technically unnecessary, but it is
 * used for more straight-forward handling.
 *   This message is actually identical to the REQ message, so the only 
 * difference is in the value of the header.requestType field. An alternative
 * would've been to define just a single data message with an additional field.
 */
typedef struct {
  p2p_HEADER header; 
  unsigned short sequenceNumber;
  unsigned long timestamp;
  unsigned long rpcIdentifier;
  unsigned long segmentNumber;
  unsigned long segmentCount;
  unsigned short dataLength;
  char data[RPC_DATA_FIELD_SIZE];
} RPC_P2PMessageRes;

/**
 * An ACK message. The ack field is used as if it was of variable size. 
 */
typedef struct {
  p2p_HEADER header; 
  unsigned short ackCount;
  unsigned short ack[RPC_ACK_COUNT_MAX];
} RPC_P2PMessageAck;

/**
 * A parameter to/from an RPC call. These (and nothing else) are stored in 
 * the Vector of the RPC_Param structure.
 */
typedef struct {
  char *name;
  size_t dataLength;
  void *data;
} Parameter;

/**
 * An RPC registered by the local node.
 */
typedef struct {
  char *name;
  void (*callback) (const RPC_Param *, RPC_Param *);
} RegisteredRPC;
  
/**
 * A per-peer structure.
 */
typedef struct {
  HostIdentity *identity;
  unsigned short sequenceNumber;
  Vector *sendWindow;
  int roundTripTime;
  int sendWindowSize;
  int ackTimerSet;
} Peer;

/**
 * A per-RPC structure.
 */
typedef struct {
  Peer *peer;
  unsigned long rpcIdentifier;
  int localRole;
  Vector *dataIn;
  int dataInCount;
  Vector *dataOut;
  Vector *ackOut;
  Semaphore *signaler; /* Used only on the source node. */
} CallInstance;

/**
 * A send window item. In addition to the message, we what call the message is
 * associated with so that when a RES segment comes in we can remove any
 * REQ segments still in the send window.
 */
typedef struct {
  p2p_HEADER *message;
  CallInstance *call; 
  int acknowledged;
} SendWindowItem;


/**
 * A set of RegisteredRPC structures, one for each RPC registered by the 
 * local node. 
 */
static Vector *registered;

/**
 * A set of CallInstance structures for active rpc calls - Either incoming
 * or outgoing.
 */
static Vector *calls;

/**
 * A set of Peer structures, one for each GNUnet peer (as identified by
 * HostIdentity) known to the RPC module. Peers are added as either RPC's
 * are made to them from the local node, or an RPC call is received from them.
 */
static Vector *peers;

/**
 * A counter whose value is used for identifying the RPC's originating
 * from the local node. The value of the counter is incremented after each
 * RPC and thus its value also tells the number of RPC's originated from the
 * local node.
 */
static unsigned long rpcIdentifier = 0;

static CoreAPIForApplication *coreAPI = NULL;

/**
 * A mutex for synchronous access to all module-wide data structures. This
 * lock must be held by the thread that accesses any module-wide accessable
 * data structures like the Vectors defined above, or the local thread-unsafe
 * functions.
 */
static Mutex rpcLock;



/**
 * Return non-zero if the two host identities are equal.
 */

static int hostIdentityEqual (HostIdentity *ha, HostIdentity *hb) {
  return equalsHashCode160 (&ha->hashPubKey, &hb->hashPubKey);
}


/**
 * Returns a fresh timestamp value for outgoing messages.
 */

static unsigned long newTimestamp (void) {
  return (unsigned long) (cronTime (NULL) % (60 * 60 * 1000));
}


/**
 * Compare two timestamps and return the time difference from a to b in
 * milliseconds.
 */

static unsigned long compareTimestamps (unsigned long a, unsigned long b) {
  if (b >= a)
    return b - a;
  else
    return b + 60 * 60 * 1000 - a;
}


/**
 * Convert a REQ message to a string representation. The return value is a 
 * pointer to a static buffer, so the caller has to deal with synchronization.
 */

static char *reqToString (RPC_P2PMessageReq *reqHBO) {
  static char buf[1024];
  (void) sprintf (buf, "[h.size=%hu,h.reqType=%hu,seq=%hu,timestamp=%lu,"
                  "rpcId=%lu,segNum=%lu,segCount=%lu,dataLength=%hu]",
                  reqHBO->header.size,
                  reqHBO->header.requestType,
                  reqHBO->sequenceNumber,
                  reqHBO->timestamp,
                  reqHBO->rpcIdentifier,
                  reqHBO->segmentNumber,
                  reqHBO->segmentCount,
                  reqHBO->dataLength);
  return buf;
}


/**
 * Converts a RES message to a string representation. The return value is a
 * pointer to a static buffer, so the caller has to deal with synchronization.
 */

static char *resToString (RPC_P2PMessageRes *resHBO) {
  static char buf[1024];
  (void) sprintf (buf, "[h.size=%hu,h.reqType=%hu,seq=%hu,timestamp=%lu,"
                  "rpcId=%lu,segNum=%lu,segCount=%lu,dataLength=%hu]",
                  resHBO->header.size,
                  resHBO->header.requestType,
                  resHBO->sequenceNumber,
                  resHBO->timestamp,
                  resHBO->rpcIdentifier,
                  resHBO->segmentNumber,
                  resHBO->segmentCount,
                  resHBO->dataLength);
  return buf;
}


/**
 * Converts an ACK message to a string representation. The return value is a
 * pointer to a static buffer, so the caller has to deal with synchronization.
 */

static char *ackToString (RPC_P2PMessageAck *ackHBO) {
  static char buf[1024];
  (void) sprintf (buf, "[h.size=%hu,h.reqType=%hu,ackCount=%hu,ack=(...)]\n",
                  ackHBO->header.size,
                  ackHBO->header.requestType,
                  ackHBO->ackCount);
  return buf;
}
                  

/**
 * Try to send a REQ message to a peer. If the peer's send window is full the
 * call returns SYSERR. Otherwise the members "sequenceNumber" and "timestamp"
 * of the structure pointed to by reqHBO are set and the message is given to
 * GNUnet. The structure pointed to by reqHBO is assumed to be in host byte
 * order, and it remains that way. On success the return value is OK.
 */

static int sendReq (CallInstance *call, RPC_P2PMessageReq *reqHBO, 
                    unsigned int importance, unsigned int maxDelay) {
  RPC_P2PMessageReq reqNBO;
  SendWindowItem *item;

  if (vectorSize (call->peer->sendWindow) < call->peer->sendWindowSize) {
    reqHBO->sequenceNumber = call->peer->sequenceNumber++;
    reqHBO->timestamp = newTimestamp();
    item = MALLOC (sizeof (SendWindowItem));
    item->message = (p2p_HEADER *) reqHBO;
    item->call = call;
    item->acknowledged = 0;
    vectorInsertLast (call->peer->sendWindow, item);
    reqNBO.header.size = htons (reqHBO->header.size);
    reqNBO.header.requestType = htons (reqHBO->header.requestType);
    reqNBO.sequenceNumber = htons (reqHBO->sequenceNumber);
    reqNBO.timestamp = htonl (reqHBO->timestamp);
    reqNBO.rpcIdentifier = htonl (reqHBO->rpcIdentifier);
    reqNBO.segmentNumber = htonl (reqHBO->segmentNumber);
    reqNBO.segmentCount = htonl (reqHBO->segmentCount);
    reqNBO.dataLength = htons (reqHBO->dataLength);
    (void) memcpy (reqNBO.data, reqHBO->data, reqHBO->dataLength);
     /* TODO: retransmission timer? */
    coreAPI->sendToNode (call->peer->identity, (p2p_HEADER *) &reqNBO, 
                         importance, maxDelay);
    LOG (LOG_DEBUG, "DEBUG: rpc.c::sendReq() - Sent REQ: %s\n", 
         reqToString (reqHBO));
    return OK;
  } else
    return SYSERR;
}


/**
 * Try to send a RES message to the peer of a call. If the peer's send window
 * is full the call returns SYSERR. Otherwise the members "sequenceNumber" 
 * and "timestamp"  of the structure pointed to by resHBO are set and the 
 * message is given to GNUnet. 
 *   If a message is given to GNUnet and there are ACK's waiting in the 
 * call->ackOut queue, the ACK's are removed from the queue and destroyed.
 *   The structure pointed to by resHBO is assumed to be in host byte 
 * order, and it remains that way. On success the return value is OK.
 */

static int sendRes (CallInstance *call, RPC_P2PMessageRes *resHBO, 
                    unsigned int importance, unsigned int maxDelay) {
  RPC_P2PMessageRes resNBO;
  SendWindowItem *item;

  if (vectorSize (call->peer->sendWindow) < call->peer->sendWindowSize) {
    /* Drop acknowledgements from the per-call queue. */
    if (vectorSize (call->ackOut) > 0) {
      LOG (LOG_DEBUG, "DEBUG: rpc.c::sendRes() - Removing %d acknowledgements "
           "from the per-RPC ackOut queue.\n", vectorSize (call->ackOut));
      while (vectorSize (call->ackOut) > 0)
        FREE (vectorRemoveLast (call->ackOut));
    }
    resHBO->sequenceNumber = call->peer->sequenceNumber++;
    resHBO->timestamp = newTimestamp();
    item = MALLOC (sizeof (SendWindowItem));
    item->message = (p2p_HEADER *) resHBO;
    item->call = call;
    item->acknowledged = 0;
    vectorInsertLast (call->peer->sendWindow, item);
    resNBO.header.size = htons (resHBO->header.size);
    resNBO.header.requestType = htons (resHBO->header.requestType);
    resNBO.sequenceNumber = htons (resHBO->sequenceNumber);
    resNBO.timestamp = htonl (resHBO->timestamp);
    resNBO.rpcIdentifier = htonl (resHBO->rpcIdentifier);
    resNBO.segmentNumber = htonl (resHBO->segmentNumber);
    resNBO.segmentCount = htonl (resHBO->segmentCount);
    resNBO.dataLength = htons (resHBO->dataLength);
    (void) memcpy (resNBO.data, resHBO->data, resHBO->dataLength);
     /* TODO: retransmission timer? */
    coreAPI->sendToNode (call->peer->identity, (p2p_HEADER *) &resNBO, 
                         importance, maxDelay);
    LOG (LOG_DEBUG, "DEBUG: rpc.c::sendRes() - Sent RES: %s\n", 
         resToString (resHBO));
    return OK;
  } else
    return SYSERR;
}


/**
 * Send an ACK message to a peer. Whether GNUnet actually is able to deliver
 * the message is unknown, so the return value is always OK. The exact size
 * of the ACK message is calculated and a copy of the message pointed to by
 * ackHBO is created with the fields in network byte order.
 */

static int sendAck (HostIdentity *receiver, RPC_P2PMessageAck *ackHBO,
                    unsigned int importance, unsigned int maxDelay) {
  RPC_P2PMessageAck ackNBO;
  int i;

  ackHBO->header.size = (void *) ackHBO->ack - (void *) ackHBO + 
    ackHBO->ackCount * sizeof (unsigned short);
  ackNBO.header.size = htons (ackHBO->header.size);
  ackNBO.header.requestType = htons (ackHBO->header.requestType);
  ackNBO.ackCount = htons (ackHBO->ackCount);
  for (i = 0; i < ackHBO->ackCount; i++)
    ackNBO.ack[i] = htons (ackHBO->ack[i]);
  coreAPI->sendToNode (receiver, (p2p_HEADER *) &ackNBO, 5, 1000);
  LOG (LOG_DEBUG, "DEBUG: rpc.c::sendAck() - Sent ACK: %s\n", 
       ackToString (ackHBO));
  return OK;
}


/**
 * Called by cron when the acknowledgement timer for a given peer expires.
 * The ackOut Vectors of CallInstances connected with the Peer are emptied
 * and sent to GNUnet in one or more ACK messages.
 */

static void ackTimerExpire (HostIdentity *target) {
  RPC_P2PMessageAck ack;
  CallInstance *call;
  unsigned short *sequenceNumber;

  MUTEX_LOCK (&rpcLock);

  LOG (LOG_DEBUG, "DEBUG: rpc.c::ackTimerExpire() - CALLED\n");

  ack.header.requestType = RPC_p2p_PROTO_ACK;
  ack.ackCount = 0;
  for (call = vectorGetFirst (calls); call; call = vectorGetNext (calls))
    if (hostIdentityEqual (call->peer->identity, target) &&
        vectorSize (call->ackOut) > 0) {
      while ((sequenceNumber = vectorRemoveAt (call->ackOut, 0)) != NULL) {
        ack.ack[ack.ackCount++] = *sequenceNumber;
        FREE (sequenceNumber);
        if (ack.ackCount == RPC_ACK_COUNT_MAX) {
          (void) sendAck (target, &ack, 5, 1000);
          ack.ackCount = 0;
        }
      }
      call->peer->ackTimerSet = 0;
    }
  if (ack.ackCount > 0) 
    (void) sendAck (target, &ack, 5, 1000);
  
  FREE (target);
  MUTEX_UNLOCK (&rpcLock);
  return;
}


/**
 * Clears any acknowledged items from the Vector sendWindow, which must
 * contain pointers to SendWindowItem.
 */

static void clearSendWindow (Vector *sendWindow) {
  SendWindowItem *item;

  for (item = vectorGetFirst (sendWindow);
       item;
       item = vectorGetNext (sendWindow))
    if (item->acknowledged) {
      /* TODO: This is slightly dumb, as the removal operation again searches
         the whole Vector linearly. Vector iterator should support removal. */
      vectorRemoveObject (sendWindow, item);
      FREE (item->message);
      FREE (item);
      LOG (LOG_DEBUG, "DEBUG: rpc.c::clearSendWindow() - Removed an "
           "acknowledged item from the send window\n");
    }
  return;
}


/**
 * Handle a peer-to-peer message of type RPC_p2p_PROTO_REQ.
 */

static int handleRPCMessageReq (HostIdentity *sender,
                                p2p_HEADER *message) {
  CallInstance *call;
  Peer *peer;
  RPC_P2PMessageReq 
    *req = (RPC_P2PMessageReq *) message,
    *reqHBO;
  unsigned short *ack;
  
  LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageReq() - CALLED\n");
  
  if (ntohs (message->requestType) != RPC_p2p_PROTO_REQ) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageReq() - requestType "
         "mismatch - Expected %d, received %d\n - Dropping message", 
         RPC_p2p_PROTO_REQ, ntohs (message->requestType));
    return SYSERR;
  }
  
  /* Create a local copy of the message with the integer fields in HBO. */
  reqHBO = MALLOC (ntohs (req->header.size));
  reqHBO->header.size = ntohs (req->header.size);
  reqHBO->header.requestType = ntohs (req->header.requestType);
  reqHBO->sequenceNumber = ntohs (req->sequenceNumber);
  reqHBO->timestamp = ntohl (req->timestamp);
  reqHBO->rpcIdentifier = ntohl (req->rpcIdentifier);
  reqHBO->segmentNumber = ntohl (req->segmentNumber);
  reqHBO->segmentCount = ntohl (req->segmentCount);
  reqHBO->dataLength = ntohs (req->dataLength);
  (void) memcpy (reqHBO->data, req->data, reqHBO->dataLength);
  
  MUTEX_LOCK (&rpcLock);

  /* Locate the Peer structure or create a new one if this is the message
     from this peer */
  for (peer = vectorGetFirst (peers); peer; peer = vectorGetNext (peers))
    if (hostIdentityEqual (peer->identity, sender))
      break;
  if (!peer) {
    peer = MALLOC (sizeof (Peer));
    peer->identity = MALLOC (sizeof (HostIdentity));
    (void) memcpy (peer->identity, sender, sizeof (HostIdentity));
    peer->sequenceNumber = 0;
    peer->sendWindow = vectorNew();
    peer->sendWindowSize = 1;
    peer->ackTimerSet = 0;
    vectorInsertLast (peers, peer);
  }
  
  /* Check that the segment number falls within the expected range. */
  if (reqHBO->segmentNumber >= reqHBO->segmentCount) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageReq() - segmentNumber "
         "out of bounds (segmentNumber=%d, segmentCount=%d) - Dropping "
         "message\n", reqHBO->segmentNumber, reqHBO->segmentCount);
    FREE (reqHBO);
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }
  
  /* Locate the CallInstance structure or create a new one if this is the
     first segment of a new incoming call. */
  for (call = vectorGetFirst (calls); call; call = vectorGetNext (calls))
    if (hostIdentityEqual (call->peer->identity, sender) &&
        call->rpcIdentifier == reqHBO->rpcIdentifier &&
        call->localRole == ROLE_TARGET)
      break;
  if (!call) {
    int n;
    call = MALLOC (sizeof (CallInstance));
    call->peer = peer;
    call->localRole = ROLE_TARGET;
    call->rpcIdentifier = reqHBO->rpcIdentifier;
    call->dataIn = vectorNew();
    /* The incoming vector is filled with NULL values for each expected segment
       so that the incoming segments are just put in their respetive slots
       once they arrive. */
    for (n = 0; n < reqHBO->segmentCount; n++)
      vectorInsertLast (call->dataIn, NULL);
    call->dataInCount = 0;
    call->dataOut = vectorNew();
    call->ackOut = vectorNew();
    /* The Semaphore call->signaler is not needed at receiving end. */
    vectorInsertLast (calls, call);
  }
  
  /* Drop the message if the segmentCount field isn't what is expected. */
  if (reqHBO->segmentCount != vectorSize (call->dataIn)) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageReq() - segmentCount "
         "mismatch (segmentCount=%d, vectorSize(dataIn)=%d) - Dropping "
         "message\n", reqHBO->segmentCount, vectorSize (call->dataIn));
    FREE (reqHBO);
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }
  
  /* The received message is placed in the dataIn vector. If it isn't a message
     already received, the counter is increased by one. */
  if (!vectorSetAt (call->dataIn, reqHBO, reqHBO->segmentNumber))
    call->dataInCount++;

  /* Add the sequence number to the ackOut Vector. */
  ack = MALLOC (sizeof (unsigned short));
  *ack = reqHBO->sequenceNumber;
  vectorInsertLast (call->ackOut, ack);
  if (!peer->ackTimerSet) {
    HostIdentity *ackTarget = MALLOC (sizeof (HostIdentity));
    (void) memcpy (ackTarget, sender, sizeof (HostIdentity));
    addCronJob ((CronJob) ackTimerExpire, 50, 0, ackTarget);
  }

  /* If the the call->dataIn vector is "full", we can now execute the RPC
     locally. */
  if (call->dataInCount == vectorSize (call->dataIn)) {
    char 
      *requestPacket,
      *returnPacket,
      *rpcName,
      *paramName;
    size_t returnPacketSize;
    int 
      i,
      n,
      pos,
      segmentNumber, 
      segmentCount;
    unsigned char requestParamCount;
    RPC_Param
      *requestParam = RPC_paramNew(),
      *returnParam = RPC_paramNew();
    unsigned long paramValueLength;
    void *paramValue;
    void (*callback) (const RPC_Param *, RPC_Param *) = NULL;
    RegisteredRPC *rrpc;
    RPC_P2PMessageRes *res;
    
    /* Construct the request packet. */
    requestPacket = MALLOC (vectorSize (call->dataIn) * RPC_DATA_FIELD_SIZE);
    pos = 0;
    while ((req = vectorRemoveAt (call->dataIn, 0)) != NULL) {
      (void) memcpy (requestPacket + pos, req->data, req->dataLength);
      pos += req->dataLength;
      FREE (req);
    }
    LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageReq() - Constructed "
         "request packet (size=%d)\n", pos);
    
    /* Scan the request packet and create the RPC_Param structure for request
       parameters. In the parameter structure, both the names of the parameters
       and their values are pointers to memory allocated for the request
       packet. */
    /* TODO: This can segfault if the packet is maligned. */

    rpcName = requestPacket;
    for (pos = 0; *(requestPacket + pos); pos++)
      ;
    pos++;

    requestParamCount = (unsigned char) *(requestPacket + pos);
    pos++;
    for (n = 0; n < requestParamCount; n++) {
      paramName = requestPacket + pos;
      for (; *(requestPacket + pos); pos++)
        ;
      pos++;
      paramValueLength = ntohl (*((unsigned long *) (requestPacket + pos)));
      pos += sizeof (unsigned long);
      paramValue = requestPacket + pos;
      pos += paramValueLength;
      RPC_paramAdd (requestParam, paramName, (size_t) paramValueLength, 
                    paramValue);
    }
    FREE (requestPacket);
    
    /* Check if the RPC is registered and pick up the callback pointer. */
    for (rrpc = vectorGetFirst (registered);
         rrpc;
         rrpc = vectorGetNext (registered))
      if (!strcmp (rrpc->name, rpcName)) {
        callback = rrpc->callback;
        break;
      }

    /* TODO: Use a different thread when calling the possibly blocking
       callback? */
    if (callback) {
      MUTEX_UNLOCK (&rpcLock);
      callback (requestParam, returnParam);
      MUTEX_LOCK (&rpcLock);
      RPC_paramFree (requestParam);
      RPC_paramAdd (returnParam, "_STATUS", 2, "0");
    } else 
      RPC_paramAdd (returnParam, "_STATUS", 2, "4");
    
    /* Regardless of whether the callback was run, we now serialize the return
       parameter structure into a return packet. */
    returnPacketSize = sizeof (unsigned char);
    returnPacket = MALLOC (returnPacketSize);
    pos = 0;
    /* TODO: what if greater than UCHAR_MAX? */
    *(returnPacket + pos++) = (unsigned char) RPC_paramCount (returnParam);
    /* This loop is rather inefficient as the returnPacket is reallocated on
       every return parameter. */
    for (i = 0; i < RPC_paramCount (returnParam); i++) {
      paramName = RPC_paramName (returnParam, i);
      paramValue = RPC_paramValueByPosition (returnParam, i);
      paramValueLength = RPC_paramValueLengthByPosition (returnParam, i);
      GROW (returnPacket, returnPacketSize, 
            returnPacketSize + strlen (paramName) + 1 + 
            sizeof (unsigned long) + paramValueLength);
      (void) strcpy (returnPacket + pos, paramName);
      pos += strlen (paramName) + 1;
      *((unsigned long *) (returnPacket + pos)) = htonl (paramValueLength);
      pos += sizeof (unsigned long);
      (void) memcpy (returnPacket + pos, paramValue, paramValueLength);
      pos += paramValueLength;
    }
    LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageReq() - Constructed return "
         "packet (size=%d)\n", pos);
           
    RPC_paramFree (returnParam);
    
    /* Now that the return packet is ready, we chop it into messages and put
       the messages into the per-call output queue. */
    segmentCount = returnPacketSize / RPC_DATA_FIELD_SIZE;
    if (returnPacketSize % RPC_DATA_FIELD_SIZE > 0)
      segmentCount++;
    segmentNumber = 0;
    pos = 0;
    while (pos < returnPacketSize) {
      res = MALLOC (sizeof (RPC_P2PMessageRes));
      res->header.requestType = RPC_p2p_PROTO_RES;
      /* res->sequenceNumber and res->timestamp are not set yet */
      res->rpcIdentifier = call->rpcIdentifier;
      res->segmentNumber = segmentNumber++;
      res->segmentCount = segmentCount;
      if (returnPacketSize - pos > RPC_DATA_FIELD_SIZE)
        res->dataLength = RPC_DATA_FIELD_SIZE;
      else
        res->dataLength = returnPacketSize - pos;
      (void) memcpy (res->data, returnPacket + pos, res->dataLength);
      res->header.size = res->data - (char *) res + res->dataLength;
      pos += res->dataLength;
      vectorInsertLast (call->dataOut, res);
    }
    FREE (returnPacket);

    /* While there's room in the send window, move segments to GNUnet. 
       There may or may not be acknowledgements in the call->ackOut queue,
       as it's possible that they were transmitted while this thread released
       the Mutex when calling the RPC callback. If there are acknowledgements
       queued, they are sent by sendRes(). */
    while (vectorSize (call->dataOut) > 0 &&
           vectorSize (peer->sendWindow) < peer->sendWindowSize) {
      res = vectorRemoveAt (call->dataOut, 0);
      if (sendRes (call, res, 5, 1000) == SYSERR) 
        LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageReq() - sendRes() "
             "somehow returned SYSERR\n");
    }
  }    
    
  MUTEX_UNLOCK (&rpcLock);
  return OK;
}


/**
 * Handle a peer-to-peer message of type RPC_p2p_PROTO_RES.
 */

static int handleRPCMessageRes (HostIdentity *sender,
                                p2p_HEADER *message) {
  RPC_P2PMessageRes 
    *res = (RPC_P2PMessageRes *) message,
    *resHBO;
  unsigned short *ack;
  Peer *peer;
  CallInstance *call;
  int n;

  LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageRes() - CALLED\n");

  if (ntohs (message->requestType) != RPC_p2p_PROTO_RES) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageRes() - requestType "
         "mismatch - Expected %d, received %d\n", RPC_p2p_PROTO_RES,
         ntohs (message->requestType));
    return SYSERR;
  }

  /* Create a local copy of the message with the integer fields in HBO. */
  resHBO = MALLOC (ntohs (res->header.size));
  resHBO->header.size = ntohs (res->header.size);
  resHBO->header.requestType = ntohs (res->header.requestType);
  resHBO->sequenceNumber = ntohs (res->sequenceNumber);
  resHBO->timestamp = ntohl (res->timestamp);
  resHBO->rpcIdentifier = ntohl (res->rpcIdentifier);
  resHBO->segmentNumber = ntohl (res->segmentNumber);
  resHBO->segmentCount = ntohl (res->segmentCount);
  resHBO->dataLength = ntohs (res->dataLength);
  (void) memcpy (resHBO->data, res->data, resHBO->dataLength);

  MUTEX_LOCK (&rpcLock);

  /* Locate the Peer structure. If it is not found, something is seriously
     wrong, as it should've been created when the RPC was initiated from this
     node. TODO: In this case:
       - Drop the message without acknowledgement? 
       - Something else? */
  for (peer = vectorGetFirst (peers); peer; peer = vectorGetNext (peers))
    if (hostIdentityEqual (peer->identity, sender))
      break;
  if (!peer) { 
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageRes() - Received a RES "
         "message (sender=%s, rpdIdentifier=%d) from a peer unknown to the "
         "RPC module - Dropping message\n", "(sender HostIdentity)", 
         resHBO->rpcIdentifier);
    FREE (resHBO);
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }

  /* Locate the CallInstance structure. If it is not found, something is 
     again seriously wrong. It should've been created when the call was 
     initiated from this node. Drop message OK? */
  for (call = vectorGetFirst (calls); call; call = vectorGetNext (calls))
    if (hostIdentityEqual (call->peer->identity, sender) &&
        call->rpcIdentifier == resHBO->rpcIdentifier &&
        call->localRole == ROLE_CALLER) 
      break;
  if (!call) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageRes() - Received a RES "
         "message (sender=%s, rpdIdentifier=%d) for a call unknown to the "
         "RPC module - Dropping message\n", "(sender HostIdentity)", 
         resHBO->rpcIdentifier);
    FREE (resHBO);
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }

  /* Check that the segment number falls within the expected range. */
  if (resHBO->segmentNumber >= resHBO->segmentCount) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageRes() - segmentNumber "
         "out of bounds (segmentNumber=%d, segmentCount=%d) - Dropping "
         "message\n", resHBO->segmentNumber, resHBO->segmentCount);
    FREE (resHBO);
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }
  
  /* If the dataIn segment is empty, this is the first RES segment and we now
     fill the dataIn vector with NULL values for every expected message. */
  if (vectorSize (call->dataIn) == 0) {
    SendWindowItem *item;
    for (n = 0; n < resHBO->segmentCount; n++)
      vectorInsertLast (call->dataIn, NULL);
    /* Scan the send window to the peer and mark any messages associated
       with this RPC in the send window. They should all be REQ messages
       still waiting acknowledgement. */
    for (item = vectorGetFirst (peer->sendWindow);
         item;
         item = vectorGetNext (peer->sendWindow))
      if (item->call == call) {
        item->acknowledged = 1;
        LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageRes() - Implicitly "
             "marked a message in the send window acknowledged\n");
      }
    clearSendWindow (peer->sendWindow);
  }

  /* Check that the segmentCount field agrees with what is expected. */
  if (resHBO->segmentCount != vectorSize (call->dataIn)) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageRes() - segmentCount "
         "mismatch (segmentCount=%d, vectorSize(dataIn)=%d) - Dropping "
         "message\n", resHBO->segmentCount, vectorSize (call->dataIn));
    FREE (resHBO);
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }

  /* The received message is placed in the dataIn vector. If it isn't a 
     message already received, the counter is increased by one. */
  if (!vectorSetAt (call->dataIn, resHBO, resHBO->segmentNumber))
    call->dataInCount++;

  /* Add the sequence number to the ackOut Vector. */
  ack = MALLOC (sizeof (unsigned short));
  *ack = resHBO->sequenceNumber;
  vectorInsertLast (call->ackOut, ack);
  if (!peer->ackTimerSet) {
    HostIdentity *ackTarget = MALLOC (sizeof (HostIdentity));
    (void) memcpy (ackTarget, sender, sizeof (HostIdentity));
    addCronJob ((CronJob) ackTimerExpire, 50, 0, ackTarget);
  }

  /* If the dataIn vector is "full", signal the thread which SHOULD be waiting
     in the blocked call to function RPC_execute(). */
  if (call->dataInCount == vectorSize (call->dataIn))
    SEMAPHORE_UP (call->signaler);

  MUTEX_UNLOCK (&rpcLock);
  return OK;
}


/**
 * Handle a peer-to-peer message of type RPC_p2p_PROTO_ACK.
 */

static int handleRPCMessageAck (HostIdentity *sender,
                                p2p_HEADER *message) {
  RPC_P2PMessageAck *ack = (RPC_P2PMessageAck *) message;
  Peer *peer;
  CallInstance *call;
  SendWindowItem *item;
  int i;
  
  LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageAck() - CALLED\n");

  if (ntohs (message->requestType) != RPC_p2p_PROTO_ACK) {
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageAck() - requestType "
         "mismatch - Expected %d, received %d\n", RPC_p2p_PROTO_ACK,
         ntohs (message->requestType));
    return SYSERR;
  }

  ack->header.size = ntohs (ack->header.size);
  ack->header.requestType = ntohs (ack->header.requestType);
  ack->ackCount = ntohs (ack->ackCount);
  for (i = 0; i < ack->ackCount; i++)
    ack->ack[i] = ntohs (ack->ack[i]);

  MUTEX_LOCK (&rpcLock);

  /* Locate the Peer structure. If it is not found, something is seriously
     wrong, as it should've been created when the RPC was initiated from this
     node. TODO: In this case:
       - Drop the message without acknowledgement? 
       - Something else? */
  for (peer = vectorGetFirst (peers); peer; peer = vectorGetNext (peers))
    if (hostIdentityEqual (peer->identity, sender))
      break;
  if (!peer) { 
    LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageAck() - Received an "
         "ACK message (sender=%s) from a peer unknown to the RPC module - "
         "Dropping message\n", "(sender HostIdentity)");
    MUTEX_UNLOCK (&rpcLock);
    return SYSERR;
  }

  /* Scan the acknowledgement message content and mark any corresponding
     messages found in the send window acknowledged. */
  for (i = 0; i < ack->ackCount; i++)
    for (item = vectorGetFirst (peer->sendWindow);
         item;
         item = vectorGetNext (peer->sendWindow))
      if (item->message->requestType == RPC_p2p_PROTO_REQ) {
        if (((RPC_P2PMessageReq *) item->message)->sequenceNumber == 
            ack->ack[i]) {
          item->acknowledged = 1;
          LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageAck() - Marked a "
               "REQ message (seq=%d, age=%dms) in the send window "
               "acknowledged\n", ack->ack[i], 
               compareTimestamps 
               (((RPC_P2PMessageReq *) item->message)->timestamp,
                newTimestamp()));
          break;
        }
      } else if (item->message->requestType == RPC_p2p_PROTO_RES)
        if (((RPC_P2PMessageReq *) item->message)->sequenceNumber == 
            ack->ack[i]) {
          item->acknowledged = 1;
          LOG (LOG_DEBUG, "DEBUG: rpc.c::handleRPCMessageAck() - Marked a "
               "RES message (seq=%d, age=%dms) in the send window "
               "acknowledged\n", ack->ack[i],
               compareTimestamps 
               (((RPC_P2PMessageRes *) item->message)->timestamp,
                newTimestamp()));
          break;
        }
  
  /* Remove acknowledged messages from the send window. */
  clearSendWindow (peer->sendWindow);
  
  /* While there's room in the send window, move messages to GNUnet. */
  /* Currently we just move any waiting message from the calls Vector -
     Round-robin is not implemented. TODO: Fix. */
  while (vectorSize (peer->sendWindow) < peer->sendWindowSize) {
    p2p_HEADER *sendWindowMessage;

    for (call = vectorGetFirst (calls);
         call;
         call = vectorGetNext (calls)) 
      if (vectorSize (call->dataOut) > 0)
        break;
    if (!call)
      break; /* Nothing to send. */
    /* The picked up message may be either REQ or RES. */
    sendWindowMessage = vectorRemoveAt (call->dataOut, 0);
    if (sendWindowMessage->requestType == RPC_p2p_PROTO_REQ) {
      if (sendReq (call, (RPC_P2PMessageReq *) sendWindowMessage, 5, 1000) == 
          SYSERR) 
        LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageAck() - sendReq() "
             "somehow returned SYSERR\n");
    } else if (sendWindowMessage->requestType == RPC_p2p_PROTO_RES)
      if (sendRes (call, (RPC_P2PMessageRes *) sendWindowMessage, 5, 1000) == 
          SYSERR) 
        LOG (LOG_WARNING, "WARNING: rpc.c::handleRPCMessageAck() - sendRes() "
             "somehow returned SYSERR\n");
  }
  
  MUTEX_UNLOCK (&rpcLock);
  return OK;
}


/**
 * Allocate a new, empty RPC parameter structure.
 *
 * @return An empty RPC_Param structure
 */

RPC_Param *RPC_paramNew (void) {
  RPC_Param *rvalue;
  rvalue = MALLOC (sizeof (RPC_Param));
  rvalue->vector = vectorNew();
  return rvalue;
}


/**
 * Free the memory used by an RPC parameter structure. All parameter names and
 * values residing in the structure are freed, and thus all pointers returned
 * by this abstractions become dangling.
 * 
 * @param param The RPC parameter structure to be freed
 */

void RPC_paramFree (RPC_Param *param) {
  while (vectorSize (param->vector) > 0) {
    Parameter *p = vectorRemoveLast (param->vector);
    FREE (p->name);
    FREE (p->data);
    FREE (p);
  }
  vectorFree (param->vector);
  FREE (param);
}


/**
 * Return the number of parameters in an RPC parameter structure.
 *
 * @param param Target RPC parameter structure
 * @return The number of parameters
 */

size_t RPC_paramCount (const RPC_Param *param) {
  return vectorSize (param->vector);
}


/**
 * Add a new parameter to the RPC parameter structure. The parameter name and
 * value are copied to memory private to the RPC parameter collection. The
 * pointers returned by other functions point to this private memory and should
 * not be freed by the user of the abstraction.
 *
 * @param param Target RPC parameter structure
 * @param name Name of the parameter
 * @param dataLength Length of the value of the parameter
 * @param data Value of the parameter
 */

void RPC_paramAdd (const RPC_Param *param, 
                   const char *name,
                   size_t dataLength,
                   const void *data) {
  Parameter *new;
  new = MALLOC (sizeof (Parameter));
  new->name = MALLOC (strlen (name) + 1);
  (void) strcpy (new->name, name);
  new->dataLength = dataLength;
  new->data = MALLOC (dataLength);
  (void) memcpy (new->data, data, dataLength);
  vectorInsertLast (param->vector, new);
  return;
}


/**
 * Return the name of the given parameter in the RPC parameter structure, the
 * first parameter being parameter number zero.
 *
 * @param param Target RPC parameter structure
 * @return Name of the parameter
 */

char *RPC_paramName (const RPC_Param *param,
                     int i) {
  Parameter *p = vectorGetAt (param->vector, i);
  if (p)
    return p->name;
  else
    return NULL;
}


/**
 * Return the value of the named parameter in the RPC parameter structure.
 *
 * @param param Target RPC parameter structure
 * @return Value of the named parameter
 */

void *RPC_paramValueByName (const RPC_Param *param,
                            const char *name) {
  Parameter *p;
  for (p = vectorGetFirst (param->vector);
       p; 
       p = vectorGetNext (param->vector)) 
    if (!strcmp (p->name, name))
      return p->data;
  return NULL;
}

/**
 * Return the length of the value of the named parameter in the RPC parameter
 * structure.
 *
 * @param param Target RPC parameter structure
 * @return Value of the named parameter
 */

size_t RPC_paramValueLengthByName (const RPC_Param *param,
                                   const char *name) {
  Parameter *p;
  for (p = vectorGetFirst (param->vector);
       p; 
       p = vectorGetNext (param->vector)) 
    if (!strcmp (p->name, name))
      return p->dataLength;
  return 0;
}

/**
 * Return the value of the given parameter in the RPC parameter structure.
 *
 * @param param Target RPC parameter structure
 * @return Value of the parameter
 */

void *RPC_paramValueByPosition (const RPC_Param *param, int i) {
  Parameter *p = vectorGetAt (param->vector, i);
  if (p) 
    return p->data;
  else
    return NULL;
}


/**
 * Return the length of the value of the given parameter in the RPC parameter 
 * structure.
 *
 * @param param Target RPC parameter structure
 * @return Value of the parameter
 */

size_t RPC_paramValueLengthByPosition (const RPC_Param *param, int i) {
  Parameter *p = vectorGetAt (param->vector, i);
  if (p) 
    return p->dataLength;
  else
    return 0;
}


/**
 * Initialize the RPC module. Nothing else is guaranteed to work prior to this.
 */
int RPC_init (CoreAPIForApplication *capi) {
  int rvalue = OK;
  
  MUTEX_CREATE (&rpcLock);

  coreAPI = capi;

  peers = vectorNew();
  calls = vectorNew();
  registered = vectorNew();

  if (capi->registerHandler (RPC_p2p_PROTO_REQ, &handleRPCMessageReq) ==
      SYSERR) {
    rvalue = SYSERR;
    LOG (LOG_WARNING, "WARNING: rpc.c::RPC_init() - Failed to register a "
         "handler for message type %d\n", RPC_p2p_PROTO_REQ);
  } else
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_init() - Registered a handler for "
         "message type %d\n", RPC_p2p_PROTO_REQ);

  if (capi->registerHandler (RPC_p2p_PROTO_RES, &handleRPCMessageRes) ==
      SYSERR) {
    rvalue = SYSERR;
    LOG (LOG_WARNING, "WARNING: rpc.c::RPC_init() - Failed to register a "
         "handler for message type %d\n", RPC_p2p_PROTO_RES);
  } else
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_init() - Registered a handler for "
         "message type %d\n", RPC_p2p_PROTO_RES);
  
  if (capi->registerHandler (RPC_p2p_PROTO_ACK, &handleRPCMessageAck) ==
      SYSERR) {
    rvalue = SYSERR;
    LOG (LOG_WARNING, "WARNING: rpc.c::RPC_init() - Failed to register a "
         "handler for message type %d\n", RPC_p2p_PROTO_ACK);
  } else
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_init() - Registered a handler for "
         "message type %d\n", RPC_p2p_PROTO_ACK);
  
  if (rvalue == SYSERR)
    LOG (LOG_WARNING, "WARNING: rpc.c::RPC_init() - RPC_init() fails");
  return rvalue;
}


void RPC_done (CoreAPIForApplication *capi) {
  if (capi->isHandlerRegistered (RPC_p2p_PROTO_REQ)) {
    capi->unregisterHandler (RPC_p2p_PROTO_REQ, &handleRPCMessageReq);
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_done() - Unregistered a handler for "
         "message type %d\n", RPC_p2p_PROTO_REQ);
  }
  if (capi->isHandlerRegistered (RPC_p2p_PROTO_RES)) {
    capi->unregisterHandler (RPC_p2p_PROTO_RES, &handleRPCMessageRes);
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_done() - Unregistered a handler for "
         "message type %d\n", RPC_p2p_PROTO_RES);
  }
  if (capi->isHandlerRegistered (RPC_p2p_PROTO_ACK)) {
    capi->unregisterHandler (RPC_p2p_PROTO_ACK, &handleRPCMessageAck);
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_done() - Unregistered a handler for "
         "message type %d\n", RPC_p2p_PROTO_ACK);
  }
  if (peers) {
    while (vectorSize (peers) > 0)
      FREE (vectorRemoveLast (peers));
    vectorFree (peers);
    peers = NULL;
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_done() - Freed per-peer vector\n");
  }
  if (calls) {
    while (vectorSize (calls) > 0) 
      FREE (vectorRemoveLast (calls));
    vectorFree (calls);
    calls = NULL;
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_done() - Freed per-call vector\n");
  }
  if (registered) {
    while (vectorSize (registered) > 0)
      FREE (vectorRemoveLast (registered));
    vectorFree (registered);
    registered = NULL;
    LOG (LOG_DEBUG, 
         "DEBUG: rpc.c::RPC_done() - Freed per-registered vector\n");
  }
  LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_done() - Returning\n");
  coreAPI = NULL;
  MUTEX_DESTROY (&rpcLock);
}



/**
 * Executes an RPC on a node, which may be the local node. The function
 * performs the following steps:
 *   - Constructs a request packet from the request parameters
 *   - Splits the request packet into one or more REQ messages
 *   - While there's room in the send window, moves REQ messages into it
 *   - Waits on a signaling semaphore until the result is ready
 *   - Constructs a return packet from the received RES messages
 *   - Extracts the return parameters from the return packet
 */

int RPC_execute (const HostIdentity *receiver, 
                 const char *name,
                 const RPC_Param *requestParam,
                 RPC_Param *returnParam,
                 int timeout) {

  char 
    *requestPacket,
    *returnPacket,
    *paramName;
  void *paramValue;
  unsigned long paramValueLength;
  Peer *peer;
  CallInstance *call;
  int 
    i,
    n,
    pos,
    returnParamCount,
    segmentCount,
    segmentNumber;
  size_t requestPacketSize;
  RPC_P2PMessageReq *req;
  RPC_P2PMessageRes *res;

  MUTEX_LOCK (&rpcLock);

  /* Special handling if the destination node is the local node. */
  if (hostIdentityEqual ((HostIdentity *) receiver, coreAPI->myIdentity)) {
    RegisteredRPC *rrpc;
    void (*callback) (const RPC_Param *, RPC_Param *) = NULL;
    
    /* TODO: Concurrency solving - Below the function pointer is copied aside 
       while holding the lock, after which the lock is freed and the callback
       is invoked. Is this correct? */
    for (rrpc = vectorGetFirst (registered);
         rrpc;
         rrpc = vectorGetNext (registered))
      if (!strcmp (rrpc->name, name))
        callback = rrpc->callback;

    MUTEX_UNLOCK (&rpcLock);

    if (callback) {
      LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_execute() - Executing RPC \"%s\" "
           "on local node\n", name);
      (*callback) (requestParam, returnParam);
      return RPC_RESULT_SUCCESS;
    } else {
      LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_execute() - Failed to execute RPC "
           "\"%s\" on local node - Not registered", name);
      return RPC_RESULT_RPC_UNREGISTERED;
    }
  }

  /* Locate the Peer structure or create a new one if the peer is formerly
     unknown to the RPC module. */
  for (peer = vectorGetFirst (peers); peer; peer = vectorGetNext (peers))
    if (hostIdentityEqual (peer->identity, (HostIdentity *) receiver))
      break;
  if (!peer) {
    peer = MALLOC (sizeof (Peer));
    peer->identity = MALLOC (sizeof (HostIdentity));
    (void) memcpy (peer->identity, receiver, sizeof (HostIdentity));
    peer->sequenceNumber = 0;
    peer->sendWindow = vectorNew();
    peer->sendWindowSize = 1;
    peer->ackTimerSet = 0;
    vectorInsertLast (peers, peer);
  }

  /* Allocate a new CallInstance structure. */
  call = MALLOC (sizeof (CallInstance));
  call->peer = peer;
  call->localRole = ROLE_CALLER;
  call->rpcIdentifier = rpcIdentifier++;
  call->dataIn = vectorNew();
  call->dataInCount = 0;
  call->dataOut = vectorNew();
  call->ackOut = vectorNew();
  call->signaler = SEMAPHORE_NEW (0);
  vectorInsertLast (calls, call);

  /* Serialize the request parameter structure into a request packet. */
  requestPacketSize = strlen (name) + 1 + sizeof (unsigned char);
  requestPacket = MALLOC (requestPacketSize);
  pos = 0;
  (void) strcpy (requestPacket + pos, name);
  pos += strlen (name) + 1;
  /* TODO: what if greater than UCHAR_MAX? */
  *(requestPacket + pos++) = (unsigned char) RPC_paramCount (requestParam);
  /* This loop is rather inefficient as the requestPacket is reallocated on
     every return parameter. */
  for (i = 0; i < RPC_paramCount (requestParam); i++) {
    paramName = RPC_paramName (requestParam, i);
    paramValue = RPC_paramValueByPosition (requestParam, i);
    paramValueLength = RPC_paramValueLengthByPosition (requestParam, i);
    GROW (requestPacket, requestPacketSize, 
          requestPacketSize + strlen (paramName) + 1 + 
          sizeof (unsigned long) + paramValueLength);
    (void) strcpy (requestPacket + pos, paramName);
    pos += strlen (paramName) + 1;
    *((unsigned long *) (requestPacket + pos)) = htonl (paramValueLength);
    pos += sizeof (unsigned long);
    (void) memcpy (requestPacket + pos, paramValue, paramValueLength);
    pos += paramValueLength;
  }
  LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_execute() - Constructed request packet "
       "(size=%d)\n", pos);

  /* Now that the request packet is ready, we chop it into messages and put
     the messages into the per-call output queue. */
  segmentCount = requestPacketSize / RPC_DATA_FIELD_SIZE;
  if (requestPacketSize % RPC_DATA_FIELD_SIZE > 0)
    segmentCount++;
  segmentNumber = 0;
  pos = 0;
  while (pos < requestPacketSize) {
    req = MALLOC (sizeof (RPC_P2PMessageReq));
    req->header.requestType = RPC_p2p_PROTO_REQ;
    /* req->sequenceNumber and req->timestamp are not set yet */
    req->rpcIdentifier = call->rpcIdentifier;
    req->segmentNumber = segmentNumber++;
    req->segmentCount = segmentCount;
    if (requestPacketSize - pos > RPC_DATA_FIELD_SIZE)
      req->dataLength = RPC_DATA_FIELD_SIZE;
    else
      req->dataLength = requestPacketSize - pos;
    (void) memcpy (req->data, requestPacket + pos, req->dataLength);
    req->header.size = req->data - (char *) req + req->dataLength;
    pos += req->dataLength;
    vectorInsertLast (call->dataOut, req);
  }
  FREE (requestPacket);

  /* While there's room in the send window, move messages to GNUnet. */
  while (vectorSize (call->dataOut) > 0 &&
         vectorSize (peer->sendWindow) < peer->sendWindowSize) {
    req = vectorRemoveAt (call->dataOut, 0);
    if (sendReq (call, req, 5, 1000) == SYSERR) 
      LOG (LOG_WARNING, "WARNING: rpc.c::RPC_execute() - sendReq() somehow "
           "returned SYSERR\n");
  }
  
  /* TODO: All locks must be freed before blocking on the semaphore. */
  /* Now we have sent as many messages as we can, and the rest will be sent
     by other threads as acknowledgements come in. This thread will now wait
     on the signaling semaphore until the result is ready in the dataIn 
     vector. */
  MUTEX_UNLOCK (&rpcLock);
  LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_execute() - Calling SEMAPHORE_DOWN\n");
  /* TODO: Implement support for timeout */
  SEMAPHORE_DOWN (call->signaler);
  MUTEX_LOCK (&rpcLock);

  /* It's possible that either no RES segments have been received or some have
     been received but more are still needed to construct a complete result. */
  if (call->dataInCount == 0) {
    LOG (LOG_WARNING, "WARNING: rpc.c::RPC_execute() - No RES segments "
         "received, dataIn vector empty - Unable to return result.\n");
    MUTEX_UNLOCK (&rpcLock);
    return RPC_RESULT_NETWORKING_PROBLEM;
  } else if (call->dataInCount < vectorSize (call->dataIn)) {
    LOG (LOG_WARNING, "WARNING: rpc.c::RPC_execute() - dataIn vector not full "
         "when waken up (%d messages received, %d expected). Unable to return "
         "result.\n", call->dataInCount, vectorSize (call->dataIn));
    MUTEX_UNLOCK (&rpcLock);
    return RPC_RESULT_NETWORKING_PROBLEM;
  }

  returnPacket = MALLOC (vectorSize (call->dataIn) * RPC_DATA_FIELD_SIZE);
  pos = 0;
  while ((res = vectorRemoveAt (call->dataIn, 0)) != NULL) {
    (void) memcpy (returnPacket + pos, res->data, res->dataLength);
    pos += res->dataLength;
    FREE (res);
  }
  LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_execute() - Constructed return "
       "packet (size=%d)\n", pos);

  /* Scan the return packet and create the RPC_Param structure for return
     parameters. In the parameter structure, the names of the parameters as 
     well as their values are separately allocated so that the user of the
     RPC module may free them as soon as they are not needed anymore. */
  /* TODO: This can segfault if the packet is maligned. */
  pos = 0;
  returnParamCount = (unsigned char) *(returnPacket + pos);
  pos++;
  for (n = 0; n < returnParamCount; n++) {
    paramName = returnPacket + pos;
    pos += strlen (returnPacket + pos) + 1;
    paramValueLength = ntohl (*((unsigned long *) (returnPacket + pos)));
    pos += sizeof (unsigned long);
    paramValue = returnPacket + pos;
    pos += paramValueLength;
    RPC_paramAdd (returnParam, paramName, (size_t) paramValueLength, 
                  paramValue);
  }
  FREE (returnPacket);

  /*TODO: Check the return parameter _STATUS and return a value accordingly.*/
  MUTEX_UNLOCK (&rpcLock);
  return RPC_RESULT_SUCCESS;
}


void RPC_register (char *name,
                   void (*callback) (const RPC_Param *, RPC_Param *)) { 
  RegisteredRPC *rrpc;

  MUTEX_LOCK (&rpcLock);

  for (rrpc = vectorGetFirst (registered);
       rrpc;
       rrpc = vectorGetNext (registered))
    if (!strcmp (rrpc->name, name)) {
      rrpc->callback = callback;
      LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_register() - Reset callback for RPC "
           "\"%s\"\n", name);
      break;
    }
  if (!rrpc) {
    rrpc = MALLOC (sizeof (RegisteredRPC));
    rrpc->name = name;
    rrpc->callback = callback;
    vectorInsertLast (registered, rrpc);
    LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_register() - Registered RPC \"%s\"\n",
         name);
  }
  LOG (LOG_DEBUG, "DEBUG: rpc.c::RPC_register() - There are now %d RPC's "
       "registered locally\n", vectorSize (registered));

  MUTEX_UNLOCK (&rpcLock);
  return;
}
