/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
*/

/**
 * This file provides the implementation of remote procedure calls over GNUnet.
 *
 * @file module/rpc.h
 * @brief Implementation of RPC's
 * @author Antti Salonen
 * @version $Id: rpc.h,v 1.17 2004/05/12 09:58:45 aksalone Exp $
 */

#ifndef RPC_H
#define RPC_H

#include <gnunet_core.h> /* HostIdentity */
#include "vector.h"

/* 
 * Return values of RPC_execute()
 */
#define RPC_RESULT_SUCCESS 0
#define RPC_RESULT_RECEIVER_UNKNOWN 1
#define RPC_RESULT_RECEIVER_UNREACHABLE 2
#define RPC_RESULT_NETWORKING_PROBLEM 3
#define RPC_RESULT_RPC_UNREGISTERED 4
#define RPC_RESULT_TIMEOUT 5

/**
 * A set of parameters to an RPC. The parameter structure is modifiable and
 * accessable with a set of functions. The parameters are stored in an array
 * and can be accessed in an indexed fashion, but generally there's no reason
 * to allocate a certain parameter to a certain position. The indexed access
 * is meant to allow iteration over a set of parameters, nothing more.
 */
typedef struct {
  Vector *vector;
} RPC_Param;

RPC_Param *RPC_paramNew (void);
void RPC_paramFree (RPC_Param *param);
size_t RPC_paramCount (const RPC_Param *param);
void RPC_paramAdd (const RPC_Param *param, const char *name, 
                   size_t dataLength, const void *data);
char *RPC_paramName (const RPC_Param *param, int i);
void *RPC_paramValueByName (const RPC_Param *param, const char *name);
size_t RPC_paramValueLengthByName (const RPC_Param *param, const char *name);
void *RPC_paramValueByPosition (const RPC_Param *param, int i);
size_t RPC_paramValueLengthByPosition (const RPC_Param *param, int i);
/* TODO: Remove by name, remove by position? */

int RPC_execute (const HostIdentity *receiver, 
                 const char *name,
                 const RPC_Param *request_param,
                 RPC_Param *return_param,
                 int timeout);

void RPC_register (char *name,
                   void (*callback) (const RPC_Param *, RPC_Param *));

/*
 * Initialization of RPC.
 * @param capi 
 */
int RPC_init(CoreAPIForApplication *capi);

/*
 * Tearing down RPC.
 * @param capi 
 */
void RPC_done(CoreAPIForApplication *capi);

#endif /* RPC_H */
