/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
*/

/**
 * An implementation of a dynamic array of objects. Like an array, the 
 * vector's elements are indexed, but it is also possible to dynamically
 * resize the vector by inserting and removing elements at any location.
 *   The vector is implemented as a double-linked list of arrays, each with a
 * static maximum length. When one array fills up, it's splitted into two
 * half-full arrays, and so forth. With functions {insert,get,remove}_last
 * the vector can also be used as a fairly efficient stack.
 *   The functions get_{at,first,last,next,previous} allow traversing the
 * vector in an efficient manner, each function call taking more or less
 * constant time. Vector_get_next and Vector_get_first may only be called
 * after a call to one of vector_get_{first,last,at}, which set the vector's
 * iterator. All functions that modify the vector's contents unset the
 * iterator.
 *
 * @file module/vector.c
 * @brief Implementation of a dynamic array
 * @author Antti Salonen
 * @version $Id: vector.c,v 1.3 2004/05/02 20:22:52 aksalone Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <gnunet_util.h>
#include "vector.h"


/* 
 * Clear an element within the segment's array by moving the trailing pointers
 * forward by one position.
 */
static void vectorSegmentNewIndex (VectorSegment *vs, int index) {
   int i;

   for (i = vs->size; i > index; i--)
      vs->data[i] = vs->data[i - 1];
}


/*
 * Remove and return the element at given index in the segment's array. The
 * trailing pointers in the array, if any, are moved backwards to fill the gap.
 */
static void *vectorSegmentRemoveIndex (VectorSegment *vs, int index) {
   void *rvalue = vs->data[index];

   for (; index < vs->size; index++)
      vs->data[index] = vs->data[index + 1];
   return rvalue;
}


/*
 * Split the full segment vs into two half-full segments.
 */
static void vectorSegmentSplit (Vector *v, VectorSegment *vs) {
   VectorSegment *oldNext = vs->next;

   vs->next = MALLOC (sizeof (VectorSegment));
   vs->next->previous = vs;
   vs->next->next = oldNext;
   if (oldNext)
      vs->next->next->previous = vs->next;
   if (!oldNext)
      v->segmentsTail = vs->next;
   (void) memcpy (vs->next->data, vs->data + (VECTOR_SEGMENT_SIZE >> 1), 
                  (VECTOR_SEGMENT_SIZE >> 1) * sizeof (void *));
   vs->size = vs->next->size = VECTOR_SEGMENT_SIZE >> 1;
}


/* 
 * Joins the given segment with the following segment. The first segment _must_
 * be empty enough to store the data of both segments.
 */
static void vectorSegmentJoin (Vector *v, VectorSegment *vs) {
   VectorSegment *oldNext = vs->next->next;

   (void) memcpy (vs->data + vs->size, vs->next->data,
                  vs->next->size * sizeof (void *));
   vs->size += vs->next->size;
   FREE (vs->next);
   vs->next = oldNext;
   if (oldNext)
      vs->next->previous = vs;
   if (!oldNext)
      v->segmentsTail = vs;
}


/* 
 * Free an empty segment, _unless_ it is the only segment.
 */
static void vectorSegmentRemove (Vector *v, VectorSegment *vs) {
   if (!vs->previous && !vs->next)
      return;
   if (vs->previous)
      vs->previous->next = vs->next;
   else
      v->segmentsHead = vs->next;
   if (vs->next)
      vs->next->previous = vs->previous;
   else
      v->segmentsTail = vs->previous;
   FREE (vs);
   return;
}


/*
 * Search for given index in the vector v. When the index is found, its
 * segment and relative index are written to parameters vs and segment_index.
 * If possible, an unused index at the end of a segment is returned, as this
 * is also a requirement for adding data in an empty vector.
 */
static void vectorFindNewIndex (Vector *v, int index, VectorSegment **vs,
                                   int *segmentIndex) {
   VectorSegment *segment;
   int segmentStartIndex;
   
   if (index <= v->size >> 1) { /* empty vector included */
      segmentStartIndex = 0;
      for (segment = v->segmentsHead; ; segment = segment->next) {
         if (index <= segmentStartIndex + segment->size)
            break;
         segmentStartIndex += segment->size;
      } 
   }
   else { /* reverse */
      segmentStartIndex = v->size;
      for (segment = v->segmentsTail; ; segment = segment->previous) {
         segmentStartIndex -= segment->size;
         /* Note that we do not need to check whether we're in the first
            segment at this point, because index 0 is never reverse searched
            in the first place. */
         if (index > segmentStartIndex)
            break;
      }
   }
   *vs = segment;
   *segmentIndex = index - segmentStartIndex;
   return;
}


/*
 * Same as above, but only returns segment indices that are actually in use.
 */
static void vectorFindIndex (Vector *v, int index, VectorSegment **vs,
                               int *segmentIndex) {
   VectorSegment *segment;
   int segmentStartIndex;

   if (index < v->size >> 1) {
      segmentStartIndex = 0;
      for (segment = v->segmentsHead; ; segment = segment->next) {
         if (index < segmentStartIndex + segment->size)
            break;
         segmentStartIndex += segment->size;
      } 
   }
   else {
      segmentStartIndex = v->size;
      for (segment = v->segmentsTail; ; segment = segment->previous) {
         segmentStartIndex -= segment->size;
         if (index >= segmentStartIndex) 
            break;
      }
   }
   *vs = segment;
   *segmentIndex = index - segmentStartIndex;
   return;
}


/*  
 * Traverse the vector looking for a given object. When found, set the pointer
 * pointed to by vs to point to the object's segment and the integer pointed
 * to by segmentIndex to the object's index in the segment. If the object is 
 * not found, *vs is set to NULL.
 */
static void vectorFindObject (Vector *v, void *object, VectorSegment **vs,
                               int *segmentIndex) {
   VectorSegment *segment;
   int i;

   for (segment = v->segmentsHead; segment; segment = segment->next)
      for (i = 0; i < segment->size; i++)
         if (segment->data[i] == object) {
            *vs = segment;
            *segmentIndex = i;
            return;
         }
   *vs = NULL;
   return;
}


/* 
 * Allocate a new vector structure with a single empty data segment.
 */
Vector *vectorNew (void) {
   Vector *rvalue;

   rvalue = MALLOC (sizeof (Vector));
   rvalue->size = 0;
   rvalue->segmentsHead = MALLOC (sizeof (VectorSegment));
   rvalue->segmentsTail = rvalue->segmentsHead;
   rvalue->segmentsHead->next = NULL;
   rvalue->segmentsHead->previous = NULL;
   rvalue->segmentsHead->size = 0;
   rvalue->iteratorSegment = NULL;
   rvalue->iteratorIndex = 0; /* redundant */
   MUTEX_CREATE_RECURSIVE (&rvalue->lock);
   return rvalue;
}


/*
 * Free vector structure including its data segments, but _not_ including the
 * stored void pointers. It is the user's responsibility to empty the vector
 * when necessary to avoid memory leakage.
 */
void vectorFree (Vector *v) {
   VectorSegment 
      *vs,
      *vsNext;

   MUTEX_LOCK (&v->lock);
   vs = v->segmentsHead;
   while (1) {
      vsNext = vs->next;
      FREE (vs);
      if (vsNext)
         vs = vsNext;
      else
         break;
   }
   MUTEX_UNLOCK (&v->lock);
   /* TODO: WTF happens if at this point somebody else is waiting on the 
      mutex? Perhaps whoever destroys this structure should be certain that
      nobody else is currently or in the future trying to use it. */
   MUTEX_DESTROY (&v->lock);
   FREE (v);
}


void vectorLock (Vector *v) {
  MUTEX_LOCK (&v->lock);
}


void vectorUnlock (Vector *v) {
  MUTEX_UNLOCK (&v->lock);
}


/*
 * Return the size of the vector.
 */
size_t vectorSize (Vector *v) {
   return v->size;
}


/*
 * Insert a new element in the vector at given index. The return value is
 * 1 on success, 0 if the index is out of bounds.
 */
int vectorInsertAt (Vector *v, void *object, int index) {
   VectorSegment *segment;
   int segmentIndex;

   MUTEX_LOCK (&v->lock);
   if (index < 0 || index > v->size) {
      MUTEX_UNLOCK (&v->lock);
      return 0;
   }
   v->iteratorSegment = NULL;
   vectorFindNewIndex (v, index, &segment, &segmentIndex);
   vectorSegmentNewIndex (segment, segmentIndex);
   segment->data[segmentIndex] = object;
   if (++segment->size == VECTOR_SEGMENT_SIZE)
      vectorSegmentSplit (v, segment);
   v->size++;
   MUTEX_UNLOCK (&v->lock);
   return 1;
}


/*
 * Insert a new element at the end of the vector.
 */
void vectorInsertLast (Vector *v, void *object) {
   MUTEX_LOCK (&v->lock);
   v->iteratorSegment = NULL;
   v->segmentsTail->data[v->segmentsTail->size++] = object;
   if (v->segmentsTail->size == VECTOR_SEGMENT_SIZE)
      vectorSegmentSplit (v, v->segmentsTail);
   v->size++;
   MUTEX_UNLOCK (&v->lock);
}


/*
 * Return the element at given index in the vector or NULL if the index is out
 * of bounds. The iterator is set to point to the returned element.
 */
void *vectorGetAt (Vector *v, int index) {
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (index < 0 || index >= v->size) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   vectorFindIndex (v, index, &v->iteratorSegment, &v->iteratorIndex);
   rvalue = v->iteratorSegment->data[v->iteratorIndex];
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/* 
 * Return the first element in the vector, whose index is 0, or NULL if the
 * vector is empty. The iterator of the vector is set to point to the first
 * element.
 */
void *vectorGetFirst (Vector *v) {
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (v->size == 0) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   v->iteratorSegment = v->segmentsHead;
   v->iteratorIndex = 0;
   rvalue = v->iteratorSegment->data[v->iteratorIndex];
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Return the last element in the vector or NULL if the vector is empty. The
 * iterator of the vector is set to point to the last element.
 */
void *vectorGetLast (Vector *v) {
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (v->size == 0) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   v->iteratorSegment = v->segmentsTail;
   v->iteratorIndex = v->segmentsTail->size - 1;
   rvalue = v->iteratorSegment->data[v->iteratorIndex];
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/* 
 * Return the next element in the vector, as called after vector_get_at() or
 * vector_get_first(). The return value is NULL if there are no more elements
 * in the vector or if the iterator has not been set. 
 */
void *vectorGetNext (Vector *v) {
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (v->iteratorSegment == NULL) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   if (++v->iteratorIndex >= v->iteratorSegment->size) {
      if (v->iteratorSegment == v->segmentsTail) {
         v->iteratorIndex--;
         MUTEX_UNLOCK (&v->lock);
         return NULL;
      } else {
         v->iteratorSegment = v->iteratorSegment->next;
         v->iteratorIndex = 0;
      }
   }
   rvalue = v->iteratorSegment->data[v->iteratorIndex];
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Return the previous element in the vector, as called after vector_get_at()
 * or vector_get_last(). The return value is NULL if there are no more
 * elements in the vector or if the iterator has not been set.
 */
void *vectorGetPrevious (Vector *v) {
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (v->iteratorSegment == NULL) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   if (--v->iteratorIndex == -1) {
      if (v->iteratorSegment == v->segmentsHead) {
         v->iteratorIndex++;
         MUTEX_UNLOCK (&v->lock);
         return NULL;
      } else {
         v->iteratorSegment = v->iteratorSegment->previous;
         v->iteratorIndex = v->iteratorSegment->size - 1;
      }
   }
   rvalue = v->iteratorSegment->data[v->iteratorIndex];
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Delete and return the element at given index. NULL is returned if index is 
 * out of bounds. 
 */
void *vectorRemoveAt (Vector *v, int index) {
   VectorSegment *segment;
   int segmentIndex;
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (index < 0 || index >= v->size) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   v->iteratorSegment = NULL;
   vectorFindIndex (v, index, &segment, &segmentIndex);
   rvalue = vectorSegmentRemoveIndex (segment, segmentIndex);
   /* If the segment ends empty remove it, otherwise join it if necessary. */
   if (--segment->size == 0)
      vectorSegmentRemove (v, segment);
   else if (segment->next && 
            segment->size + segment->next->size < VECTOR_SEGMENT_SIZE >> 1)
      vectorSegmentJoin (v, segment);
   else if (segment->previous &&
            segment->size + segment->previous->size < VECTOR_SEGMENT_SIZE >> 1)
      vectorSegmentJoin (v, segment->previous);
   v->size--;
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Delete and return the last element in the vector, or NULL if the vector
 * is empty.
 */
void *vectorRemoveLast (Vector *v) {
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (v->size == 0) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   v->iteratorSegment = NULL;
   rvalue = v->segmentsTail->data[v->segmentsTail->size - 1];
   /* If the segment ends empty remove it, otherwise join it if necessary. */
   if (--v->segmentsTail->size == 0)
      vectorSegmentRemove (v, v->segmentsTail);
   else if (v->segmentsTail->previous &&
            v->segmentsTail->size + v->segmentsTail->previous->size <
            VECTOR_SEGMENT_SIZE >> 1)
      vectorSegmentJoin (v, v->segmentsTail->previous);
   v->size--;
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Delete and return given object from the vector, or return NULL if the object
 * is not found.
 */
void *vectorRemoveObject (Vector *v, void *object) {
   VectorSegment *segment;
   int segmentIndex;
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   v->iteratorSegment = NULL;
   vectorFindObject (v, object, &segment, &segmentIndex);
   if (segment == NULL) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   rvalue = vectorSegmentRemoveIndex (segment, segmentIndex);
   /* If the segment ends empty remove it, otherwise join it if necessary. */
   if (--segment->size == 0)
      vectorSegmentRemove (v, segment);
   else if (segment->next && 
            segment->size + segment->next->size < VECTOR_SEGMENT_SIZE >> 1)
      vectorSegmentJoin (v, segment);
   else if (segment->previous &&
            segment->size + segment->previous->size < VECTOR_SEGMENT_SIZE >> 1)
      vectorSegmentJoin (v, segment->previous);
   v->size--;
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Set the given index in the vector. The old value of the index is 
 * returned, or NULL if the index is out of bounds.
 */
void *vectorSetAt (Vector *v, void *object, int index) {
   VectorSegment *segment;
   int segmentIndex;
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   if (index < 0 || index >= v->size) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   v->iteratorSegment = NULL;
   vectorFindIndex (v, index, &segment, &segmentIndex);
   rvalue = segment->data[segmentIndex];
   segment->data[segmentIndex] = object;
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/*
 * Set the index occupied by the given object to point to the new object.
 * The old object is returned, or NULL if it's not found.
 */
void *vectorSetObject (Vector *v, void *object, void *oldObject) {
   VectorSegment *segment;
   int segmentIndex;
   void *rvalue;

   MUTEX_LOCK (&v->lock);
   v->iteratorSegment = NULL;
   vectorFindObject (v, oldObject, &segment, &segmentIndex);
   if (segment == NULL) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   rvalue = segment->data[segmentIndex];
   segment->data[segmentIndex] = object;
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


/* 
 * Swaps the contents of index1 and index2. Return value is non-zero on
 * success, zero if either index is out of bounds.
 */
int vectorSwap (Vector *v, int index1, int index2) {
   VectorSegment 
      *segment1,
      *segment2;
   int 
      segmentIndex1,
      segmentIndex2;
   void *temp;

   MUTEX_LOCK (&v->lock);
   if (index1 < 0 || index1 >= v->size ||
       index2 < 0 || index2 >= v->size) {
      MUTEX_UNLOCK (&v->lock);
      return 0;
   }
   v->iteratorSegment= NULL;
   vectorFindIndex (v, index1, &segment1, &segmentIndex1);
   vectorFindIndex (v, index2, &segment2, &segmentIndex2);
   temp = segment1->data[segmentIndex1];
   segment1->data[segmentIndex1] = segment2->data[segmentIndex2];
   segment2->data[segmentIndex2] = temp;
   MUTEX_UNLOCK (&v->lock);
   return 1;
}


/*
 * Return the index of given element or -1 if the element is not found.
 */
int vectorIndexOf (Vector *v, void *object) {
   VectorSegment *segment;
   int 
      i, 
      segmentStartIndex;

   MUTEX_LOCK (&v->lock);
   segmentStartIndex = 0;
   for (segment = v->segmentsHead; segment; segment = segment->next) {
      for (i = 0; i < segment->size; i++)
         if (segment->data[i] == object) {
            MUTEX_UNLOCK (&v->lock);
            return segmentStartIndex + i;
         }
      segmentStartIndex += segment->size;
   }
   MUTEX_UNLOCK (&v->lock);
   return -1;
}


/* 
 * Return the data stored in the vector as a single dynamically allocated
 * array of (void *), which must be free(3)d by the user. Use the functions
 * get_{at,first,last,next,previous} instead, unless you really need to access
 * everything in the vector as fast as possible.
 */
void **vectorElements (Vector *v) {
   void **rvalue;
   VectorSegment *vs;
   size_t i = 0;

   MUTEX_LOCK (&v->lock);
   if (!(rvalue = MALLOC (v->size * sizeof (void *)))) {
      MUTEX_UNLOCK (&v->lock);
      return NULL;
   }
   for (vs = v->segmentsHead; vs; vs = vs->next) {
      (void) memcpy (rvalue + i, vs->data, vs->size * sizeof (void *));
      i += vs->size;
   }
   MUTEX_UNLOCK (&v->lock);
   return rvalue;
}


#ifdef INCLUDE_MAIN
/*
 * A debug function that traverses the linked list and prints the sizes of the
 * segments. This currently isn't used even by main.
 */
#if 0
static void vectordebug (Vector *v) {
   VectorSegment *vs;
   int 
      n = 0,
      sum = 0;
   for (vs = v->segments_head; vs; vs = vs->next) {
      fprintf (stderr, "%4d: %3d / %d: %d...%d\n", n++, vs->size, 
               VECTOR_SEGMENT_SIZE, sum, sum + vs->size - 1);
      sum += vs->size;
   }
   fprintf (stderr, "Vector size: %d\n", sum);
}
#endif

Semaphore *s1, *s2;

void *worker1 (Vector *v) {
  int n;

  while ((n = vectorSize (v)) < 100000)
    if (!vectorInsertAt (v, "kala", rand() % (n + 1)))
      fputs ("1: Insertion failure\n", stderr);
    else 
      printf ("1: Insert\n");

  printf ("1: Waiting\n");
  SEMAPHORE_DOWN (s1);
  printf ("1: Signaling\n");
  SEMAPHORE_UP (s2);
  
  while ((n = vectorSize (v)) > 0)
    if (!vectorRemoveAt (v, rand() % n))
      fputs ("1: Deletion failure\n", stderr);
    else
      printf ("1: Delete\n");
}

void *worker2 (Vector *v) {
  int n;

  while ((n = vectorSize (v)) < 100000)
    if (!vectorInsertAt (v, "kala", rand() % (n + 1)))
      fputs ("2: Insertion failure\n", stderr);
    else
      printf ("2: Insert\n");
  
  printf ("2: Signaling\n");
  SEMAPHORE_UP (s1);
  printf ("2: Waiting\n");
  SEMAPHORE_DOWN (s2);
  
  while ((n = vectorSize (v)) > 0)
    if (!vectorRemoveAt (v, rand() % n))
      fputs ("2: Deletion failure\n", stderr);
    else
      printf ("2: Delete\n");
}

int main (void) {
  pthread_t thread1, thread2;
  Vector *v;
  void *unused;
   
  v = vectorNew(); 
  
  s1 = SEMAPHORE_NEW (0);
  s2 = SEMAPHORE_NEW (0);

  PTHREAD_CREATE (&thread1, worker1, v, 128 * 1024);
  PTHREAD_CREATE (&thread2, worker2, v, 128 * 1024);
  PTHREAD_JOIN (thread1, &unused);
  PTHREAD_JOIN (thread2, &unused);

  SEMAPHORE_FREE (s1);
  SEMAPHORE_FREE (s2);

  vectorFree (v);
  return 0;
}
#endif
