/*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
*/

/**
 * An implementation of a dynamic array of objects. Like an array, the 
 * vector's elements are indexed, but it is also possible to dynamically
 * resize the vector by inserting and removing elements at any location.
 *   The vector is implemented as a double-linked list of arrays, each with a
 * static maximum length. When one array fills up, it's splitted into two
 * half-full arrays, and so forth. With functions {insert,get,remove}_last
 * the vector can also be used as a fairly efficient stack.
 *   The functions get_{at,first,last,next,previous} allow traversing the
 * vector in an efficient manner, each function call taking more or less
 * constant time. Vector_get_next and Vector_get_first may only be called
 * after a call to one of vector_get_{first,last,at}, which set the vector's
 * iterator. All functions that modify the vector's contents unset the
 * iterator.
 *
 * @file module/vector.h
 * @brief Implementation of a dynamic array
 * @author Antti Salonen
 * @version $Id: vector.h,v 1.3 2004/05/02 20:22:52 aksalone Exp $
 */

#ifndef VECTOR_H
#define VECTOR_H

#include <sys/types.h>
#include <gnunet_util.h>

/*
 * Size of the VectorSegment data area. The "correct" value for this is a bit
 * of a gamble, as it depends on both the operations you perform on the vectors
 * and how much data is stored in them. In general, the more data you store
 * the bigger the segments should be, or otherwise the increased length of the
 * linked list will become a bottleneck for operations that are performed on
 * arbitrary indexes.
 * 
 * TODO: There really should be a smart way to handle this runtime.
 */
#ifndef VECTOR_SEGMENT_SIZE
#define VECTOR_SEGMENT_SIZE 50
#endif

struct vector_segment_t {
   void *data[VECTOR_SEGMENT_SIZE];
   struct vector_segment_t *next;
   struct vector_segment_t *previous;
   size_t size;
};
typedef struct vector_segment_t VectorSegment;

struct vector_t {
   VectorSegment *segmentsHead;
   VectorSegment *segmentsTail;
   VectorSegment *iteratorSegment;
   int iteratorIndex;
   size_t size;
   Mutex lock;
};
typedef struct vector_t Vector;

Vector *vectorNew (void);
void vectorFree (Vector *v);
void vectorLock (Vector *v);
void vectorUnlock (Vector *v);
size_t vectorSize (Vector *v);
int vectorInsertAt (Vector *v, void *object, int index);
void vectorInsertLast (Vector *v, void *object);
void *vectorGetAt (Vector *v, int index);
void *vectorGetFirst (Vector *v);
void *vectorGetLast (Vector *v);
void *vectorGetNext (Vector *v);
void *vectorGetPrevious (Vector *v);
void *vectorRemoveAt (Vector *v, int index);
void *vectorRemoveLast (Vector *v);
void *vectorRemoveObject (Vector *v, void *object);
void *vectorSetAt (Vector *v, void *object, int index);
void *vectorSetObject (Vector *v, void *object, void *old_object);
int vectorSwap (Vector *v, int index1, int index2);
int vectorIndexOf (Vector *v, void *object);
void **vectorElements (Vector *v);

#endif
